//#####################################################################
// Copyright 2015, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Basic_Geometry/SEGMENT_2D.h>
#include <Core/Basic_Geometry/TRIANGLE_3D.h>
#include <Core/Boundaries/BOUNDARY_PERIODIC.h>
#include <Core/Boundaries/PERIODIC_BOUNDARY_HELPER.h>
#include <Core/Data_Structures/DEPTH_FIRST_SEARCH.h>
#include <Core/Data_Structures/HASHTABLE.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/CELL_ITERATOR_THREADED.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/FACE_ITERATOR_THREADED.h>
#include <Core/Grids/FACE_RANGE_ITERATOR.h>
#include <Core/Grids_Uniform_Computations/MARCHING_CUBES.h>
#include <Core/Interpolation/CUBIC_INTERPOLATION.h>
#include <Core/Interpolation/DIV_FREE_INTERPOLATION.h>
#include <Core/Interpolation/INTERPOLATED_COLOR_MAP.h>
#include <Core/Interpolation/LINEAR_INTERPOLATION_FACE.h>
#include <Core/Interpolation/WENO_INTERPOLATION.h>
#include <Core/Krylov_Solvers/CONJUGATE_GRADIENT.h>
#include <Core/Krylov_Solvers/KRYLOV_SOLVER.h>
#include <Core/Krylov_Solvers/KRYLOV_VECTOR_WRAPPER.h>
#include <Core/Krylov_Solvers/MATRIX_SYSTEM.h>
#include <Core/Level_Sets/EXTRAPOLATION_HIGHER_ORDER.h>
#include <Core/Level_Sets/EXTRAPOLATION_HIGHER_ORDER_FACE.h>
#include <Core/Level_Sets/FAST_MARCHING_METHOD_UNIFORM.h>
#include <Core/Level_Sets/LEVELSET_UTILITIES.h>
#include <Core/Log/DEBUG_SUBSTEPS.h>
#include <Core/Log/FINE_TIMER.h>
#include <Core/Log/SCOPE.h>
#include <Core/Matrices/DIAGONAL_MATRIX.h>
#include <Core/Matrices/LAPACK.h>
#include <Core/Matrices/SPARSE_MATRIX_THREADED_CONSTRUCTION.h>
#include <Core/Matrices/SYMMETRIC_MATRIX.h>
#include <Core/Matrices/SYSTEM_MATRIX_HELPER.h>
#include <Core/Ordinary_Differential_Equations/RUNGEKUTTA.h>
#include <Core/Parallel_Computation/APPEND_HOLDER.h>
#include <Core/Poisson/HEAT_LAPLACE.h>
#include <Core/Poisson/HELMHOLTZ.h>
#include <Core/Poisson/LAPLACE.h>
#include <Core/Projection/CUT_CELL_PROJECTION.h>
#include <Core/Projection/LAPLACE_COLLIDABLE_UNIFORM.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Read_Write/OCTAVE_OUTPUT.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_PARTICLES.h>
#include <Hybrid_Methods/Iterators/GATHER_SCATTER.h>
#include <Hybrid_Methods/Iterators/PARTICLE_GRID_FACE_ITERATOR.h>
#include <Hybrid_Methods/Iterators/PARTICLE_GRID_WEIGHTS.h>
#include <Hybrid_Methods/Projection/MPM_PROJECTION_SYSTEM.h>
#include <Hybrid_Methods/Projection/MPM_PROJECTION_VECTOR.h>
#include "MPM_MAC_DRIVER.h"
#include "MPM_MAC_EXAMPLE.h"
#ifdef USE_OPENMP
#include <omp.h>
#endif
using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class TV> MPM_MAC_DRIVER<TV>::
MPM_MAC_DRIVER(MPM_MAC_EXAMPLE<TV>& example)
    :example(example),
    sol(*new KRYLOV_VECTOR_WRAPPER<T,ARRAY<T> >),
    rhs(*new KRYLOV_VECTOR_WRAPPER<T,ARRAY<T> >),
    poisson_tmp(*new KRYLOV_VECTOR_WRAPPER<T,ARRAY<T> >)
{
    DEBUG_SUBSTEPS::write_substeps_level=example.write_substeps_level;
    DEBUG_SUBSTEPS::writer=[=,this](const std::string& title,int level){this->example.viewer_output.Write_Substep(title,level);};
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> MPM_MAC_DRIVER<TV>::
~MPM_MAC_DRIVER()
{
    projection_system.nullspace_vectors.Delete_Pointers_And_Clean_Memory();
    DEBUG_SUBSTEPS::writer=0;
    delete &sol;
    delete &rhs;
    delete &poisson_tmp;
}
//#####################################################################
// Execute_Main_Program
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Execute_Main_Program()
{
    Step([this](){Initialize();},"initialize");
    Simulate_To_Frame(example.last_frame);
    for(const auto& it:example.time_step_callbacks)
        if(!it.second.first)
            PHYSBAM_FATAL_ERROR("Unrecognized time step callback: "+it.first);
    FINE_TIMER::Dump_Timing_Info();
}
//#####################################################################
// Initialize
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Initialize()
{
    TIMER_SCOPE_FUNC;
    LOG::cout<<std::setprecision(16);
    DEBUG_SUBSTEPS::write_substeps_level=example.substeps_delay_frame<0?example.write_substeps_level:-1;

    if(example.auto_restart){
        example.viewer_dir.Read_Last_Frame(0);
        example.restart=example.viewer_dir.frame_stack(0);}
    else if(example.restart)
        example.viewer_dir.Set(example.restart);
    if(example.restart) example.viewer_dir.Make_Common_Directory(true);
    current_frame=example.restart;
    example.time=current_frame*example.frame_dt;

    example.Initialize();
    for(int i=0;i<TV::m;i++){
        bool a=example.side_bc_type(i*2)==MPM::BC_PERIODIC;
        bool b=example.side_bc_type(i*2+1)==MPM::BC_PERIODIC;
        PHYSBAM_ASSERT(a==b);
        example.periodic_boundary.is_periodic(i)=a;}

    // must have level sets in order to use level set projection
    if(example.use_level_set_projection) example.use_mls_xfers=true;
    if(example.use_mls_xfers && !example.interpolated_advection) example.use_object_extrap=false;

    if(example.xpic)
        example.particles.template Add_Array<TV>("effective_v",&example.effective_v);
    if(example.opt.flip)
        example.particles.template Add_Array<TV>("flip_adv_velocity",&example.flip_adv_velocity);
    if(example.use_bdf2 || example.adv_v_timepoint){
        example.particles.template Add_Array<TV>("V_prev");
        if(example.opt.use_affine){
            example.particles.template Add_Array<MATRIX<T,TV::m>>("C_prev");
        }
        if(example.opt.use_polypic) example.particles.template Add_Array<SYMMETRIC_TENSOR<T,0,TV::m,TV::m>>("H_prev");
    }
    if(example.init_prev&&(example.use_bdf2||example.adv_v_timepoint)) example.Initialize_Prev_Velocity_Fields();
    PHYSBAM_ASSERT(example.predictive_pressure || !example.ab_pressure);
    if(example.predictive_pressure){
        example.pressure.Resize(example.grid.numbers_of_cells.Product(),init_all,0);
        if(example.ab_pressure) example.pressure_prev.Resize(example.grid.numbers_of_cells.Product(),init_all,0);
        example.cell_index.Resize(example.grid.numbers_of_cells,init_all,0);
        for(CELL_ITERATOR<TV> it(example.grid);it.Valid();it.Next()){
            example.cell_index(it.index)=example.grid.numbers_of_cells(1)*it.index(0)+it.index(1);
            if(example.init_prev){
                example.pressure(example.cell_index(it.index))=example.Compute_Analytic_Pressure(it.Location(),0);
                if(example.ab_pressure) example.pressure_prev(example.cell_index(it.index))=example.Compute_Analytic_Pressure(it.Location(),-example.max_dt);
            }
        }
    }
    
    if(example.restart) example.Read_Output_Files();

    PHYSBAM_ASSERT(example.opt.use_affine || !example.opt.use_polypic);
    example.particles.Store_C(example.opt.use_affine);
    example.particles.Store_H(example.opt.use_polypic);

    if(example.particle_vort)
        example.particles.template Add_Array<typename TV::SPIN>("vort");

    example.gather_scatter=new GATHER_SCATTER<TV>(example.grid,example.simulated_particles);
    example.gather_scatter->threads=example.threads;
    example.gather_scatter->face_weights=example.weights;
    example.gather_scatter->weights=example.weights(0);

    if(example.use_phi){
        example.levelset.Set_Grid(example.grid.Get_MAC_Grid_At_Regular_Positions());
        example.phi.Resize(example.grid.Node_Indices(example.ghost));
        example.levelset.boundary=&example.periodic_boundary;}

    if(example.use_warm_start){
        example.pressure_save.Resize(example.grid.Cell_Indices(example.ghost));
        example.pressure_valid.Resize(example.grid.Cell_Indices(example.ghost));}

    if(example.output_face_fractions)
        face_fractions.Resize(example.grid.Cell_Indices(example.ghost),init_all,0);

    RANGE<IV> range(example.grid.Cell_Indices(example.ghost));

    example.velocity.u.Resize(example.density.m);
    example.mass.u.Resize(example.density.m);
    Update_Simulated_Particles();
    // Need grid velocities for initial advection step
    if(example.rk_particle_order || example.xpic){
        Update_Particle_Weights();
        Prepare_Scatter();
        if(!example.restart) Particle_To_Grid();
        if(example.xpic && !example.restart)
            Compute_Effective_Velocity();} // TODO: This is run before velocity_save is populated.  broken?
    if(example.opt.flip && !example.restart) example.flip_adv_velocity=example.particles.V;

    if(example.interpolated_advection)
    {
        example.particles.template Add_Array<TV>("advection_velocity");
        auto& advection_v=*example.particles.template Get_Array<TV>("advection_velocity");
        advection_v=example.particles.V;
    }

    if(example.write_output)
    {
        example.Register_Output_Files();
        if(!example.restart)
        {
            LOG::SCOPE scope("writing output files");
            example.viewer_output.Flush_Frame("");
        }
    }
    else example.viewer_output.viewer_dir.Make_Common_Directory();
}
//#####################################################################
// Function Advance_One_Time_Step
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Advance_One_Time_Step()
{
    TIMER_SCOPE_FUNC;
    Step([this](){Move_Particles();},"move-particles",true,example.enable_advection);
    Step([this](){Time_Integration();},"time-integration",true);
    Step([this](){Update_Simulated_Particles();},"simulated-particles",false);
    Step([this](){Update_Particle_Weights();},"update-weights",false);
    Step([this](){Prepare_Scatter();},"prepare-scatter",false);
    Step([this](){Compute_Regions();},"compute-regions",false);
    Step([this](){Particle_To_Grid();},"p2g");
    Step([this](){Build_Level_Sets();},"build-level-sets",false,example.use_phi);
    Step([this](){Reseeding();},"reseeding",true,example.use_reseeding);
    Step([this](){Apply_Forces();},"forces",true,example.enable_force);
    Step([this](){Prepare_Pressure_And_Viscosity();},"prepare p mu",true,example.enable_viscosity||example.enable_pressure_projection);
    Step([this](){Apply_Predictive_Pressure();},"predictive pressure",true,example.predictive_pressure&&example.enable_pressure_projection);
    Step([this](){Apply_Viscosity();},"viscosity",true,example.viscosity.Max()>0&&(!example.use_level_set_projection)&&example.enable_viscosity);
    Step([this](){Level_Set_Apply_Viscosity();},"level set viscosity",true,example.viscosity.Max()>0&&example.use_level_set_projection&&example.enable_viscosity);
    Step([this](){Pressure_Projection();},"projection",true,(!example.use_level_set_projection)&&example.enable_pressure_projection);
    Step([this](){Level_Set_Pressure_Projection();},"level set projection",true,example.use_level_set_projection&&example.enable_pressure_projection);
    Step([this](){Extrapolate_Inside_Object();},"extrapolate",true,example.use_object_extrap);
    Step([this](){Reflect_Boundary_Velocity_Copy_Only();},"velocity-extrapolation",true);
    Step([this](){Compute_Effective_Velocity();},"compute effective velocity",false,example.xpic);
    Step([this](){Grid_To_Particle();},"g2p",false,!example.opt.use_polypic);
    Step([this](){Grid_To_Particle_Polypic();},"g2p_least_squares",false,example.opt.use_polypic);
}
//#####################################################################
// Function Simulate_To_Frame
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Simulate_To_Frame(const int frame)
{
    TIMER_SCOPE_FUNC;
    for(;current_frame<frame;current_frame++){
        LOG::SCOPE scope("FRAME","frame %d",current_frame+1);
        for(int i=0;i<example.begin_frame.m;i++)
            example.begin_frame(i)(current_frame);
        if(example.substeps_delay_frame==current_frame)
            DEBUG_SUBSTEPS::write_substeps_level=example.write_substeps_level;
        T time_at_frame=example.time+example.frame_dt;
        bool done=false;
        for(int substep=0;!done;substep++){
            LOG::SCOPE scope("SUBSTEP","substep %d",substep+1);
            example.dt=Compute_Dt();
            example.dt=clamp(example.dt,example.min_dt,example.max_dt);
            T next_time=example.time+example.dt;
            if(next_time+(T).0001*example.dt>time_at_frame){ // Account for roundoff errors
                next_time=time_at_frame;
                done=true;}
            else if(next_time+example.dt>time_at_frame) next_time=(example.time+time_at_frame)/2;
            example.dt=next_time-example.time;
            LOG::cout<<"substep dt: "<<example.dt<<std::endl;

            Step([this](){Advance_One_Time_Step();},"time-step");
            PHYSBAM_DEBUG_WRITE_SUBSTEP("end substep %i",1,substep);
            LOG::cout<<"max velocity: "<<Max_Particle_Speed()<<"  bound: "<<Grid_V_Upper_Bound()<<std::endl;
            example.time=next_time;
            if(example.one_substep){
                Particle_To_Grid();
                break;
            }
        }
        for(int i=0;i<example.end_frame.m;i++)
            example.end_frame(i)(current_frame);
        if(example.write_output)
        {
            LOG::SCOPE scope("writing output files");
            example.viewer_output.Flush_Frame("");
        }
    }
}
//#####################################################################
// Function Update_Particle_Weights
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Update_Particle_Weights()
{
    TIMER_SCOPE_FUNC;
    for(int i=0;i<TV::m;++i)
        example.weights(i)->Update(example.particles.X);
}
//#####################################################################
// Function Particle_To_Grid
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Particle_To_Grid()
{
    TIMER_SCOPE_FUNC;
    const MPM_PARTICLES<TV>& particles=example.particles;

#pragma omp parallel for
    for(REGION_FACE_FLAT_ID i=REGION_FACE_FLAT_ID(0);i<example.mass.raw.m;i++){
        example.mass.raw(i)=0;
        example.velocity.raw(i)=0;}

    TIMER_SCOPE("Particle_To_Grid_Scatter");
    example.gather_scatter->Scatter(true,false,0,0,
        [this,&particles](int p,PARTICLE_GRID_FACE_ITERATOR<TV>& it,int data)
        {
            for(;it.Valid();it.Next())
            {
                T w=it.Weight();
                REGION_FACE_INDEX<TV::m> ri={example.particles.Region(p),it.Index()};
                example.mass(ri)+=w*particles.mass(p);
                T V=particles.V(p)(ri.f.axis);
                if(example.opt.use_affine)
                {
                    TV Y=example.grid.Face_X(ri.f)-particles.X(p);
                    TV dV=particles.C(p).Row(ri.f.axis);
                    if(example.opt.use_polypic) dV+=(T).5*(particles.H(p).x(ri.f.axis)*Y);
                    V+=dV.Dot(Y);
                }
                example.velocity(ri)+=w*particles.mass(p)*V;
            }
        });
        
    TIMER_SCOPE("Particle_To_Grid_Fixups");
    for(auto& a:example.mass.u) Fix_Periodic_Accum(a,PBH::ADD());
    for(auto& a:example.velocity.u) Fix_Periodic_Accum(a,PBH::ADD());
    
    if(example.xpic) example.mass_save=example.mass;

    if(!example.use_mls_xfers)
    {
        PHYSBAM_DEBUG_WRITE_SUBSTEP("before reflect",2);
        Reflect_Boundary_Mass_Momentum(example.velocity);
    }

    example.valid_indices.Remove_All();
    example.valid_flat_indices.Remove_All();

#pragma omp parallel for
    for(REGION_FACE_FLAT_ID i=REGION_FACE_FLAT_ID();i<example.mass.raw.m;i++)
    {
        if(example.mass(i)){
            example.velocity(i)/=example.mass(i);
            if(example.use_mls_xfers)
                example.valid_xfer_data(i)=true;}
        else{
            example.velocity(i)=0;
            if(example.use_mls_xfers)
                example.valid_xfer_data(i)=false;}
    }

    TIMER_SCOPE("Particle_To_Grid_Indices");
    APPEND_HOLDER<REGION_FACE_FLAT_ID> flat_h(example.valid_flat_indices);
    APPEND_HOLDER<REGION_FACE_INDEX<TV::m> > indices_h(example.valid_indices);
#pragma omp parallel
    {
#pragma omp single
        {
            flat_h.Init();
            indices_h.Init();
        }
#pragma omp barrier
        ARRAY<REGION_FACE_FLAT_ID>& flat_t=flat_h.Array();
        ARRAY<REGION_FACE_INDEX<TV::m> >& indices_t=indices_h.Array();
        for(REGION_ID r(0);r<example.mass.u.m;r++)
        {
            for(FACE_ITERATOR_THREADED<TV> it(example.grid,example.mass.u(r).domain);it.Valid();it.Next()){
                REGION_FACE_INDEX<TV::m> ri={r,it.Full_Index()};
                auto i=example.mass.Flat_Index(ri);
                if(example.mass(i)){
                    flat_t.Append(i);
                    indices_t.Append(ri);}}
        }
    }
    TIMER_SCOPE("Particle_To_Grid_Indices2");
    flat_h.Combine();
    indices_h.Combine();
    if(example.opt.flip || example.xpic)
        example.velocity_save=example.velocity;
}
//#####################################################################
// Function Build_Level_Sets
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Build_Level_Sets()
{
    TIMER_SCOPE_FUNC;
    PHYSBAM_ASSERT(example.use_phi);
    T dx=example.grid.dX.Max();
//    T fmm_offset=3*dx;
    int fill_radius_cells=5;
    T particle_radius=(T)0.55*sqrt(3+TV::m)*dx;
    example.phi.array.Fill(fill_radius_cells*dx-particle_radius);
    RANGE<IV> grid_domain=example.grid.Node_Indices(example.ghost);
    for(int k=0;k<example.simulated_particles.m;k++){
        int p=example.simulated_particles(k);
        TV X=example.particles.X(p);
        IV cell=example.grid.Node_IC(X);
        RANGE<IV> range=RANGE<IV>(cell).Thickened(fill_radius_cells).Intersect(grid_domain);
        for(RANGE_ITERATOR<TV::m> it(range);it.Valid();it.Next()){
            T d=(X-example.grid.Node_X(it.index)).Magnitude()-particle_radius;
            example.phi(it.index)=min(example.phi(it.index),d);}}
    Fix_Periodic(example.phi);

    // ARRAY<IV> seed_indices;
    // for(FACE_RANGE_ITERATOR<TV::m> it(example.grid.Node_Indices(example.ghost),RF::skip_outer);it.Valid();it.Next()){
    //     IV a=it.face.First_Cell_Index(),b=it.face.Second_Cell_Index();
    //     bool pa=example.phi(a)<0,pb=example.phi(b)<0;
    //     if(pa!=pb){
    //         seed_indices.Append(a);
    //         seed_indices.Append(b);}}
    // Prune_Duplicates(seed_indices);

    // FAST_MARCHING_METHOD_UNIFORM<TV> fmm;
    // fmm.seed_indices=&seed_indices;
    // fmm.correct_interface_phi=false;
    // fmm.Fast_Marching_Method(example.levelset.Native_Grid(),example.ghost,example.phi,
    //     fmm_offset+example.ghost*dx);
    // example.phi.array+=fmm_offset-particle_radius;
    // Fix_Periodic(example.phi);
}
//#####################################################################
// Function Reseeding
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Reseeding()
{
    TIMER_SCOPE_FUNC;
    REGION_ID rid(0);
    PHYSBAM_ASSERT(example.velocity.u.m==REGION_ID(1));
    bool added_particle=false;
    T dx=example.grid.dX.Max();
    LINEAR_INTERPOLATION_FACE<TV,T> li;
    int target_ppc=1<<TV::m;
    ARRAY<ARRAY<int>,IV> particle_counts(example.grid.Cell_Indices());
    ARRAY_VIEW<VECTOR<T,3> >* color=example.particles.template Get_Array<VECTOR<T,3> >("color");
    for(int i=0;i<example.particles.X.m;i++){
        if(!example.particles.valid(i)) continue;
        TV X=example.particles.X(i);
        IV index=example.grid.Cell_IC(X);
        particle_counts(index).Append(i);}
    T uniform_mass=example.particles.mass(0);
    auto* V_prev=example.particles.template Get_Array<TV>("V_prev");
    auto* C_prev=example.particles.template Get_Array<MATRIX<T,TV::m>>("C_prev");
    auto* H_prev=example.particles.template Get_Array<SYMMETRIC_TENSOR<T,0,TV::m,TV::m>>("H_prev");
    auto* vort=example.particles.template Get_Array<TV>("vort");
    for(CELL_ITERATOR<TV> it(example.grid);it.Valid();it.Next()){
        ARRAY<int>& a=particle_counts(it.index);
        int target=target_ppc;
        target=2;
        // if(abs(example.phi(it.index))<dx) target*=4;
        if(a.m<target){
            bool broad_phase_inside=false;
            for(int j=0;j<example.collision_objects.m;j++){
                MPM_COLLISION_OBJECT<TV>* o=example.collision_objects(j);
                RANGE<TV> box=example.grid.Cell_Domain(example.grid.Clamp_To_Cell(o->Get_Implicit_Object(example.time)->box));
                if(box.Lazy_Inside(example.grid.Cell_X(it.index))){
                    broad_phase_inside=true;
                    break;}}
            RANGE<TV> range=example.grid.Cell_Domain(it.index);
            for(int i=0,c=target-a.m;i<c;i++){
                TV X;
                example.random.Fill_Uniform(X,range);
                T phi=example.use_phi?example.levelset.Phi(X):-2*dx;
                bool inside_obj=false;
                for(int j=0;broad_phase_inside && j<example.collision_objects.m;j++){
                    MPM_COLLISION_OBJECT<TV>* o=example.collision_objects(j);
                    if(o->Phi(X,example.time)<=0){
                        inside_obj=true;
                        break;}}
                if(!inside_obj && abs(phi)>=(T).3*dx)
                {
                    int p=example.particles.Add_Element_From_Deletion_List();
                    example.particles.valid(p)=true;
                    (*color)(p)=VECTOR<T,3>(0,1,0);
                    added_particle=true;
                    example.particles.mass(p)=uniform_mass;
                    example.particles.V(p)=li.Interpolate_Face(example.grid,example.velocity.u(rid),X);
                    example.particles.X(p)=X;
                    if(example.particles.C.m) example.particles.C(p)={};
                    if(example.particles.H.m) example.particles.H(p)={};
                    if(V_prev && V_prev->m) (*V_prev)(p)=example.particles.V(p);
                    if(C_prev && C_prev->m) (*C_prev)(p)={};
                    if(H_prev && H_prev->m) (*H_prev)(p)={};
                    if(example.particles.region.m) example.particles.region(p)={};
                    if(example.particles.volume.m) example.particles.volume(p)=example.reseed_volume;
                    if(vort && vort->m) (*vort)(p)={};
                }}}
        if(a.m>2*(1<<TV::m)){
            example.random.Random_Shuffle(a);
            while(a.m>2*(1<<TV::m))
                Invalidate_Particle(a.Pop_Value());}}

    if(added_particle){
        Update_Simulated_Particles();
        Update_Particle_Weights();
        Prepare_Scatter();
        Particle_To_Grid();}
}
//#####################################################################
// Function Grid_To_Particle
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Grid_To_Particle_Polypic()
{
    TIMER_SCOPE_FUNC;
    example.gather_scatter->Gather(true,false,0,0,
        [this](int p,PARTICLE_GRID_FACE_ITERATOR<TV>& it,int)
        {
            TV h_V;
            MATRIX<T,TV::m> h_B,B;
            SYMMETRIC_TENSOR<T,0,TV::m> h_H;
            for(;it.Valid();it.Next())
            {
                T w=it.Weight();
                REGION_FACE_INDEX<TV::m> ri={example.particles.Region(p),it.Index()};
                T V=example.velocity(ri);
                h_V(ri.f.axis)+=w*V;
                TV Z=example.grid.Face_X(ri.f),dX=Z-example.particles.X(p);
                h_B.Add_Row(ri.f.axis,w*V*dX);
                h_H.x(ri.f.axis)+=w*V*Outer_Product(dX);
            }

            T d=1/example.weights(0)->Dp_inv;
            auto H=h_H/(2*d*d);
            auto v=h_V;
            for(int a=0;a<TV::m;a++)
            {
                TV e=it.scratch(a).m3,f=it.scratch(a).m4;
                TV g=e*e-d*f+d*d*d;
                for(int b=0;b<TV::m;b++)
                    H.x(a)(b,b)=(d*d*h_V(a)-d*h_H(a,b,b)+e(b)*h_B(a,b))/g(b);
                for(int b=0;b<TV::m;b++)
                    B(a,b)=((d*d-f(b))*h_B(a,b)-d*e(b)*h_V(a)+e(b)*h_H(a,b,b))/g(b);
                v(a)-=d*H.x(a).Trace();
            }

            example.particles.V(p)=v;
            example.particles.C(p)=B;
            example.particles.H(p)=H*2;
        });

    if(!example.is_first_timestep) example.is_first_2_timesteps=false;
    example.is_first_timestep=false;
}
//#####################################################################
// Function Grid_To_Particle
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Grid_To_Particle()
{
    TIMER_SCOPE_FUNC;
    MPM_PARTICLES<TV>& particles=example.particles;
    bool use_flip=example.opt.flip;
    auto* advection_v=example.particles.template Get_Array<TV>("advection_velocity");

    INTERPOLATION_FACE<TV, T>* interp = nullptr;
    if (example.div_free_advection) {
        interp = new DIV_FREE_INTERPOLATION<TV,T>();
    } else if (example.cubic_advection) {
        interp = new CUBIC_INTERPOLATION_FACE<TV,T>();
    } else if (example.cubic_advection) {
        interp = new LINEAR_INTERPOLATION_FACE<TV,T>();
    }
    
    example.gather_scatter->Gather(true,false,0,0,
        [this,use_flip,&particles,interp,advection_v](int p,PARTICLE_GRID_FACE_ITERATOR<TV>& it,int)
        {
            TV h_V;
            TV h_flip_V;
            MATRIX<T,TV::m> h_B;
            T h_a[TV::m]={};
            TV h_b[TV::m];
            SYMMETRIC_MATRIX<T,TV::m> h_G[TV::m];
            bool h_partial[TV::m]={};
            for(;it.Valid();it.Next())
            {
                T w=it.Weight();
                REGION_FACE_INDEX<TV::m> ri={example.particles.Region(p),it.Index()};
                if(example.use_mls_xfers && !example.valid_xfer_data(ri)){
                    h_partial[ri.f.axis]=true;
                    h_a[ri.f.axis]+=w;
                    if(example.opt.use_affine){
                        TV Z=example.grid.Face_X(ri.f),dX=Z-particles.X(p);
                        h_b[ri.f.axis]+=w*dX;
                        h_G[ri.f.axis]+=w*Outer_Product(dX);}
                    return;}
                T V=example.velocity(ri);
                h_V(ri.f.axis)+=w*V;
                if(use_flip) h_flip_V(ri.f.axis)+=w*(V-example.velocity_save(ri));
                if(example.opt.use_affine){
                    TV Z=example.grid.Face_X(ri.f),dX=Z-particles.X(p);
                    h_B.Add_Row(ri.f.axis,w*V*dX);}
            }
            if(example.use_mls_xfers)
                for(int d=0;d<TV::m;d++)
                    if(h_partial[d]){
                        T a=1-h_a[d];
                        if(!example.opt.use_affine){
                            if(a){
                                h_V(d)/=a;
                                if(use_flip) h_flip_V(d)/=a;}
                            else{
                                h_V(d)=0;
                                if(use_flip) h_flip_V(d)=0;}
                            continue;}
                        TV b=-h_b[d];
                        MATRIX<T,TV::m> G(example.weights(d)->Dp-h_G[d]);
                        MATRIX<T,TV::m+1> A;
                        A(TV::m,TV::m)=a;
                        for(int i=0;i<TV::m;i++)
                            A(TV::m,i)=A(i,TV::m)=b(i);
                        for(int i=0;i<TV::m;i++)
                            for(int j=0;j<TV::m;j++)
                                A(i,j)=G(i,j);
                        VECTOR<T,TV::m+1> rhs=h_B.Row(d).Append(h_V(d)),sol;
                        Least_Squares_Solve(A,rhs,sol);
                        h_V(d)=sol(TV::m);
                        h_B.Set_Row(d,example.weights(d)->Dp*sol.Remove_Index(TV::m));}

            if(particles.store_C) particles.C(p)=h_B*example.weights(0)->Dp_inv;
            if(use_flip){
                particles.V(p)=(1-example.opt.flip)*h_V+example.opt.flip*(particles.V(p)+h_flip_V);
                example.flip_adv_velocity(p)=h_V;}
            else if(example.xpic) particles.V(p)+=example.effective_v(p);
            else particles.V(p)=h_V;
            if(example.interpolated_advection)
            {
                REGION_ID i=example.particles.Region(p);
                const auto& V=example.velocity.u(i);
                if (example.second_order) {
                    TV X = example.particles.X(p);
                    TV v = interp->Interpolate_Face(example.grid,V,X);

                    TV X1 = X + 0.5 * v * example.dt;
                    (*advection_v)(p) = interp->Interpolate_Face(example.grid,V,X1);
                } else {
                    (*advection_v)(p)=interp->Interpolate_Face(example.grid,V,example.particles.X(p));
                }
            }
        });
        
    if(!example.is_first_timestep) example.is_first_2_timesteps=false;
    example.is_first_timestep=false;
}
//#####################################################################
// Function Compute_Effective_Velocity
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Compute_Effective_Velocity()
{
    TIMER_SCOPE_FUNC;
    example.xpic_v=example.velocity_save;
    example.xpic_v_star.Resize(example.xpic_v);

    MPM_PARTICLES<TV>& particles=example.particles;
    T s=1;
    example.xpic_v_star.raw.Fill(0);
    for(int r=2;r<=example.xpic;r++,s*=-1){
        T f=((T)example.xpic-r+1)/r;

        example.gather_scatter->Gather(false,false,0,0,
            [this](int p,PARTICLE_GRID_FACE_ITERATOR<TV>& it,int data)
            {
                TV ev;
                for(;it.Valid();it.Next())
                {                
                    T w=it.Weight();
                    REGION_FACE_INDEX<TV::m> ri={example.particles.Region(p),it.Index()};
                    T V=example.xpic_v(ri);
                    ev(ri.f.axis)+=w*V;
                }
                example.effective_v(p)=ev;
            });

        example.xpic_v.raw.Fill(0);
        example.gather_scatter->Scatter(false,false,0,0,
            [this,&particles,f](int p,PARTICLE_GRID_FACE_ITERATOR<TV>& it,int data)
            {
                for(;it.Valid();it.Next())
                {
                    T w=it.Weight();
                    REGION_FACE_INDEX<TV::m> ri={example.particles.Region(p),it.Index()};
                    T V=w*example.effective_v(p)(ri.f.axis);
                    example.xpic_v(ri)+=f*particles.mass(p)*V;
                }
            });
        for(auto& a:example.xpic_v.u) Fix_Periodic_Accum(a,PBH::ADD());
        
        if(!example.use_mls_xfers){
            example.mass=example.mass_save;
            PHYSBAM_DEBUG_WRITE_SUBSTEP("before reflect xpic",2);
            Reflect_Boundary_Mass_Momentum(example.xpic_v);}

        for(REGION_ID r(0);r<example.mass.u.m;r++)
        {
            for(FACE_ITERATOR_THREADED<TV> it(example.grid,example.mass.u(r).domain);it.Valid();it.Next())
            {
                REGION_FACE_INDEX<TV::m> ri={r,it.Full_Index()};
                auto i=example.mass.Flat_Index(ri);
                if(example.mass(i)){
                    example.xpic_v(i)/=example.mass(i);
                    example.xpic_v_star(i)+=s*example.xpic_v(i);}
                else example.xpic_v(i)=0;
            }
        }
    }

    example.gather_scatter->Gather(false,false,0,0,
        [this,&particles](int p,PARTICLE_GRID_FACE_ITERATOR<TV>& it,int data)
        {
            auto E=-particles.V(p);
            for(;it.Valid();it.Next())
            {
                T w=it.Weight();
                REGION_FACE_INDEX<TV::m> ri={example.particles.Region(p),it.Index()};
                T V=example.velocity(ri)+(example.xpic-1)*example.velocity_save(ri)-example.xpic*example.xpic_v_star(ri);
                E(ri.f.axis)+=w*V;
            }
            example.effective_v(p)=E;
        });
}
//#####################################################################
// Function Neumann_Boundary_Condition
//#####################################################################
template<class TV> bool MPM_MAC_DRIVER<TV>::
Neumann_Boundary_Condition(const FACE_INDEX<TV::m>& face,T& bc) const
{
    REGION_ID rid(0);
    PHYSBAM_ASSERT(example.velocity.u.m==REGION_ID(1));
    VECTOR<RANGE<IV>,TV::m> domains;
    for(int i=0;i<TV::m;i++){
        domains(i)=example.grid.Cell_Indices();
        domains(i).min_corner(i)++;}
    for(int i=0;i<TV::m;i++){
        bool wall=false;
        TV Y=example.grid.domain.Clamp(example.grid.Face_X(face));
        if(face.index(i)<domains(face.axis).min_corner(i)){
            wall=example.side_bc_type(2*i)==MPM::BC_SLIP||example.side_bc_type(2*i)==MPM::BC_NOSLIP;
            if(example.bc_type(2*i)){
                auto bc_type=example.bc_type(2*i)(Y,example.time);
                wall=bc_type==MPM::BC_SLIP||bc_type==MPM::BC_NOSLIP;}}
        else if(face.index(i)>=domains(face.axis).max_corner(i)){
            wall=example.side_bc_type(2*i+1)==MPM::BC_SLIP||example.side_bc_type(2*i+1)==MPM::BC_NOSLIP;
            if(example.bc_type(2*i+1)){
                auto bc_type=example.bc_type(2*i+1)(Y,example.time);
                wall=bc_type==MPM::BC_SLIP||bc_type==MPM::BC_NOSLIP;}}
        if(!wall) continue;
        if(example.mass.u(rid)(face)){
            bc=0;
            if(example.bc_velocity)
                bc=example.bc_velocity(Y,example.time)(face.axis);}
        return true;}
    TV X=example.grid.Face_X(face);
    for(int i=0;i<example.collision_objects.m;i++){
        MPM_COLLISION_OBJECT<TV>* o=example.collision_objects(i);
        if(o->Phi(X,example.time)<0){
            bc=o->Velocity(X,example.time)(face.axis);
            return true;}}
    return false;
}
//#####################################################################
// Function Compute_Boundary_Conditions
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Compute_Boundary_Conditions()
{
    TIMER_SCOPE_FUNC;
    auto domain=example.grid.Cell_Indices();
    example.psi_N.Resize(example.grid.Cell_Indices(example.ghost),init_all,false);
    for(int side=0;side<2*TV::m;side++)
    {
        bool have_psi_N=false,all_psi_N=false;
        auto side_bc_type=example.side_bc_type(side);
        if(side_bc_type==MPM::BC_SLIP || side_bc_type==MPM::BC_NOSLIP)
        {
            all_psi_N=have_psi_N=true;
            for(FACE_RANGE_ITERATOR<TV::m> it(example.psi_N.domain,
                    domain,RF::ghost,side);it.Valid();it.Next())
                example.psi_N(it.face)=true;
        }
        else if(example.bc_type(side))
            for(FACE_RANGE_ITERATOR<TV::m> it(example.psi_N.domain,
                    domain,RF::ghost,side);it.Valid();it.Next())
            {
                TV Y=example.grid.domain.Clamp(example.grid.Face_X(it.face));
                auto bc_type=example.bc_type(side)(Y,example.time);
                if(bc_type==MPM::BC_SLIP || bc_type==MPM::BC_NOSLIP)
                {
                    example.psi_N(it.face)=true;
                    have_psi_N=true;
                }
            }
        if(!have_psi_N) continue;
        // Assign velocity boundary conditions for psi_N faces.  Visit ghost
        // region portions of region domains to avoid the need for bounds checks
        // on the region arrays.
        for(REGION_ID r(0);r<example.mass.u.m;r++)
        {
            auto& mass=example.mass.u(r);
            auto& u=example.velocity.u(r);
            RANGE<IV> valid_region(example.psi_N.domain.Intersect(mass.domain));
            for(FACE_RANGE_ITERATOR<TV::m> it(valid_region,domain,
                    RF::ghost,side);it.Valid();it.Next())
            {
                if(mass(it.face) && (all_psi_N || example.psi_N(it.face)))
                {
                    T bc=0;
                    if(example.bc_velocity)
                    {
                        TV Y=example.grid.domain.Clamp(example.grid.Face_X(it.face));
                        bc=example.bc_velocity(Y,example.time)(it.face.axis);
                    }
                    u(it.face)=bc;
                }
            }
        }
    }

    Visit_Colliding_Region_Faces(
        [this](const REGION_FACE_INDEX<TV::m>& ri,MPM_COLLISION_OBJECT<TV>* obj,T phi)
        {
            example.psi_N(ri.f)=true;
            T bc=obj->Velocity(example.grid.Face_X(ri.f),example.time)(ri.f.axis);
            example.velocity(ri)=bc;
        });

    Fix_Periodic_Accum(example.psi_N,PBH::OR());
}
//#####################################################################
// Function Allocate_Pressure_Dofs
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Allocate_Pressure_Dofs()
{
    TIMER_SCOPE_FUNC;
    example.cell_index.Resize(example.grid.Cell_Indices(example.ghost),init_all,0);

    // Set the low 2*TV::m bits to record which of the surrounding faces have
    // mass from some region.
    for(REGION_ID r(0);r<example.mass.u.m;r++)
    {
        auto& mass=example.mass.u(r);
        for(FACE_RANGE_ITERATOR<TV::m> it(mass.domain);it.Valid();it.Next())
            if(mass(it.face))
            {
                example.cell_index(it.face.First_Cell_Index())|=1<<(it.face.axis*2+1);
                example.cell_index(it.face.Second_Cell_Index())|=1<<(it.face.axis*2);
            }
    }
    // Set the low 2*TV::m bits starting at 8 to record which of the surrounding
    // faces have psi_N set.
    for(FACE_RANGE_ITERATOR<TV::m> it(example.psi_N.domain,
            RF::skip_outer);it.Valid();it.Next())
        if(example.psi_N(it.face))
        {
            example.cell_index(it.face.First_Cell_Index())|=1<<(8+it.face.axis*2+1);
            example.cell_index(it.face.Second_Cell_Index())|=1<<(8+it.face.axis*2);
        }
    
    // Prevent Dirichlet cells from being registered as dofs by flagging them as
    // all Neumann.
    int side_mask=0;
    for(int i=0;i<2*TV::m;i++)
        if(example.side_bc_type(i)==MPM::BC_FREE)
            side_mask|=1<<i;
    for(RANGE_ITERATOR<TV::m> it(example.grid.Cell_Indices(),example.ghost,0,
            RI::ghost|RI::side_mask,side_mask);it.Valid();it.Next())
    {
        if(example.bc_type(it.side))
        {
            TV Y=example.grid.domain.Clamp(example.grid.Cell_X(it.index));
            if(example.bc_type(it.side)(Y,example.time)==MPM::BC_FREE)
                example.cell_index(it.index)=-1;
        }
        else example.cell_index(it.index)=-1;
    }
    Fix_Periodic_Accum(example.cell_index,PBH::OR());

    // A cell is a degree of freedom if (1) all surrounding faces have mass in
    // some region and (2) at least one of those faces is not Neumann.
    int all_faces=(1<<(2*TV::m))-1;
    int all_N=all_faces<<8;
    example.dof=0;
    for(RANGE_ITERATOR<TV::m> it(example.grid.Cell_Indices());it.Valid();it.Next())
    {
        auto& c=example.cell_index(it.index);
        if((c&all_faces)==all_faces && (c&all_N)!=all_N)
            c=example.dof++;
        else c=-1;
    }
    for(RANGE_ITERATOR<TV::m> it(example.cell_index.domain,example.grid.Cell_Indices(),RI::ghost);it.Valid();it.Next())
        example.cell_index(it.index)=-1;
    Fix_Periodic(example.cell_index);
}
//#####################################################################
// Function Compute_Poisson_Matrix
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Compute_Poisson_Matrix()
{
    TIMER_SCOPE_FUNC;
    if(example.dump_bc)
    {
        for(FACE_RANGE_ITERATOR<TV::m> it(example.psi_N.domain);it.Valid();it.Next())
            if(example.psi_N(it.face))
                Add_Point(example.grid.Face_X(it.face),{0,1,1});
        for(RANGE_ITERATOR<TV::m> it(example.cell_index.domain);it.Valid();it.Next())
            if(example.cell_index(it.index)<0)
                Add_Point(example.grid.Cell_X(it.index),{1,0,1});
    }
    face_volumes.Resize(example.grid.Cell_Indices(example.ghost),init_all,0);
    for(REGION_ID r(0);r<example.mass.u.m;r++)
    {
        auto& mass=example.mass.u(r);
        T density=example.density(r);
        for(FACE_RANGE_ITERATOR<TV::m> it(mass.domain);it.Valid();it.Next())
            face_volumes(it.face)+=mass(it.face)/density;
    }
    Fix_Periodic_Accum(face_volumes,PBH::ADD());

    gradient_data.Remove_All();
    region_starts.Resize(example.mass.u.m+1);
    region_starts(REGION_ID())=0;

    poisson.Reset(example.dof);
    poisson.m=example.dof;
    poisson.A.Resize((TV::m*2+1)*example.dof,init_all,{-1,0});
    poisson.offsets.Resize(example.dof+1);
    for(int i=0;i<=example.dof;i++)
        poisson.offsets(i)=i*(TV::m*2+1);

    ARRAY<char> is_dc(example.dof);
    ENTRY e;
    T velocity_scale=1;
    if(example.use_bdf2&&(example.init_prev||!example.is_first_timestep))
        velocity_scale=(T)1.5;
    for(REGION_ID r(0);r<example.mass.u.m;r++)
    {
        auto& mass=example.mass.u(r);
        T density=example.density(r);
        for(FACE_RANGE_ITERATOR<TV::m> it(mass.domain);it.Valid();it.Next())
        {
            e.face=example.mass.Flat_Index({r,it.face});
            T m=example.mass(e.face);
            if(!m)
            {
                if(example.dump_bc)
                    Add_Point(example.grid.Face_X(it.face),{1,.5,0});
                continue;
            }
            e.axis=it.face.axis;
            T entry=example.grid.one_over_dX(e.axis);
            e.dof(0)=example.cell_index(it.face.First_Cell_Index());
            e.dof(1)=example.cell_index(it.face.Second_Cell_Index());
            if(example.psi_N(it.face))
            {
                if(e.dof(0)>=0) rhs.v(e.dof(0))-=entry*example.velocity(e.face)*velocity_scale;
                if(e.dof(1)>=0) rhs.v(e.dof(1))+=entry*example.velocity(e.face)*velocity_scale;
                if(example.dump_bc)
                    Add_Point(example.grid.Face_X(it.face),{1,0,0});
                continue;
            }
            e.theta=m/(density*face_volumes(it.face));
            e.data={-entry,entry};
            if(e.dof.Max()<0)
            {
                if(example.dump_bc)
                    Add_Point(example.grid.Face_X(it.face),{1,1,0});
                continue;
            }
            if(example.dump_bc)
            {
                if(e.dof.Min()<0)
                    Add_Point(example.grid.Face_X(it.face),{0,1,0});
                else Add_Point(example.grid.Face_X(it.face),{0,1,1});
            }

            T sigma=e.theta/density;
            if(e.dof(0)>=0)
            {
                int i00=(TV::m*2+1)*e.dof(0)+TV::m;
                T d0=e.data(0)*sigma;
                poisson.A(i00).a+=e.data(0)*d0;
                poisson.A(i00).j=e.dof(0);
                if(e.dof(1)>=0)
                {
                    int i01=(TV::m*2+1)*e.dof(0)+2*TV::m-e.axis;
                    int i10=(TV::m*2+1)*e.dof(1)+e.axis;
                    T d01=e.data(1)*d0;
                    poisson.A(i01).a+=d01;
                    poisson.A(i10).a+=d01;
                    poisson.A(i01).j=e.dof(1);
                    poisson.A(i10).j=e.dof(0);
                }
            }
            else
            {
                // Ensure all dof indices are valid; if necessary, re-use the valid
                // index and give it zero coefficient.  This avoids the need to test
                // for valid indices later.
                e.dof(0)=e.dof(1);
                e.data(0)=0;
                is_dc(e.dof(1))=true;
            }
            if(e.dof(1)>=0)
            {
                int i11=(TV::m*2+1)*e.dof(1)+TV::m;
                T d1=e.data(1)*sigma;
                poisson.A(i11).a+=e.data(1)*d1;
                poisson.A(i11).j=e.dof(1);
            }
            else
            {
                e.dof(1)=e.dof(0);
                e.data(1)=0;
                is_dc(e.dof(0))=true;
            }
            gradient_data.Append(e);
        }
        region_starts(r+1)=gradient_data.m;
    }
    poisson.Prune_Zero_Entries();
    poisson.Sort_Entries();

    ARRAY<int> component_dofs;
    bool region_has_dc=false;
    projection_system.nullspace_vectors.Delete_Pointers_And_Clean_Memory();
    DEPTH_FIRST_SEARCH dfs(example.dof);
    dfs.Traverse(
        [&](int i)
        {
            for(int b=poisson.offsets(i),e=poisson.offsets(i+1);b<e;b++)
                dfs.Add_Neighbor(poisson.A(b).j);
            component_dofs.Append(i);
            if(is_dc(i)) region_has_dc=true;
        },
        [](int i){},
        [&]()
        {
            if(!region_has_dc)
            {
                auto * nullspace=new KRYLOV_VECTOR_WRAPPER<T,ARRAY<T> >;
                nullspace->v.Resize(example.dof);
                nullspace->v.Subset(component_dofs).Fill(1/sqrt((T)component_dofs.m));
                projection_system.nullspace_vectors.Append(nullspace);
            }
            region_has_dc=false;
            component_dofs.Remove_All();
        });

    if(example.use_preconditioner)
    {
        poisson.Construct_Incomplete_Cholesky_Factorization();
        poisson_tmp.v.Resize(example.dof);
        projection_system.Set_Preconditioner(*poisson.C,poisson_tmp);
    }
}
//#####################################################################
// Function Prepare_Pressure_And_Viscosity
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Prepare_Pressure_And_Viscosity()
{
    TIMER_SCOPE_FUNC;
    Compute_Boundary_Conditions();
    Allocate_Pressure_Dofs();
    sol.v.Resize(example.dof,init_all,0);
    rhs.v.Resize(example.dof,init_all,0);
    Compute_Poisson_Matrix();
}
//#####################################################################
// Function Pressure_Projection
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Pressure_Projection()
{
    TIMER_SCOPE_FUNC;
    if(example.use_bdf2&&(example.init_prev||!example.is_first_timestep))
#pragma omp parallel for
        for(REGION_FACE_FLAT_ID i=REGION_FACE_FLAT_ID();i<example.mass.raw.m;i++)
        {
            example.velocity(i)*=1.5;
        }
    
    Apply_Divergence(rhs.v,example.velocity,-1);

    static int solve_id=-1;
    solve_id++;
    LOG::cout<<"solve id: "<<solve_id<<std::endl;

    if(example.test_system) Test_Poisson();
    if(example.print_matrix) Print_Poisson_Matrix(solve_id);

    ARRAY<KRYLOV_VECTOR_BASE<T>*> av;
    CONJUGATE_GRADIENT<T> cg;
    cg.relative_tolerance=true;
    bool converged=cg.Solve(projection_system,sol,rhs,
        av,example.solver_tolerance,0,example.solver_iterations);
    if(!converged) LOG::printf("SOLVER DID NOT CONVERGE.\n");

    if(example.print_matrix)
        OCTAVE_OUTPUT<T>(LOG::sprintf("x-%i.txt",solve_id).c_str()).Write("x",sol);

    if(example.test_system){
        projection_system.Multiply(sol,*av(0));
        *av(0)-=rhs;
        T r=projection_system.Convergence_Norm(*av(0));
        LOG::cout<<"residual: "<<r<<std::endl;}
    av.Delete_Pointers_And_Clean_Memory();

    Apply_Gradient(example.velocity,sol.v,-1);
    
    if(example.use_bdf2&&(example.init_prev||!example.is_first_timestep))
#pragma omp parallel for
        for(REGION_FACE_FLAT_ID i=REGION_FACE_FLAT_ID();i<example.mass.raw.m;i++)
        {
            example.velocity(i)/=1.5;
        }
        
    if(example.predictive_pressure) example.pressure+=sol.v/example.dt; // sol.v is dt*dp
    else if(example.check_pressure_error) example.pressure=sol.v/example.dt;
}
//#####################################################################
// Function Apply_Particle_Forces
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Apply_Forces()
{
    TIMER_SCOPE_FUNC;
    example.gather_scatter->Scatter(false,false,0,0,
        [this](int p,PARTICLE_GRID_FACE_ITERATOR<TV>& it,int data)
        {
            for(;it.Valid();it.Next())
            {
                T w=it.Weight();
                REGION_FACE_INDEX<TV::m> ri={example.particles.Region(p),it.Index()};
                example.force(ri)+=w*example.gravity(ri.f.axis)*example.particles.mass(p);
            }
        });
    if(!example.use_mls_xfers)
        Reflect_Boundary_Particle_Force(example.force);
#pragma omp parallel for
    for(int i=0;i<example.valid_flat_indices.m;i++){
        auto k=example.valid_flat_indices(i);
        auto f=example.valid_indices(i);
        TV af=example.Compute_Analytic_Force(example.grid.Face_X(f.f),example.time+example.dt*example.f_timepoint);
        example.force(k)+=af(f.f.axis)*example.mass(k);}
    if(!example.use_mls_xfers)
        Reflect_Boundary_Grid_Force(example.force);
#pragma omp parallel for
    for(REGION_FACE_FLAT_ID i=REGION_FACE_FLAT_ID();i<example.velocity.raw.m;i++){
        T acc=0;
        if(example.mass(i)) acc=example.force(i)/example.mass(i);
        if(example.use_bdf2&&(example.init_prev||!example.is_first_timestep)) acc/=1.5; // Is this using f^{n+1}? if f is not analytic
        example.velocity(i)+=acc*example.dt;
        example.force(i)=0;}
    for(auto&u:example.velocity.u)
        Fix_Periodic(u);
}
//#####################################################################
// Function Compute_Dt
//#####################################################################
template<class TV> typename TV::SCALAR MPM_MAC_DRIVER<TV>::
Compute_Dt() const
{
    TIMER_SCOPE_FUNC;
    T critical_speed=example.cfl*example.grid.dX.Min()/example.max_dt;
    T v=Grid_V_Upper_Bound();
    return (v>critical_speed)?(example.cfl*example.grid.dX.Min()/v):example.max_dt;
}
//#####################################################################
// Function Max_Particle_Speed
//#####################################################################
template<class TV> typename TV::SCALAR MPM_MAC_DRIVER<TV>::
Max_Particle_Speed() const
{
    T v2=0;
#pragma omp parallel for reduction(max:v2)
    for(int k=0;k<example.simulated_particles.m;k++){
        int p=example.simulated_particles(k);
        v2=max(v2,example.particles.V(p).Magnitude_Squared());}
    return sqrt(v2);
}
//#####################################################################
// Function Grid_V_Upper_Bound
//#####################################################################
template<class TV> typename TV::SCALAR MPM_MAC_DRIVER<TV>::
Grid_V_Upper_Bound() const
{
    T result=0;
    T xi=(T)6*sqrt((T)TV::m)*example.grid.one_over_dX.Min()*example.weights(0)->Dp;
#pragma omp parallel for reduction(max:result)
    for(int k=0;k<example.simulated_particles.m;k++){
        int p=example.simulated_particles(k);
        T v=example.particles.V(p).Magnitude();
        if(example.particles.store_C) v+=example.particles.C(p).Frobenius_Norm()*xi;
        if(example.opt.use_polypic){ // Not a tight maximum; (C is not tight either?)
            SYMMETRIC_MATRIX<T,TV::m> sumH;
            for(int i=0;i<TV::m;++i) sumH+=example.particles.H(p).x(i);
            v+=sqrt((sumH.Times_Transpose(sumH)).Trace())*xi;
        }
        result=max(result,v);}
    return result;
}
//#####################################################################
// Function Update_Simulated_Particles
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Update_Simulated_Particles()
{
    TIMER_SCOPE_FUNC;

    example.simulated_particles.Remove_All();
    for(int p=0;p<example.particles.number;p++)
        if(example.particles.valid(p))
            example.simulated_particles.Append(p);
}
//#####################################################################
// Function Prepare_Scatter
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Prepare_Scatter()
{
    TIMER_SCOPE_FUNC;
    example.gather_scatter->Prepare_Scatter(example.particles);
}
//#####################################################################
// Function Fix_Periodic
//#####################################################################
template<class TV> template<class T2> void MPM_MAC_DRIVER<TV>::
Fix_Periodic(ARRAY<T2,IV>& u) const
{
    if(!example.side_bc_type.Contains(MPM::BC_PERIODIC)) return;
    Fill_Ghost_Cells_Periodic(example.grid,u,example.periodic_boundary.is_periodic);
}
//#####################################################################
// Function Fix_Periodic
//#####################################################################
template<class TV> template<class T2> void MPM_MAC_DRIVER<TV>::
Fix_Periodic(ARRAY<T2,FACE_INDEX<TV::m> >& u) const
{
    if(!example.side_bc_type.Contains(MPM::BC_PERIODIC)) return;
    Fill_Ghost_Faces_Periodic(example.grid,u,example.periodic_boundary.is_periodic);
}
//#####################################################################
// Function Fix_Periodic_Accum
//#####################################################################
template<class TV> template<class T2,class OP> void MPM_MAC_DRIVER<TV>::
Fix_Periodic_Accum(ARRAY<T2,IV>& u,const OP& op) const
{
    if(!example.side_bc_type.Contains(MPM::BC_PERIODIC)) return;
    Fill_Ghost_Cells_Periodic_Accum(example.grid,u,
        example.periodic_boundary.is_periodic,op);
}
//#####################################################################
// Function Fix_Periodic_Accum
//#####################################################################
template<class TV> template<class T2,class OP> void MPM_MAC_DRIVER<TV>::
Fix_Periodic_Accum(ARRAY<T2,FACE_INDEX<TV::m> >& u,const OP& op) const
{
    if(!example.side_bc_type.Contains(MPM::BC_PERIODIC)) return;
    Fill_Ghost_Faces_Periodic_Accum(example.grid,u,
        example.periodic_boundary.is_periodic,op);
}
//#####################################################################
// Function Reflect_Boundary_Particle_Force
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Reflect_Boundary_Particle_Force(REGION_ARRAY<T,FACE_INDEX<TV::m> >& force) const
{
    Reflect_Boundary(
        [&](const auto& in,const auto& out,int side)
        {force(in)-=force(out);},
        [&](const auto& in,const auto& out,int side)
        {force(in)+=force(out);},
        RF::duplicate_corners);
}
//#####################################################################
// Function Reflect_Boundary_Grid_Force
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Reflect_Boundary_Grid_Force(REGION_ARRAY<T,FACE_INDEX<TV::m> >& force) const
{
    Reflect_Boundary(
        [&](const auto& in,const auto& out,int side)
        {force(out)=-force(in);},
        [&](const auto& in,const auto& out,int side)
        {force(out)=force(in);},
        RF::delay_corners);
}
//#####################################################################
// Function Reflect_Boundary_Mass_Momentum
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Reflect_Boundary_Mass_Momentum(REGION_ARRAY<T,FACE_INDEX<TV::m> >& P) const
{
    Reflect_Boundary(
        [&](const auto& in,const auto& out,int side)
        {
            T bc=0;
            if(example.bc_velocity){
                TV nearest_point=example.grid.domain.Clamp(example.grid.Face_X(out.f));
                bc=example.bc_velocity(nearest_point,example.time)(out.f.axis);}
            T& moment_in=P(in),&moment_out=P(out);
            T mv_in_old=moment_in,mv_out_old=moment_out;
            T& m_in=example.mass(in),&m_out=example.mass(out);
            moment_in+=2*bc*m_out-moment_out;
            moment_out=mv_out_old+2*bc*m_in-mv_in_old;
            m_in+=m_out;
            m_out=m_in;
        },
        [&](const auto& in,const auto& out,int side)
        {
            T& moment_in=P(in),&moment_out=P(out);
            moment_in+=moment_out;
            moment_out=moment_in;
            T& m_in=example.mass(in),&m_out=example.mass(out);
            m_in+=m_out;
            m_out=m_in;
        },
        RF::duplicate_corners);
}
//#####################################################################
// Function Reflect_Boundary_Copy_Only
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Reflect_Boundary_Velocity_Copy_Only() const
{
    TIMER_SCOPE_FUNC;
    Reflect_Boundary(
        [&](const auto& in,const auto& out,int side)
        {
            T bc=0;
            if(example.bc_velocity){
                TV nearest_point=example.grid.domain.Clamp(example.grid.Face_X(out.f));
                bc=example.bc_velocity(nearest_point,example.time)(out.f.axis);}
            example.velocity(out)=(2*bc-example.velocity(in));
        },
        [&](const auto& in,const auto& out,int side)
        {
            example.velocity(out)=example.velocity(in);
        },
        RF::delay_corners);
    for(auto& a:example.velocity.u) Fix_Periodic(a);
}
//#####################################################################
// Function Reflect_Boundary
//#####################################################################
template<class TV> template<class D,class N> void MPM_MAC_DRIVER<TV>::
Reflect_Boundary(D func_d,N func_n,RF flag) const
{
    for(REGION_ID r(0);r<example.mass.u.m;r++)
    {
        RANGE<IV> domain=example.grid.Cell_Indices();
        IV corner[2]={domain.min_corner,domain.max_corner};
        const RANGE<IV>& outer=example.mass.u(r).domain;
        for(FACE_RANGE_ITERATOR<TV::m> it(outer,domain,flag|RF::ghost);it.Valid();it.Next())
        {
            REGION_FACE_INDEX<TV::m> ri={r,it.face},si=ri;
            auto bc_type=example.side_bc_type(it.side);
            if(bc_type==MPM::BC_PERIODIC) continue;
            int side_axis=it.side/2;
            si.f.index(side_axis)=2*corner[it.side%2](side_axis)-it.face.index(side_axis);
            if(side_axis!=it.face.axis) si.f.index(side_axis)-=1;
            if(example.bc_type(it.side)){
                TV Y=example.grid.domain.Clamp(example.grid.Face_X(it.face));
                bc_type=example.bc_type(it.side)(Y,example.time);}
            if((bc_type==MPM::BC_SLIP && it.side/2==it.face.axis) || bc_type==MPM::BC_NOSLIP)
                func_d(si,ri,it.side);
            else func_n(si,ri,it.side);
        }
    }
}
//#####################################################################
// Function Apply_Viscosity
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Apply_Viscosity()
{
    REGION_ID rid(0);
    PHYSBAM_ASSERT(example.velocity.u.m==REGION_ID(1));
    TIMER_SCOPE_FUNC;
    for(const auto& t:example.bc_type) PHYSBAM_ASSERT(!t);
    typedef KRYLOV_VECTOR_WRAPPER<T,ARRAY<T> > VEC;
    typedef MATRIX_SYSTEM<SPARSE_MATRIX_FLAT_MXN<T>,T,VEC> MAT;
    if(!example.viscosity.m) return;

    // TODO: cut cell does its boundary conditions differently that this function does.
    PHYSBAM_ASSERT(!example.use_level_set_projection);
    
    auto p_psi_D=[this](const IV& c){return example.cell_index(c)==pressure_D;};
    // Based on logic from IMPLICIT_VISCOSITY_UNIFORM::Setup_Boundary_Conditions
    for(int axis=0;axis<TV::m;axis++){
        GRID<TV> face_grid(example.grid.Get_Face_MAC_Grid(axis));
        ARRAY<int,IV> velocity_index(face_grid.Cell_Indices(1),init_new,-1);
        VEC rhs,sol,tmp_vector;

        auto range=example.grid.Cell_Indices().Intersect(example.velocity.u(rid).domain);
        
        // Allocate velocities that will be corrected; copy over initial guess
        for(FACE_RANGE_ITERATOR<TV::m> it(range,RF::none,axis);it.Valid();it.Next())
        {
            if(example.side_bc_type(axis)==MPM::BC_PERIODIC)
                if(it.face.index(axis)==example.grid.numbers_of_cells(axis))
                    continue;
            if(p_psi_D(it.face.First_Cell_Index()) && p_psi_D(it.face.Second_Cell_Index()))
                continue;
            if(!example.psi_N(it.face))
                velocity_index(it.face.index)=rhs.v.Append(example.velocity.u(rid)(it.face));
        }
        Fix_Periodic(velocity_index);

        // Compute as needed rather than store.  face_grid index.
        auto psi_N=[axis,this,p_psi_D](const FACE_INDEX<TV::m>& face,int s)
        {
            IV a=face.index,b=a;
            b(axis)--;
            if(face.axis==axis) return p_psi_D(b);
            int index=face.index(face.axis);
            bool boundary=index==example.grid.Cell_Indices().min_corner(face.axis)||
                index==example.grid.Cell_Indices().max_corner(face.axis);
            if(boundary && example.side_bc_type(2*face.axis+s)==MPM::BC_SLIP) return true;
            if(p_psi_D(a) && p_psi_D(b)) return true;
            a(face.axis)--;
            b(face.axis)--;
            return p_psi_D(a) && p_psi_D(b);
        };
        

        SPARSE_MATRIX_FLAT_MXN<T> A;
        MAT sys(A);
        TV scale=example.viscosity(rid)/example.density(rid)*example.dt*sqr(example.grid.one_over_dX);
        if(example.use_bdf2&&(example.init_prev||!example.is_first_timestep)) scale /= 1.5;

        A.Reset(rhs.v.m);
        sol.v=rhs.v;
        tmp_vector.v.Resize(rhs.v.m);

        for(FACE_RANGE_ITERATOR<TV::m> it(range,RF::none,axis);it.Valid();it.Next())
        {
            int center_index=velocity_index(it.face.index);
            if(center_index<0) continue;
            if(example.side_bc_type(axis)==MPM::BC_PERIODIC)
                if(it.face.index(axis)==example.grid.numbers_of_cells(axis))
                    continue;
            T diag=1;
            for(int a=0;a<TV::m;a++){
                FACE_INDEX<TV::m> face(a,it.face.index);
                for(int s=0;s<2;s++){
                    IV cell=it.face.index;
                    cell(a)+=2*s-1;
                    PHYSBAM_ASSERT(face.Cell_Index(s)==cell);
                    if(psi_N(face,s)){
                        // Assume for now that the boundary condition is zero.
                    }
                    else{
                        diag+=scale(a);
                        int ci=velocity_index(cell);
                        if(ci>=0) A.Append_Entry_To_Current_Row(ci,-scale(a));
                        else rhs.v(center_index)+=scale(a)*example.velocity.u(rid)(FACE_INDEX<TV::m>(axis,cell));}
                    face.index(a)++;}}
            A.Append_Entry_To_Current_Row(center_index,diag);
            A.Finish_Row();
        }
        A.Sort_Entries();

        if(example.use_preconditioner){
            A.Construct_Incomplete_Cholesky_Factorization();
            sys.Set_Preconditioner(*A.C,tmp_vector);}

        ARRAY<KRYLOV_VECTOR_BASE<T>*> av;
        static int solve_id=-1;
        solve_id++;
        if(example.test_system) sys.Test_System(sol);
        if(example.print_matrix){
            LOG::cout<<"solve id: "<<solve_id<<std::endl;
            OCTAVE_OUTPUT<T>(LOG::sprintf("visc-M-%i.txt",solve_id).c_str()).Write("M",sys,rhs);
            OCTAVE_OUTPUT<T>(LOG::sprintf("visc-C-%i.txt",solve_id).c_str()).Write_Preconditioner("C",sys,rhs);
            OCTAVE_OUTPUT<T>(LOG::sprintf("visc-b-%i.txt",solve_id).c_str()).Write("b",rhs);}
        CONJUGATE_GRADIENT<T> cg;
        cg.finish_before_indefiniteness=true;
        cg.relative_tolerance=true;
        bool converged=cg.Solve(sys,sol,rhs,
            av,example.solver_tolerance,0,example.solver_iterations);
        if(!converged) LOG::printf("SOLVER DID NOT CONVERGE.\n");

        if(example.print_matrix)
            OCTAVE_OUTPUT<T>(LOG::sprintf("visc-x-%i.txt",solve_id).c_str()).Write("x",sol);

        if(example.test_system){
            sys.Multiply(sol,*av(0));
            *av(0)-=rhs;
            T r=sys.Convergence_Norm(*av(0));
            LOG::cout<<"visc residual: "<<r<<std::endl;}
        av.Delete_Pointers_And_Clean_Memory();
        
        for(CELL_ITERATOR<TV> it(face_grid);it.Valid();it.Next()){
            FACE_INDEX<TV::m> face(axis,it.index);
            int i=velocity_index(it.index);
            if(i>=0) example.velocity.u(rid)(face)=sol.v(i);}}

    Fix_Periodic(example.velocity.u(rid));
}
//#####################################################################
// Function Apply_Viscosity
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Level_Set_Apply_Viscosity()
{
    REGION_ID rid(0);
    PHYSBAM_ASSERT(example.velocity.u.m==REGION_ID(1));
    TIMER_SCOPE_FUNC;
    for(const auto& t:example.bc_type) PHYSBAM_ASSERT(!t);
    typedef KRYLOV_VECTOR_WRAPPER<T,ARRAY<T> > VEC;
    typedef MATRIX_SYSTEM<SPARSE_MATRIX_FLAT_MXN<T>,T,VEC> MAT;
    if(!example.viscosity.m) return;

    // TODO: cut cell does its boundary conditions differently that this function does.
    PHYSBAM_ASSERT(example.use_level_set_projection);
    PHYSBAM_ASSERT(example.disable_free_surface);
    PHYSBAM_ASSERT(example.velocity.u.m==REGION_ID(1));

    for(const auto& t:example.bc_type) PHYSBAM_ASSERT(!t);
    ARRAY<T,IV> object_phi(example.grid.Node_Indices(example.ghost));
    T def_dep=3*example.grid.dX.Max();
    for(RANGE_ITERATOR<TV::m> it(object_phi.domain);it.Valid();it.Next()){
        TV X=example.grid.Node_X(it.index);
        T dep=def_dep;//example.grid.domain.Signed_Distance(X);
        for(int i=0;i<example.collision_objects.m;i++)
            dep=min(dep,example.collision_objects(i)->Phi(X,example.time));
        object_phi(it.index)=-dep;}

    ARRAY<bool,IV> p_psi_D(example.cell_index.domain);
    for(RANGE_ITERATOR<TV::m> it(example.cell_index.domain);it.Valid();it.Next())
        p_psi_D(it.index)=example.cell_index(it.index)==pressure_D;

    for(int axis=0;axis<TV::m;axis++)
    {
        ARRAY<T,IV> tmp_u(example.grid.Cell_Indices(example.ghost));
        tmp_u.Put(example.velocity.u(rid).Component(axis),tmp_u);
        HEAT_LAPLACE<LAPLACE_COLLIDABLE_UNIFORM<TV> > heat(
            example.grid.Get_Face_Grid(axis),tmp_u);
        LEVELSET<TV> levelset(example.grid,at_cells,object_phi);
        heat.Use_External_Level_Set(levelset);
        heat.Set_Up_Second_Order_Cut_Cell_Method(true);
        heat.coefficient=example.dt*example.viscosity(rid)/example.density(rid);
        heat.psi_D=example.psi_N.Component(axis);

        for(FACE_RANGE_ITERATOR<TV::m> it(example.velocity.u(rid).domain,RF::none,axis);it.Valid();it.Next())
            if(p_psi_D(it.face.First_Cell_Index()) && p_psi_D(it.face.Second_Cell_Index()))
                heat.psi_D(it.face.index)=true;

        for(FACE_RANGE_ITERATOR<TV::m> it(heat.psi_N.domain);it.Valid();it.Next())
        {
            IV a=it.face.index,b=a.Add_Axis(axis,-1);
            if(it.face.axis==axis)
            {
                heat.psi_N(it.face)=p_psi_D(b);
                continue;
            }
            int index=it.face.index(it.face.axis);
            if(index==example.grid.Cell_Indices().min_corner(it.face.axis) && example.side_bc_type(2*it.face.axis)==MPM::BC_SLIP)
            {
                heat.psi_N(it.face)=true;
                continue;
            }
            if(index==example.grid.Cell_Indices().max_corner(it.face.axis) && example.side_bc_type(2*it.face.axis+1)==MPM::BC_SLIP)
            {
                heat.psi_N(it.face)=true;
                continue;
            }
            if(p_psi_D(a) && p_psi_D(b))
            {
                heat.psi_N(it.face)=true;
                continue;
            }
            a(it.face.axis)--;
            b(it.face.axis)--;
            heat.psi_N(it.face)=p_psi_D(a) && p_psi_D(b);
        }

        heat.f=tmp_u;
        heat.f.array*=-1;
        heat.Solve(example.time);
        tmp_u.Put(tmp_u,example.velocity.u(rid).Component(axis));
    }

    Fix_Periodic(example.velocity.u(rid));
}
//#####################################################################
// Function Step
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Step(std::function<void()> func,const char* name,bool dump_substep,bool do_step)
{
    auto& p=example.time_step_callbacks.Get_Or_Insert(name);
    p.first=true; // Flag the callback name as recognized, for sanity checking later
    if(!do_step) return;
    for(int i=0;i<p.second(1).m;i++) p.second(1)(i)();
    func();
    // Note: p may be invalidated by func(), so we must re-access it here.
    const auto& q=example.time_step_callbacks.Get_Or_Insert(name);
    for(int i=0;i<q.second(0).m;i++) q.second(0)(i)();
    if(dump_substep) PHYSBAM_DEBUG_WRITE_SUBSTEP(name,2);
}
//#####################################################################
// Function Invalidate_Particle
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Invalidate_Particle(int p)
{
    example.particles.valid(p)=false;
    example.particles.Add_To_Deletion_List(p);
}
//#####################################################################
// Function Move_Particles
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Move_Particles()
{
    TIMER_SCOPE_FUNC;
    ARRAY<TV> v_temp;
    ARRAY<MATRIX<T,TV::m>> C_temp;
    
    if(example.adv_v_timepoint) // Compute C and v used for advection
    {
        v_temp=example.particles.V;
        auto& V_prev=*example.particles.template Get_Array<TV>("V_prev");
        if(!example.is_first_timestep) 
#pragma omp parallel for
            for(int p=0;p<example.particles.X.m;p++)
                if(example.particles.valid(p))
                    example.particles.V(p)+=(example.particles.V(p)-V_prev(p))*example.adv_v_timepoint;

        if(example.opt.use_affine && example.C_JH){
            C_temp=example.particles.C;
            auto& C_prev=*example.particles.template Get_Array<MATRIX<T,TV::m>>("C_prev");
#pragma omp parallel for
            for(int p=0;p<example.particles.X.m;p++)
                if(example.particles.valid(p))
                    if(!example.is_first_timestep) example.particles.C(p)+=(example.particles.C(p)-C_prev(p))*example.adv_v_timepoint;
        }
    }


    TV wall[2]={example.grid.domain.min_corner,example.grid.domain.max_corner};
    ARRAY<ARRAY<int> > invalidate_lists(example.threads);
    PARTICLE_BOUNDARY_CONDITIONS<TV> bc_helper(example.particles,example.grid,example.side_bc_type);
    if(example.bc_velocity)
        bc_helper.bc_v=[this](const TV& X){return example.bc_velocity(X,example.time);};
    for(int i=0;i<2*TV::m;i++)
        if(example.bc_type(i))
            bc_helper.bc_mixed_type(i)=
                [this,i](const TV& X){return example.bc_type(i)(X,example.time);};
    
#pragma omp parallel
    {
        ARRAY<int> invalidate_list;

        auto Clip=[this,wall,&invalidate_list,&bc_helper](int p)
            {
                bc_helper.Enforce(p);
                if(!example.particles.valid(p))
                    invalidate_list.Append(p);

                TV X=example.particles.X(p);
                if(example.push_particles_from_objects)
                    for(int i=0;i<example.collision_objects.m;i++){
                        MPM_COLLISION_OBJECT<TV>* o=example.collision_objects(i);
                        if(o->Phi(X,example.time)<0){
                            example.particles.X(p)=o->Get_Implicit_Object(example.time)->Closest_Point_On_Boundary(X);
                            break;}}
            };

        if(example.rk_particle_order && example.position_update=='d'){
            LINEAR_INTERPOLATION_FACE<TV,T> li;
#pragma omp for
            for(int p=0;p<example.particles.X.m;p++)
                if(example.particles.valid(p)){
                    for(RUNGEKUTTA<TV> rk(example.particles.X(p),example.rk_particle_order,example.dt,0);rk.Valid();rk.Next())
                        example.particles.X(p)+=example.dt*li.Interpolate_Face(example.grid,example.velocity.u(example.particles.Region(p)),example.particles.X(p));
                    Clip(p);}}
        else if(example.opt.flip && example.position_update=='d'){
#pragma omp for
            for(int p=0;p<example.particles.X.m;p++)
                if(example.particles.valid(p)){
                    example.particles.X(p)+=example.dt*example.flip_adv_velocity(p);
                    Clip(p);}}
        else if(example.interpolated_advection){
            auto& advection_v=*example.particles.template Get_Array<TV>("advection_velocity");
#pragma omp for
            for(int p=0;p<example.particles.X.m;p++)
                if(example.particles.valid(p)){
                    example.particles.X(p)+=example.dt*advection_v(p);
                    Clip(p);}}
        else if(example.opt.flip && example.position_update=='x'){
            example.gather_scatter->Gather(false,false,0,0,
                [this,&Clip](int p,PARTICLE_GRID_FACE_ITERATOR<TV>& it,int data)
                {
                    for(;it.Valid();it.Next())
                    {
                        REGION_FACE_INDEX<TV::m> ri={example.particles.Region(p),it.Index()};
                        example.particles.X(p)(ri.f.axis)+=example.dt*it.Weight()*
                            (example.velocity(ri)+example.velocity_save(ri))/2;
                    }
                    Clip(p);
                });}
        else if(example.xpic && example.position_update=='d'){
#pragma omp for
            for(int p=0;p<example.particles.X.m;p++)
                if(example.particles.valid(p)){
                    example.particles.X(p)+=example.dt*example.effective_v(p)*0.5;}
            example.gather_scatter->Gather(false,false,0,0,
                [this,&Clip](int p,PARTICLE_GRID_FACE_ITERATOR<TV>& it,int data)
                {
                    for(;it.Valid();it.Next())
                    {
                        REGION_FACE_INDEX<TV::m> ri={example.particles.Region(p),it.Index()};
                        example.particles.X(p)(ri.f.axis)+=example.dt*it.Weight()*example.velocity_save(ri);
                    }
                    Clip(p);
                });}
        else{
#pragma omp for
            for(int p=0;p<example.particles.X.m;p++)
                if(example.particles.valid(p)){
                    example.particles.X(p)+=example.dt*example.particles.V(p);
                    Clip(p);}}
        if(example.position_update=='p')
            example.position_update='d';

#ifdef USE_OPENMP
        invalidate_lists(omp_get_thread_num()).Exchange(invalidate_list);
#else
        invalidate_lists(0).Exchange(invalidate_list);
#endif
    }
    
    ARRAY<MATRIX<T,TV::m>> C_adv;
    // Compute C advection factor C_adv
    if(example.opt.use_affine){
        C_adv.Resize(example.particles.X.m);
#pragma omp parallel for
        for(int p=0;p<example.particles.X.m;p++)
            if(example.particles.valid(p))
            {
                auto& C=example.particles.C(p);
                C_adv(p) = C.Identity_Matrix();
                if(example.C_order==2){
                    C_adv(p)=(1-example.dt*C);
                }
                else if(example.C_order>2){
                    C_adv(p)=(1+example.dt*C).Inverse();
                }
            }
    }
    
    if(example.adv_v_timepoint) // Restore C and v used for advection
    {
#pragma omp parallel for
        for(int p=0;p<example.particles.X.m;p++)
            if(example.particles.valid(p))
                example.particles.V(p)=v_temp(p);
        if(example.opt.use_affine && example.C_JH){
#pragma omp parallel for
            for(int p=0;p<example.particles.X.m;p++)
                if(example.particles.valid(p))
                    example.particles.C(p)=C_temp(p);
        }
    }

    if(example.particles.store_C) // advect C and C_prev
    {
        auto& C_prev=*example.particles.template Get_Array<MATRIX<T,TV::m>>("C_prev");
#pragma omp parallel for
        for(int p=0;p<example.particles.X.m;p++)
            if(example.particles.valid(p))
            {
                auto& C=example.particles.C(p);
                C=C*C_adv(p);
                if(!example.is_first_timestep && (example.adv_v_timepoint || example.use_bdf2) && example.C_JH) C_prev(p)=C_prev(p)*C_adv(p);
            }
    }
    for(int t=0;t<invalidate_lists.m;t++)
        for(int i=0;i<invalidate_lists(t).m;i++)
            Invalidate_Particle(invalidate_lists(t)(i));
}
//#####################################################################
// Function Extrapolate_Inside_Object
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Extrapolate_Inside_Object()
{
    TIMER_SCOPE_FUNC;
    REGION_ID rid(0);
    PHYSBAM_ASSERT(example.velocity.u.m==REGION_ID(1));
    ARRAY<T,IV> phi_data(example.grid.Cell_Indices(3));
    LEVELSET<TV> phi(example.grid,at_cells,phi_data,3);
    RANGE<TV> box=example.grid.domain;
    ARRAY<bool,FACE_INDEX<TV::m> > face_inside(example.grid.Cell_Indices(3));
    T def_dep=-3*example.grid.dX.Max();
    for(RANGE_ITERATOR<TV::m> it(example.grid.Cell_Indices(),3,0);it.Valid();it.Next()){
        TV X=example.grid.Cell_X(it.index);
        T dep=def_dep;//example.grid.domain.Signed_Distance(X);
        for(int i=0;i<example.collision_objects.m;i++)
            dep=max(dep,-example.collision_objects(i)->Phi(X,example.time));
        phi_data(it.index)=dep;}
    for(FACE_RANGE_ITERATOR<TV::m> it(example.grid.Cell_Indices(3));it.Valid();it.Next()){
        TV X=example.grid.Face_X(it.face);
        T dep=def_dep;//example.grid.domain.Signed_Distance(X);
        for(int i=0;i<example.collision_objects.m;i++)
            dep=max(dep,-example.collision_objects(i)->Phi(X,example.time));
        face_inside(it.face)=dep<0;}

    EXTRAPOLATION_HIGHER_ORDER_FACE<TV> eho(phi,example.grid,
        [&face_inside](const FACE_INDEX<TV::m>& index){return face_inside(index);},
        example.velocity.u(rid).domain,3,3);
    eho.Extrapolate(example.velocity.u(rid),20);
    if(example.opt.flip)
        eho.Extrapolate(example.velocity_save.u(rid),20);
}
//#####################################################################
// Function Level_Set_Pressure_Projection
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Level_Set_Pressure_Projection()
{
    REGION_ID rid(0);
    PHYSBAM_ASSERT(example.velocity.u.m==REGION_ID(1));
    TIMER_SCOPE_FUNC;
    for(const auto& t:example.bc_type) PHYSBAM_ASSERT(!t);
    ARRAY<T,IV> object_phi(example.grid.Node_Indices(example.ghost));
    T def_dep=3*example.grid.dX.Max();
    for(RANGE_ITERATOR<TV::m> it(object_phi.domain);it.Valid();it.Next()){
        TV X=example.grid.Node_X(it.index);
        T dep=def_dep;//example.grid.domain.Signed_Distance(X);
        for(int i=0;i<example.collision_objects.m;i++)
            dep=min(dep,example.collision_objects(i)->Phi(X,example.time));
        object_phi(it.index)=-dep;}

    CUT_CELL_PROJECTION<TV> proj;
    proj.object_phi=&object_phi;
    if(!example.disable_free_surface) proj.surface_phi=&example.phi;
    example.valid_xfer_data.Resize(example.mass);
    proj.valid_u=&example.valid_xfer_data.u(rid);
    if(example.output_face_fractions) proj.face_fractions=&face_fractions;
    
    for(int s=0;s<2*TV::m;s++)
        switch(example.side_bc_type(s))
        {
            case MPM::BC_FREE:proj.bc_type(s)=1;break;
            case MPM::BC_SLIP:case MPM::BC_NOSLIP:proj.bc_type(s)=2;break;
            case MPM::BC_PERIODIC:proj.bc_type(s)=0;continue;
            case MPM::BC_SEP:PHYSBAM_FATAL_ERROR();
            case MPM::BC_MIXED:PHYSBAM_FATAL_ERROR();
        }
    proj.use_preconditioner=example.use_preconditioner;
    proj.print_matrix=example.print_matrix;
    proj.print_residual=example.test_system;
    proj.solver_tolerance=example.solver_tolerance;
    proj.solver_iterations=example.solver_iterations;
    if(example.bc_velocity)
        proj.bc_v=[this](const TV& X){return example.bc_velocity(X,example.time);};
    if(example.bc_pressure)
        proj.bc_p=[this](const TV& X){return example.bc_pressure(X,example.time);};

    proj.use_warm_start=example.use_warm_start;
    if(example.use_warm_start){
        proj.pressure=&example.pressure_save;
        proj.valid_p=&example.pressure_valid;}
    
    proj.Cut_Cell_Projection(example.grid,example.velocity.u(rid),example.density(rid),example.dt);

    for(FACE_RANGE_ITERATOR<TV::m> it(example.grid.Cell_Indices(example.ghost));it.Valid();it.Next())
        if(!(*proj.valid_u)(it.face))
            example.velocity.u(rid)(it.face)=0;

    if(example.use_mls_xfers && example.zero_invalid){
        for(FACE_ITERATOR<TV> it(example.grid);it.Valid();it.Next()){
            for(int i=0;i<example.collision_objects.m;i++){
                if(example.collision_objects(i)->Phi(it.Location(),example.time)<0 && !example.valid_xfer_data.u(rid)(it.face)){
                    example.valid_xfer_data.u(rid)(it.face)=true;
                    example.velocity.u(rid)(it.face)=0;}}}}
}
//#####################################################################
// Function Compute_Regions
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Compute_Regions()
{
    TIMER_SCOPE_FUNC;
    PHYSBAM_ASSERT(example.density.m);
    PHYSBAM_ASSERT(example.density.m==example.viscosity.m);
    VECTOR<ARRAY<RANGE<IV> >,TV::m> particle_ranges;
    ARRAY<RANGE<IV>,REGION_ID> region_ranges(example.density.m);
    for(int i=0;i<TV::m;i++)
        example.weights(i)->Particle_Ranges(particle_ranges(i));
    for(int p=0;p<example.particles.X.m;p++)
        if(example.particles.valid(p))
        {
            REGION_ID i=example.particles.Region(p);
            for(int a=0;a<TV::m;a++)
            {
                auto range=particle_ranges(a)(p);
                range.max_corner(a)--;
                region_ranges(i).Enlarge_To_Include_Box(range);
            }
        }
    example.mass.Resize(region_ranges);
    example.velocity.Resize(example.mass);
    example.velocity_save.Resize(example.mass);
    example.force.Resize(example.mass);
    if(example.use_mls_xfers)
        example.valid_xfer_data.Resize(example.mass);
    LOG::printf("regions %P\n",example.mass.u.m);
}
//#####################################################################
// Function Update_Collision_Object_Lookup
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Update_Collision_Object_Lookup()
{
    T diag_dist=(T).5*example.grid.dX.Magnitude();
    collision_object_lookup.Remove_All();
    for(int i=0;i<example.collision_objects.m;i++)
    {
        RANGE<TV> box=example.collision_objects(i)->Bounding_Box(example.time);
        RANGE<IV> range(example.grid.Clamp_To_Cell(box,example.ghost));
        for(RANGE_ITERATOR<TV::m> it(range);it.Valid();it.Next())
        {
            T phi=example.collision_objects(i)->Phi(example.grid.Cell_X(it.index),example.time);
            if(phi<=diag_dist)
                collision_object_lookup.Insert(it.index,i);
        }
    }
}
//#####################################################################
// Function Visit_Collision_Objects
//#####################################################################
template<class TV> template<class F> void MPM_MAC_DRIVER<TV>::
Visit_Collision_Objects(const TV& X,const F& func)
{
    auto list=collision_object_lookup.Get(example.grid.Cell_IC(X,example.ghost));
    for(auto i:list)
    {
        T phi=example.collision_objects(i)->Phi(X,example.time);
        if(phi<=0)
            func(example.collision_objects(i),phi);
    }
}
//#####################################################################
// Function Visit_Colliding_Region_Faces
//#####################################################################
template<class TV> template<class F> void MPM_MAC_DRIVER<TV>::
Visit_Colliding_Region_Faces(const F& func)
{
    for(auto co:example.collision_objects)
    {
        RANGE<TV> box=co->Bounding_Box(example.time);
        // Handle infinite objects
        box=box.Intersect(example.grid.Ghost_Domain(example.ghost));
        RANGE<IV> range(example.grid.Clamp_To_Cell(box,example.ghost));
        for(REGION_ID r(0);r<example.mass.u.m;r++)
        {
            RANGE<IV> sub_range=range.Intersect(example.mass.u(r).domain);
            for(FACE_RANGE_ITERATOR<TV::m> it(sub_range);it.Valid();it.Next())
            {
                T phi=co->Phi(example.grid.Face_X(it.face),example.time);
                if(phi<=0)
                {
                    REGION_FACE_INDEX<TV::m> ri={r,it.face};
                    func(ri,co,phi);
                }
            }
        }
    }
}
//#####################################################################
// Function Apply_Gradient
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Apply_Gradient(REGION_ARRAY<T,FACE_INDEX<TV::m> >& u,const ARRAY<T>& p,
    T scale) const
{
    for(REGION_ID r(0);r<region_starts.m-1;r++)
    {
        T a=scale/example.density(r);
        for(int f=region_starts(r);f<region_starts(r+1);f++)
        {
            auto& e=gradient_data(f);
            for(int i=0;i<2;i++){
                u(e.face) += a*e.data(i)*p(e.dof(i));
            }
        }
    }
}
//#####################################################################
// Function Apply_Divergence
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Apply_Divergence(ARRAY<T>& div_u,const REGION_ARRAY<T,FACE_INDEX<TV::m> >& u,
    T scale) const
{
    for(auto& e:gradient_data)
    {
        T a=scale*u(e.face)*e.theta;
        for(int i=0;i<2;i++)
            div_u(e.dof(i)) -= e.data(i)*a;
    }
}
//#####################################################################
// Function Test_Poisson
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Test_Poisson()
{
    projection_system.Test_System(sol);

    RANDOM_NUMBERS<T> rand;
    ARRAY<T> u(example.dof),v(u.m),r(u.m);
    rand.Fill_Uniform(u,-1,1);
    REGION_ARRAY<T,FACE_INDEX<TV::m> > w(example.mass);
    Apply_Gradient(w,u,1);
    Apply_Divergence(v,w,-1);
    poisson.Times(r,u);
    LOG::printf("poisson = div grad: %P %P %P\n",v.Magnitude(),r.Magnitude(),(v-r).Magnitude());
}
//#####################################################################
// Function Print_Poisson_Matrix
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Print_Poisson_Matrix(int solve_id) const
{
    OCTAVE_OUTPUT<T>(LOG::sprintf("A-%i.txt",solve_id).c_str()).Write("A",poisson);
    OCTAVE_OUTPUT<T>(LOG::sprintf("B-%i.txt",solve_id).c_str()).Write_Preconditioner("B",projection_system,rhs);
    OCTAVE_OUTPUT<T>(LOG::sprintf("P-%i.txt",solve_id).c_str()).Write_Projection("P",projection_system,rhs);
    OCTAVE_OUTPUT<T>(LOG::sprintf("b-%i.txt",solve_id).c_str()).Write("b",rhs);
    OCTAVE_OUTPUT<T>(LOG::sprintf("Grad-%i.txt",solve_id).c_str()).Write("Grad",
        [this](ARRAY<T>& out,const ARRAY<T>& in)
        {
            REGION_ARRAY<T,FACE_INDEX<TV::m> > u(example.mass);
            u.raw.Fill(0);
            Apply_Gradient(u,in,1);
            for(REGION_FACE_FLAT_ID r(0);r<u.raw.m;r++)
                out(Value(r))=u.raw(r);
        },Value(example.mass.raw.m),example.dof);
    OCTAVE_OUTPUT<T>(LOG::sprintf("Div-%i.txt",solve_id).c_str()).Write("Div",
        [this](ARRAY<T>& out,const ARRAY<T>& in)
        {
            out.Fill(0);
            REGION_ARRAY<T,FACE_INDEX<TV::m> > u(example.mass);
            for(REGION_FACE_FLAT_ID r(0);r<u.raw.m;r++)
                u.raw(r)=in(Value(r));
            Apply_Divergence(out,u,1);
        },example.dof,Value(example.mass.raw.m));

    ARRAY<T> mass(Value(example.mass.raw.m)),theta(mass.m);
    OCTAVE_OUTPUT<T> og(LOG::sprintf("ST_G-%i.txt",solve_id).c_str());
    og.Begin_Sparse_Matrix("ST_G",Value(example.mass.raw.m),example.dof);
    for(REGION_ID r(0);r<region_starts.m-1;r++)
    {
        for(int f=region_starts(r);f<region_starts(r+1);f++)
        {
            auto& e=gradient_data(f);
            for(int i=0;i<2;i++)
            {
                og.Add_Sparse_Entry(Value(e.face),e.dof(i),e.data(i));
                mass.Append(example.density(r));
                theta.Append(e.theta);
            }
        }
    }
    og.End_Sparse_Matrix();
    OCTAVE_OUTPUT<T>(LOG::sprintf("M-%i.txt",solve_id).c_str()).Write_Diagonal("M",mass);
    OCTAVE_OUTPUT<T>(LOG::sprintf("T-%i.txt",solve_id).c_str()).Write_Diagonal("T",theta);
}
//#####################################################################
// Function Apply_Predictive_Pressure
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Apply_Predictive_Pressure()
{
    TIMER_SCOPE_FUNC;
    if(example.init_prev||!example.is_first_timestep){
        T scale=1;
        if(example.use_bdf2) scale=2./3;
        Apply_Gradient(example.velocity,example.pressure,-scale*example.dt);
    }
}
//#####################################################################
// Function Time_Integration
//#####################################################################
template<class TV> void MPM_MAC_DRIVER<TV>::
Time_Integration()
{
    TIMER_SCOPE_FUNC;
    if(example.use_bdf2)
    {
        auto& V_prev=*example.particles.template Get_Array<TV>("V_prev");
#pragma omp parallel for
        for(int p=0;p<example.particles.X.m;p++)
            if(example.particles.valid(p))
            {
                auto temp_v=example.particles.V(p);
                if(example.init_prev||!example.is_first_timestep) example.particles.V(p)+=(example.particles.V(p)-V_prev(p))/3;
                V_prev(p)=temp_v;
            }
        if(example.opt.use_affine){
            auto& C_prev=*example.particles.template Get_Array<MATRIX<T,TV::m>>("C_prev");
#pragma omp parallel for
            for(int p=0;p<example.particles.X.m;p++)
                if(example.particles.valid(p))
                {
                    auto temp_C=example.particles.C(p);
                    if((example.init_prev||!example.is_first_timestep)&& example.C_INT) example.particles.C(p)+=(example.particles.C(p)-C_prev(p))/3;
                    C_prev(p)=temp_C;
                }
        }
        if(example.opt.use_polypic){
            auto& H_prev=*example.particles.template Get_Array<SYMMETRIC_TENSOR<T,0,TV::m,TV::m>>("H_prev");
#pragma omp parallel for
            for(int p=0;p<example.particles.X.m;p++)
                if(example.particles.valid(p))
                {
                    auto temp_H=example.particles.H(p);
                    if((example.init_prev||!example.is_first_timestep)&& example.H_INT) example.particles.H(p)+=(example.particles.H(p)-H_prev(p))/3;
                    H_prev(p)=temp_H;
                }
        }
    }
    else if(example.adv_v_timepoint)
    {
        auto& V_prev=*example.particles.template Get_Array<TV>("V_prev");
#pragma omp parallel for
        for(int p=0;p<example.particles.X.m;p++)
            if(example.particles.valid(p))
            {
                V_prev(p)=example.particles.V(p);
            }
        if(example.opt.use_affine){
            auto& C_prev=*example.particles.template Get_Array<MATRIX<T,TV::m>>("C_prev");
#pragma omp parallel for
            for(int p=0;p<example.particles.X.m;p++)
                if(example.particles.valid(p))
                {
                    C_prev(p)=example.particles.C(p);
                }
        }
        if(example.opt.use_polypic){
            auto& H_prev=*example.particles.template Get_Array<SYMMETRIC_TENSOR<T,0,TV::m,TV::m>>("H_prev");
#pragma omp parallel for
            for(int p=0;p<example.particles.X.m;p++)
                if(example.particles.valid(p))
                {
                    H_prev(p)=example.particles.H(p);
                }
        }
    }
    if(example.ab_pressure){
        if(example.init_prev){
            auto temp_p=example.pressure;
            example.pressure=2*example.pressure-example.pressure_prev;
            example.pressure_prev=temp_p;
        }
        else if(!example.is_first_timestep){
            if(example.is_first_2_timesteps){
                example.pressure_prev=example.pressure;
            }
            else{
                auto temp_p=example.pressure;
                example.pressure=2*example.pressure-example.pressure_prev;
                example.pressure_prev=temp_p;
            }
        }
    }
}
//#####################################################################
namespace PhysBAM{
template class MPM_MAC_DRIVER<VECTOR<float,2> >;
template class MPM_MAC_DRIVER<VECTOR<float,3> >;
template class MPM_MAC_DRIVER<VECTOR<double,2> >;
template class MPM_MAC_DRIVER<VECTOR<double,3> >;
}

