//#####################################################################
// Copyright 2015, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __MPM_MAC_DRIVER__
#define __MPM_MAC_DRIVER__
#include <Core/Data_Structures/CHAINED_ARRAY.h>
#include <Core/Grids/FACE_INDEX.h>
#include <Core/Interpolation/AVERAGING_UNIFORM.h>
#include <Core/Krylov_Solvers/KRYLOV_VECTOR_WRAPPER.h>
#include <Core/Krylov_Solvers/MATRIX_SYSTEM.h>
#include <Core/Matrices/SPARSE_MATRIX_FLAT_MXN.h>
#include <Core/Vectors/VECTOR.h>
#include <Hybrid_Methods/Collisions/MPM_COLLISION_OBJECT.h>
#include <Hybrid_Methods/Regions/REGION_ARRAY.h>
#include <Hybrid_Methods/Regions/REGION_FACE_ARRAY.h>
#include <climits>
#include <functional>
namespace PhysBAM{

template<class TV> class MPM_MAC_EXAMPLE;
template<class TV> class PARTICLE_GRID_WEIGHTS;
template<class TV> class LAPLACE;
enum class RF;

template<class TV>
class MPM_MAC_DRIVER
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
public:
    enum {pressure_D=-1,pressure_N=-2,pressure_uninit=-3};

    int current_frame=0;

    MPM_MAC_EXAMPLE<TV>& example;

    CHAINED_ARRAY<int,IV> collision_object_lookup;
    ARRAY<T,FACE_INDEX<TV::m> > face_volumes;
    ARRAY<T,FACE_INDEX<TV::m> > face_fractions;

    struct ENTRY
    {
        REGION_FACE_FLAT_ID face;
        T theta;
        VECTOR<int,2> dof;
        VECTOR<T,2> data;
        int axis;
    };

    ARRAY<int,REGION_ID> region_starts;
    ARRAY<ENTRY> gradient_data;

    SPARSE_MATRIX_FLAT_MXN<T> poisson;
    MATRIX_SYSTEM<SPARSE_MATRIX_FLAT_MXN<T>,T,KRYLOV_VECTOR_WRAPPER<T,ARRAY<T> > > projection_system{poisson};
    KRYLOV_VECTOR_WRAPPER<T,ARRAY<T> >& sol;
    KRYLOV_VECTOR_WRAPPER<T,ARRAY<T> >& rhs;
    KRYLOV_VECTOR_WRAPPER<T,ARRAY<T> >& poisson_tmp;
    
    AVERAGING_UNIFORM<TV> avg;
    
    MPM_MAC_DRIVER(MPM_MAC_EXAMPLE<TV>& example);
    virtual ~MPM_MAC_DRIVER();

    void Execute_Main_Program();
    void Initialize();

    void Advance_One_Time_Step();
    void Simulate_To_Frame(const int frame_input);
    void Register_Output_Files();
    void Time_Integration();
    void Compute_Velocities_For_Move();
    void Update_Particle_Weights();
    void Prepare_Scatter();
    void Particle_To_Grid();
    void Grid_To_Particle();
    void Grid_To_Particle_Polypic();
    void Compute_Effective_Velocity();
    void Build_Level_Sets();
    void Prepare_Pressure_And_Viscosity();
    void Apply_Predictive_Pressure();
    void Pressure_Projection();
    void Apply_Forces();
    void Apply_Viscosity();
    void Level_Set_Apply_Viscosity();
    void Move_Particles();
    template <class D,class N>
    void Reflect_Boundary(D func_d,N func_n,RF flag) const;
    void Reflect_Boundary_Mass_Momentum(REGION_ARRAY<T,FACE_INDEX<TV::m> >& P) const;
    void Reflect_Boundary_Particle_Force(REGION_ARRAY<T,FACE_INDEX<TV::m> >& force) const;
    void Reflect_Boundary_Grid_Force(REGION_ARRAY<T,FACE_INDEX<TV::m> >& force) const;
    void Reflect_Boundary_Velocity_Copy_Only() const;
    T Compute_Dt() const;
    T Max_Particle_Speed() const;
    T Grid_V_Upper_Bound() const;
    void Update_Simulated_Particles();
    void Compute_Poisson_Matrix();
    template<class T2> void Fix_Periodic(ARRAY<T2,IV>& u) const;
    template<class T2> void Fix_Periodic_Node(ARRAY<T2,IV>& u) const;
    template<class T2> void Fix_Periodic(ARRAY<T2,FACE_INDEX<TV::m> >& u) const;
    template<class T2,class OP> void Fix_Periodic_Accum(ARRAY<T2,IV>& u,
        const OP& op) const;
    template<class T2,class OP> void Fix_Periodic_Accum(ARRAY<T2,FACE_INDEX<TV::m> >& u,
        const OP& op) const;

    void Compute_Boundary_Conditions();
    void Allocate_Pressure_Dofs();
    void Apply_Gradient(REGION_ARRAY<T,FACE_INDEX<TV::m> >& u,
        const ARRAY<T>& p,T scale) const;
    void Apply_Divergence(ARRAY<T>& div_u,
        const REGION_ARRAY<T,FACE_INDEX<TV::m> >& u,T scale) const;
    void Reseeding();
    void Step(std::function<void()> func,const char* name,bool dump_substep=true,bool do_step=true);
    void Invalidate_Particle(int p);
    bool Neumann_Boundary_Condition(const FACE_INDEX<TV::m>& face,T& bc) const;
    void Extrapolate_Inside_Object();
    void Level_Set_Pressure_Projection();
    void Compute_Regions();
    void Update_Collision_Object_Lookup();
    void Add_Viscosity(ARRAY<T,FACE_INDEX<TV::m> >& u_out,const ARRAY<T,FACE_INDEX<TV::m> >& u_in);

    // void func(MPM_COLLISION_OBJECT<TV>* obj,T phi);
    template<class F>
    void Visit_Collision_Objects(const TV& X,const F& func);

    // void func(const REGION_FACE_INDEX<TV::m>& ri,MPM_COLLISION_OBJECT<TV>* obj,T phi);
    template<class F>
    void Visit_Colliding_Region_Faces(const F& func);
    void Test_Poisson();
    void Print_Poisson_Matrix(int solve_id) const;
//#####################################################################
};
}
#endif
