//#####################################################################
// Copyright 2015, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Boundaries/BOUNDARY_PERIODIC.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Krylov_Solvers/KRYLOV_VECTOR_WRAPPER.h>
#include <Core/Level_Sets/LEVELSET.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Read_Write/OCTAVE_OUTPUT.h>
#include <Hybrid_Methods/Collisions/MPM_COLLISION_OBJECT.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_PARTICLES.h>
#include <Hybrid_Methods/Iterators/GATHER_SCATTER.h>
#include <Hybrid_Methods/Iterators/PARTICLE_GRID_WEIGHTS_SPLINE.h>
#include <Hybrid_Methods/Projection/MPM_PROJECTION_SYSTEM.h>
#include "MPM_MAC_EXAMPLE.h"
using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class TV> MPM_MAC_EXAMPLE<TV>::
MPM_MAC_EXAMPLE(const STREAM_TYPE stream_type)
    :stream_type(stream_type),levelset(*new LEVELSET<TV>(phi,ghost)),
    particles(*new MPM_PARTICLES<TV>),
    periodic_boundary(*new BOUNDARY_PERIODIC<TV,T>),
    debug_particles(DEBUG_PARTICLES<TV>::Get())
{
    side_bc_type.Fill(MPM::BC_SLIP);
    particles.Store_Region(true);
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> MPM_MAC_EXAMPLE<TV>::
~MPM_MAC_EXAMPLE()
{
    delete &particles;
    delete &periodic_boundary;
    delete &levelset;
    delete gather_scatter;
    for(int i=0;i<TV::m;i++) delete weights(i);
    collision_objects.Delete_Pointers_And_Clean_Memory();
    fluid_walls.Delete_Pointers_And_Clean_Memory();
}
//#####################################################################
// Function Register_Output_Files
//#####################################################################
template<class TV> void MPM_MAC_EXAMPLE<TV>::
Register_Output_Files()
{
    if(TV::m==2) viewer_output.Add_Global({.is_2d=true});
    if(TV::m==3) viewer_output.template Add_Axes<TV::m>("axes",{});
    viewer_output.Add("grid",grid,true,{.draw=(TV::m==2)});
    viewer_output.Add("debug_particles",DEBUG_PARTICLES<TV>::Get(),false,{});
    if(use_warm_start)
    {
        viewer_output.Add("pressure","grid",pressure_save,false,{.draw=false,.color_map=Color_Map_Jet()});
        viewer_output.Add("pressure_valid","grid",pressure_valid,false,{.draw=false});
    }
    viewer_output.Add("mpm_particles",particles,false,{});
    if(use_phi) viewer_output.Add_Levelset("levelset.phi","grid",levelset.phi,false,{},{0,1,1});

    if(!only_write_particles)
    {
        for(REGION_ID r(0);r<velocity.u.m;r++)
            viewer_output.Add("mac_velocities_"+std::to_string(Value(r)),"grid",velocity.u(r),false,{});
        for(REGION_ID r(0);r<mass.u.m;r++)
            viewer_output.Add_Scalar("mass_"+std::to_string(Value(r)),"grid",mass.u(r),false,{.draw=false});
        if(xpic)
            for(REGION_ID r(0);r<velocity_save.u.m;r++)
                viewer_output.Add("velocity_save_"+std::to_string(Value(r)),"grid",velocity_save.u(r),false,{.draw=false});
    }

    viewer_output.Add([this](){
        if(only_log){
            debug_particles.Clear_Debug_Particles();
            return;}
        if(this->use_test_output){
            std::string file=LOG::sprintf("%s/%s-%03d.txt",viewer_dir.output_directory.c_str(),test_output_prefix.c_str(),viewer_dir.frame_stack(0));
            OCTAVE_OUTPUT<T> oo(file.c_str());
            oo.Write("X",particles.X.Flattened());
            oo.Write("V",particles.V.Flattened());
            oo.Write("u",velocity.raw);}

#pragma omp parallel
#pragma omp single
        {
#pragma omp task
            Write_To_File(stream_type,viewer_dir.Frame_File("restart_data"),
                time,random,last_linear_momentum,last_angular_momentum,last_grid_te,last_grid_ke,
                last_part_te,last_part_ke);
#pragma omp task
            Write_To_File(stream_type,viewer_dir.Frame_File("mpm_particles_dl"),particles.deletion_list);
#pragma omp task
            if(use_phi)
                Write_To_File(stream_type,viewer_dir.Frame_File("levelset"),levelset);

            if(!only_write_particles){
#pragma omp task
                if(velocity.u.m)
                    Write_To_File(stream_type,viewer_dir.Frame_File("mac_velocities"),velocity.u(REGION_ID()));
                if(xpic){
#pragma omp task
                    if(velocity_save.u.m)
                        Write_To_File(stream_type,viewer_dir.Frame_File("velocity_save"),velocity_save.u(REGION_ID()));}
// #pragma omp task
//             {
//                 GRID<TV> ghost_grid=grid.Get_Ghost_Grid(ghost);
//                 for(int i=0;i<collision_objects.m;i++)
//                     if(IMPLICIT_OBJECT<TV>* io=collision_objects(i)->Get_Implicit_Object(time))
//                         Dump_Levelset(ghost_grid,at_nodes,*io,VECTOR<T,3>(0.7,0.3,0.3));
//                 debug_particles.Write_Debug_Particles(stream_type,viewer_dir);
//             }
            }
        }
    },false);
}
//#####################################################################
// Function Read_Output_Files
//#####################################################################
template<class TV> void MPM_MAC_EXAMPLE<TV>::
Read_Output_Files()
{
    Read_From_File(viewer_dir.Frame_File("mpm_particles"),particles);
    Read_From_File(viewer_dir.Frame_File("mpm_particles_dl"),particles.deletion_list);
    Read_From_File(viewer_dir.Frame_File("restart_data"),
        time,random,last_linear_momentum,last_angular_momentum,last_grid_te,last_grid_ke,
        last_part_te,last_part_ke);
    if(xpic)
        if(velocity_save.u.m)
            Read_From_File(viewer_dir.Frame_File("velocity_save"),velocity_save.u(REGION_ID()));
    if(use_warm_start)
    {
        Read_From_File(viewer_dir.Frame_File("pressure"),pressure_save);
        Read_From_File(viewer_dir.Frame_File("pressure_valid"),pressure_valid);
    }
    for(int i=0;i<read_output_files.m;i++) read_output_files(i)();
}
//#####################################################################
// Function Potential_Energy
//#####################################################################
template<class TV> typename TV::SCALAR MPM_MAC_EXAMPLE<TV>::
Potential_Energy(const T time) const
{
    T potential=0;
    for(int i=0;i<valid_flat_indices.m;i++){
        auto vi=valid_indices(i);
        potential-=mass(vi)*gravity(vi.f.axis)*grid.Face_X(vi.f)(vi.f.axis);}
    return potential;
}
//#####################################################################
// Function Total_Particle_Vorticity
//#####################################################################
template<class TV> typename TV::SCALAR MPM_MAC_EXAMPLE<TV>::
Total_Particle_Vorticity() const
{
    T l2_vort=0;
    auto* a_vort=particles.template Get_Array<typename TV::SPIN>("vort");
    gather_scatter->Gather(true,true,T(),[&l2_vort](T l2){l2_vort+=l2;},
        [this,a_vort](int p,PARTICLE_GRID_FACE_ITERATOR<TV>& it,T& l2)
        {
            typename TV::SPIN vort;
            for(;it.Valid();it.Next())
            {
                REGION_FACE_INDEX<TV::m> ri={particles.Region(p),it.Index()};
                vort+=it.Gradient().Cross(TV::Axis_Vector(it.axis)*velocity(ri));
            }
            if(a_vort) (*a_vort)(p)=vort;
            l2+=(T)0.5*particles.mass(p)*vort.Magnitude_Squared();
        });
    return l2_vort;
}
//#####################################################################
// Function Apply_Forces
//#####################################################################
template<class TV> void MPM_MAC_EXAMPLE<TV>::
Apply_Forces(const T time)
{
    for(int i=0;i<valid_flat_indices.m;i++){
        auto k=valid_flat_indices(i);
        auto f=valid_indices(i);
        TV af=Compute_Analytic_Force(grid.Face_X(f.f),time);
        velocity(k)+=dt*af(f.f.axis);
        velocity(k)+=dt*gravity(f.f.axis);}
}
//#####################################################################
// Function Compute_Analytic_Force
//#####################################################################
template<class TV> TV MPM_MAC_EXAMPLE<TV>::
Compute_Analytic_Force(const TV& X,T time) const
{
    return TV();
}
//#####################################################################
// Function Compute_Analytic_Pressure
//#####################################################################
template<class TV> TV::SCALAR MPM_MAC_EXAMPLE<TV>::
Compute_Analytic_Pressure(const TV& X,T time) const
{
    return T();
}
//#####################################################################
// Function Initialize_Prev_Velocity_Fields
//#####################################################################
template<class TV> void MPM_MAC_EXAMPLE<TV>::
Initialize_Prev_Velocity_Fields() const
{
    return;
}
//#####################################################################
// Function Set_Weights
//#####################################################################
template<class TV> void MPM_MAC_EXAMPLE<TV>::
Set_Weights(int order)
{
    for(int i=0;i<TV::m;i++){
        GRID<TV> face_grid=grid.Get_Face_Grid(i);
        if(order==2)
            weights(i)=new PARTICLE_GRID_WEIGHTS_SPLINE<TV,2>(face_grid);
        else if(order==3)
            weights(i)=new PARTICLE_GRID_WEIGHTS_SPLINE<TV,3>(face_grid);
        else PHYSBAM_FATAL_ERROR("Unrecognized interpolation order");}
}
//#####################################################################
// Function Total_Particle_Linear_Momentum
//#####################################################################
template<class TV> TV MPM_MAC_EXAMPLE<TV>::
Total_Particle_Linear_Momentum() const
{
    TV result;
#pragma omp parallel for
    for(int t=0;t<threads;t++){
        int a=t*simulated_particles.m/threads;
        int b=(t+1)*simulated_particles.m/threads;
        TV result_local;
        for(int k=a;k<b;k++){
            int p=simulated_particles(k);
            result_local+=particles.mass(p)*particles.V(p);}
#pragma omp critical
        result+=result_local;}
    return result;
}
//#####################################################################
// Function Total_Grid_Linear_Momentum
//#####################################################################
template<class TV> TV MPM_MAC_EXAMPLE<TV>::
Total_Grid_Linear_Momentum() const
{
    TV result;
#pragma omp parallel for
    for(int t=0;t<threads;t++){
        int a=t*valid_flat_indices.m/threads;
        int b=(t+1)*valid_flat_indices.m/threads;
        TV result_local;
        for(int k=a;k<b;k++){
            auto j=valid_flat_indices(k);
            result_local(valid_indices(k).f.axis)+=mass(j)*velocity(j);}
#pragma omp critical
        result+=result_local;}
    return result;
}
//#####################################################################
// Function Total_Particle_Angular_Momentum
//#####################################################################
template<class TV> typename TV::SPIN MPM_MAC_EXAMPLE<TV>::
Total_Particle_Angular_Momentum() const
{
    typename TV::SPIN result;
#pragma omp parallel for
    for(int t=0;t<threads;t++){
        int a=t*simulated_particles.m/threads;
        int b=(t+1)*simulated_particles.m/threads;
        typename TV::SPIN result_local;
        for(int k=a;k<b;k++){
            int p=simulated_particles(k);
            result_local+=particles.mass(p)*particles.X(p).Cross(particles.V(p));
            if(particles.store_C) result_local-=particles.mass(p)*Contract_Permutation_Tensor(particles.C(p));}
#pragma omp critical
        result+=result_local;}
    return result;
}
//#####################################################################
// Function Total_Grid_Angular_Momentum
//#####################################################################
template<class TV> typename TV::SPIN MPM_MAC_EXAMPLE<TV>::
Total_Grid_Angular_Momentum(T dt) const
{
    typename TV::SPIN result;
#pragma omp parallel for
    for(int t=0;t<threads;t++){
        int a=t*valid_indices.m/threads;
        int b=(t+1)*valid_indices.m/threads;
        typename TV::SPIN result_local;
        for(int k=a;k<b;k++){
            auto i=valid_indices(k);
            TV X=grid.Face_X(i.f);
            result_local+=mass(i)*X.Cross(velocity(i)*TV::Axis_Vector(valid_indices(k).f.axis));}
#pragma omp critical
        result+=result_local;}
    return result;
}
//#####################################################################
// Function Total_Grid_Kinetic_Energy
//#####################################################################
template<class TV> typename TV::SCALAR MPM_MAC_EXAMPLE<TV>::
Total_Grid_Kinetic_Energy() const
{
    T result=0;
#pragma omp parallel for reduction(+:result)
    for(int i=0;i<valid_flat_indices.m;i++){
        auto j=valid_flat_indices(i);
        auto f=valid_indices(i);
        if(!grid.Valid_Face(f.f)) continue;
        T scale=(T).5;
        if(f.f.index(f.f.axis)==0 || f.f.index(f.f.axis)==grid.numbers_of_cells(f.f.axis))
            scale=(T).25;
        result+=scale*mass(j)*sqr(velocity(j));
        //Add_Ball(this->grid.Face_X(f.f),(T).5*mass(j)*sqr(velocity(j)),VECTOR<T,3>(1,1,1));
    }
    return result;
}
//#####################################################################
// Function Total_Particle_Kinetic_Energy
//#####################################################################
template<class TV> typename TV::SCALAR MPM_MAC_EXAMPLE<TV>::
Total_Particle_Kinetic_Energy() const
{
    T result=0;
#pragma omp parallel for reduction(+:result)
    for(int k=0;k<simulated_particles.m;k++){
        int p=simulated_particles(k);
        auto V=particles.V(p);
        T result_local=V.Magnitude_Squared();
        if(particles.store_C)
        {
            T xi=weights(0)->Dp;
            auto C=particles.C(p);
            T a=C.Frobenius_Norm_Squared();
            result_local+=xi*a;
            if(particles.store_H)
            {
                auto H=particles.H(p);
                auto hc=Contract<1,2>(H);
                result_local+=(T).5*xi*xi*Sum_Of_Squares(H);
                result_local+=(T).25*xi*xi*hc.Magnitude_Squared();
                result_local+=xi*hc.Dot(V);
                for(int i=0;i<TV::m;i++)
                {
                    typename PARTICLE_GRID_WEIGHTS<TV>::SCRATCH sc;
                    weights(i)->Compute(p,sc,false,false);
                    TV e=sc.m3,f=sc.m4,c=C.Row(i),h=H.x(i).Diagonal_Part().x;
                    result_local+=(T).25*f.Dot(h*h)+e.Dot(c*h);
                    result_local-=(T).75*h.Dot(h)*xi*xi;
                }
            }
        }
        result+=particles.mass(p)/2*result_local;
        //Add_Ball(this->particles.X(p),result_local,VECTOR<T,3>(0,1,0));
    }
    return result;
}
//#####################################################################
// Function Average_Particle_Mass
//#####################################################################
template<class TV> typename TV::SCALAR MPM_MAC_EXAMPLE<TV>::
Average_Particle_Mass() const
{
    T result=0;
#pragma omp parallel for reduction(+:result)
    for(int k=0;k<simulated_particles.m;k++)
        result+=particles.mass(simulated_particles(k));
    return result/(T)particles.number;
}
//#####################################################################
// Function Add_Collision_Object
//#####################################################################
template<class TV> void MPM_MAC_EXAMPLE<TV>::
Add_Collision_Object(IMPLICIT_OBJECT<TV>* io,MPM::BC_TYPE type,T friction,std::function<FRAME<TV>(T)> func_frame,std::function<TWIST<TV>(T)> func_twist)
{
    auto* co=new MPM_COLLISION_OBJECT<TV>(io,type,friction);
    co->func_frame=func_frame;
    co->func_twist=func_twist;
    collision_objects.Append(co);
}
//#####################################################################
// Function Add_Fluid_Wall
//#####################################################################
template<class TV> void MPM_MAC_EXAMPLE<TV>::
Add_Fluid_Wall(IMPLICIT_OBJECT<TV>* io)
{
    fluid_walls.Append(io);
}
//#####################################################################
// Function Add_Callbacks
//#####################################################################
template<class TV> void MPM_MAC_EXAMPLE<TV>::
Add_Callbacks(bool is_begin,const char* func_name,std::function<void()> func)
{
    time_step_callbacks.Get_Or_Insert(func_name).second(is_begin).Append(func);
}
//#####################################################################
// Function Print_Grid_Stats
//#####################################################################
template<class TV> void MPM_MAC_EXAMPLE<TV>::
Print_Grid_Stats(const char* str)
{
    typename TV::SPIN am=Total_Grid_Angular_Momentum(dt);
    TV lm=Total_Grid_Linear_Momentum();
    T ke=Total_Grid_Kinetic_Energy();
    T pe=Potential_Energy(time);
    T te=ke+pe;
    LOG::cout<<str<<" linear  "<<"time " <<time<<" value "<<lm<<"  diff "<<(lm-last_linear_momentum)<<std::endl;
    LOG::cout<<str<<" angular "<<"time " <<time<<" value "<<am<<"  diff "<<(am-last_angular_momentum)<<std::endl;
    LOG::cout<<str<<" ke "<<"time " <<time<<" value "<<ke<<"  diff "<<(ke-last_grid_ke)<<std::endl;
    LOG::cout<<str<<" pe "<<"time " <<time<<" value "<<pe<<std::endl;
    LOG::cout<<str<<" total energy "<<"time " <<time<<" value "<<te<<" diff "<<(te-last_grid_te)<<std::endl;
    last_linear_momentum=lm;
    last_angular_momentum=am;
    last_grid_ke=ke;
    last_grid_te=te;
}
//#####################################################################
// Function Print_Grid_Stats
//#####################################################################
template<class TV> void MPM_MAC_EXAMPLE<TV>::
Print_Particle_Stats(const char* str)
{
    typename TV::SPIN am=Total_Particle_Angular_Momentum();
    TV lm=Total_Particle_Linear_Momentum();
    T ke=Total_Particle_Kinetic_Energy();
    T pe=Potential_Energy(time);
    T te=ke+pe;
    LOG::cout<<str<<" linear  "<<"time " <<time<<" value "<<lm<<"  diff "<<(lm-last_linear_momentum)<<std::endl;
    LOG::cout<<str<<" angular "<<"time " <<time<<" value "<<am<<"  diff "<<(am-last_angular_momentum)<<std::endl;
    LOG::cout<<str<<" ke "<<"time " <<time<<ke<<"  diff "<<(ke-last_part_ke)<<std::endl;
    LOG::cout<<str<<" pe "<<"time " <<time<<" value "<<pe<<std::endl;
    LOG::cout<<str<<" total energy "<<"time " <<time<<" value "<<te<<" diff "<<(te-last_part_te)<<std::endl;
    last_linear_momentum=lm;
    last_angular_momentum=am;
    last_part_ke=ke;
    last_part_te=te;
}
//#####################################################################
namespace PhysBAM{
template class MPM_MAC_EXAMPLE<VECTOR<float,2> >;
template class MPM_MAC_EXAMPLE<VECTOR<float,3> >;
template class MPM_MAC_EXAMPLE<VECTOR<double,2> >;
template class MPM_MAC_EXAMPLE<VECTOR<double,3> >;
}
