//#####################################################################
// Copyright 2015, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __MPM_MAC_EXAMPLE__
#define __MPM_MAC_EXAMPLE__
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Data_Structures/HASHTABLE.h>
#include <Core/Data_Structures/TUPLE.h>
#include <Core/Grids/GRID.h>
#include <Core/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>
#include <Hybrid_Methods/Collisions/MPM_COLLISION_OBJECT.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_HELPERS.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_TRANSFER_OPTIONS.h>
#include <Hybrid_Methods/Regions/REGION_FACE_ARRAY.h>
#include <functional>
namespace PhysBAM{

template<class TV> class LEVELSET;
template<class TV> class DEBUG_PARTICLES;
template<class TV> class GATHER_SCATTER;
template<class TV> class IMPLICIT_OBJECT;
template<class TV> class MPM_PARTICLES;
template<class TV> class PARTICLE_GRID_WEIGHTS;
template<class TV> class MPM_PLASTICITY_MODEL;
template<class TV> class PROJECTION_UNIFORM;
template<class TV> class KRYLOV_VECTOR_BASE;
template<class TV,class T2> class BOUNDARY_PERIODIC;

template<class TV>
class MPM_MAC_EXAMPLE
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
public:
    STREAM_TYPE stream_type;
    int ghost=3;
    GRID<TV> grid;
    
    REGION_ARRAY<T,FACE_INDEX<TV::m> > mass,mass_save;
    REGION_ARRAY<T,FACE_INDEX<TV::m> > velocity,velocity_save;
    REGION_ARRAY<bool,FACE_INDEX<TV::m> > valid_xfer_data;

    ARRAY<T> pressure,pressure_prev;
    ARRAY<T,IV> pressure_save;
    ARRAY<bool,IV> pressure_valid;
    bool use_warm_start=false;
    
    ARRAY<REGION_FACE_FLAT_ID> valid_flat_indices;
    ARRAY<REGION_FACE_INDEX<TV::m> > valid_indices;
    ARRAY<int> simulated_particles;
    GATHER_SCATTER<TV>* gather_scatter=0;
    ARRAY<T,REGION_ID> density; // if not using per-particle mass
    ARRAY<T,REGION_ID> viscosity; // if not using per-particle mass

    // signed distance field & level sets
    ARRAY<T,IV> phi;
    LEVELSET<TV>& levelset;
    bool disable_free_surface=false;
    bool push_particles_from_objects=true;

    // XPIC stuff
    int xpic=0;
    REGION_ARRAY<T,FACE_INDEX<TV::m> > xpic_v,xpic_v_star;
    ARRAY_VIEW<TV> effective_v;

    // particle stuff
    MPM_PARTICLES<TV>& particles;

    // grid stuff
    VECTOR<MPM::BC_TYPE,TV::m*2> side_bc_type; // -x, +x, -y, +y, -z, +z
    VECTOR<std::function<MPM::BC_TYPE(const TV& X,T)>,2*TV::m> bc_type;
    // Valid if BC_SLIP or BC_NOSLIP; velocity at face. null=0
    std::function<TV(const TV& X,T)> bc_velocity=0;
    // Valid if BC_FREE; pressure at ghost cell. null=0
    std::function<T(TV,T)> bc_pressure=0;
    BOUNDARY_PERIODIC<TV,T>& periodic_boundary;
    ARRAY<int,IV> cell_index;
    int dof=0;
    ARRAY<bool,FACE_INDEX<TV::m> > psi_N;
    REGION_ARRAY<T,FACE_INDEX<TV::m> > force;
    bool clamp_particles=true;

    // transfer stuff
    VECTOR<PARTICLE_GRID_WEIGHTS<TV>*,TV::m> weights;
    ARRAY_VIEW<TV> flip_adv_velocity;
    bool div_free_advection=false;
    bool cubic_advection=false;
    bool linear_advection=false;
    bool interpolated_advection=false;
    bool second_order=false;

    // objects
    ARRAY<MPM_COLLISION_OBJECT<TV>*> collision_objects;
    ARRAY<IMPLICIT_OBJECT<TV>* > fluid_walls;
    RANDOM_NUMBERS<T> random;

    // initialization & output
    int last_frame=100;
    std::string frame_title;
    int write_substeps_level=-1;
    int substeps_delay_frame=-1;
    VIEWER_DIR viewer_dir{"output"};
    VIEWER_OUTPUT viewer_output{stream_type,viewer_dir};
    std::string data_directory="../../Public_Data";
    std::string test_output_prefix;
    bool use_test_output=false;
    bool auto_restart=false;
    int restart=0;
    T dt=0,time=0,frame_dt=(T)1/24,min_dt=0,max_dt=(T)1/24;
    bool only_write_particles=false;
    bool only_log=false;
    bool is_first_timestep=true;
    bool is_first_2_timesteps=true;

    MPM_TRANSFER_OPTIONS<T> opt;

    // parameters
    TV gravity;
    T cfl=1;
    T solver_tolerance=std::numeric_limits<T>::epsilon()*10;
    int solver_iterations=1000;
    int threads=1;
    bool use_preconditioner=true;
    bool use_phi=false;
    int rk_particle_order=0;
    bool use_massless_particles=false;
    bool use_level_set_projection=false;
    bool use_reseeding=false;
    bool use_periodic_test_shift=false;
    bool use_bdf2=false;
    bool predictive_pressure=false;
    bool ab_pressure=false;
    IV periodic_test_shift;
    // d: default; c: always use pic update; p: use pic in the first step only
    char position_update='d';
    // for 2nd order C
    int C_order=1;
    bool enable_advection = true;
    bool enable_force = true;
    bool enable_pressure_projection = true;
    bool enable_viscosity = true;
    bool C_INT=false;
    bool H_INT=false;
    bool C_JH = false;
    bool init_prev = false;
    bool init_from_grid = false;
    bool one_substep = false;
    T f_timepoint = 0;
    T adv_v_timepoint = 0;
    bool write_output=true;
    bool check_particle_error=true;
    bool check_grid_error=true;
    bool check_pressure_error=true;
    
    // debugging
    DEBUG_PARTICLES<TV>& debug_particles;
    TV last_linear_momentum;
    typename TV::SPIN last_angular_momentum;
    T last_grid_te=0;
    T last_grid_ke=0;
    T last_part_te=0;
    T last_part_ke=0;
    bool test_system=false;
    bool print_matrix=false;
    bool particle_vort=false;
    bool use_object_extrap=false;
    bool use_mls_xfers=false;
    bool zero_invalid=false;
    bool dump_bc=false;
    bool output_face_fractions=false;
    T reseed_volume=0;
    
    MPM_MAC_EXAMPLE(const STREAM_TYPE stream_type_input);
    MPM_MAC_EXAMPLE(const MPM_MAC_EXAMPLE&) = delete;
    void operator=(const MPM_MAC_EXAMPLE&) = delete;
    virtual ~MPM_MAC_EXAMPLE();
    
    void Register_Output_Files();
    void Read_Output_Files();
    virtual void Initialize()=0;
    ARRAY<std::function<void(int frame)> > begin_frame;
    ARRAY<std::function<void(int frame)> > end_frame;
    ARRAY<std::function<void()> > read_output_files;

    HASHTABLE<std::string,pair<bool,VECTOR<ARRAY<std::function<void()> >,2> > > time_step_callbacks; // begin, end

    T Potential_Energy(const T time) const;
    T Total_Particle_Vorticity() const;
    void Apply_Forces(const T time);
    virtual TV Compute_Analytic_Force(const TV& X,const T time) const;
    virtual T Compute_Analytic_Pressure(const TV& X,const T time) const;
    virtual void Initialize_Prev_Velocity_Fields() const;
    void Set_Weights(int order);
    void Add_Collision_Object(IMPLICIT_OBJECT<TV>* io,MPM::BC_TYPE type,T friction,
        std::function<FRAME<TV>(T)> func_frame=0,std::function<TWIST<TV>(T)> func_twist=0);
    void Add_Fluid_Wall(IMPLICIT_OBJECT<TV>* io);
    template<class OBJECT> enable_if_t<!is_pointer<OBJECT>::value>
    Add_Collision_Object(const OBJECT& object,MPM::BC_TYPE type,T friction,
        std::function<FRAME<TV>(T)> func_frame=0,std::function<TWIST<TV>(T)> func_twist=0)
    {Add_Collision_Object(Make_IO(object),type,friction,func_frame,func_twist);}

    TV Total_Particle_Linear_Momentum() const;
    TV Total_Grid_Linear_Momentum() const;
    typename TV::SPIN Total_Grid_Angular_Momentum(T dt) const;
    typename TV::SPIN Total_Particle_Angular_Momentum() const;
    T Total_Grid_Kinetic_Energy() const;
    T Total_Particle_Kinetic_Energy() const;
    T Average_Particle_Mass() const;
    void Execute_Callbacks(bool is_begin,const char* func_name);
    void Add_Callbacks(bool is_begin,const char* func_name,std::function<void()> func);
    void Print_Grid_Stats(const char* str);
    void Print_Particle_Stats(const char* str);
//#####################################################################
};
}
#endif
