//#####################################################################
// Copyright 2015, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Basic_Geometry/SPHERE.h>
#include <Core/Grids_Uniform_Computations/LEVELSET_MAKER_UNIFORM.h>
#include <Core/Implicit_Objects/IMPLICIT_OBJECT_INTERSECTION.h>
#include <Core/Implicit_Objects/IMPLICIT_OBJECT_INVERT.h>
#include <Core/Implicit_Objects/IMPLICIT_OBJECT_UNION.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_PARTICLES.h>
#include "STANDARD_TESTS_3D.h"
#include <fstream>
namespace PhysBAM{
//#####################################################################
// Constructor
//#####################################################################
template<class T> STANDARD_TESTS<VECTOR<T,3> >::
STANDARD_TESTS(const STREAM_TYPE stream_type_input,PARSE_ARGS& parse_args)
    :STANDARD_TESTS_BASE<TV>(stream_type_input,parse_args)
{
    parse_args.Parse();
    if(!this->override_output_directory) viewer_dir.output_directory=LOG::sprintf("Test_3d_%i",test_number);
}
//#####################################################################
// Destructor
//#####################################################################
template<class T> STANDARD_TESTS<VECTOR<T,3> >::
~STANDARD_TESTS()
{
}
//#####################################################################
// Function Initialize
//#####################################################################
template<class T> void STANDARD_TESTS<VECTOR<T,3> >::
Initialize()
{
    REGION_ID rid(0);
    density.Append(unit_rho*scale_mass);
    viscosity.Append(this->arg_mu);
    switch(test_number)
    {
        case 1:{ // rotating sphere
            Set_Grid(RANGE<TV>::Unit_Box()*m);
            SPHERE<TV> sphere(TV(.5,.5,.5)*m,.3*m);
            VECTOR<T,3> angular_velocity(TV(0.4,0,0)/s);
            density(rid)=2*unit_rho*scale_mass;
            Seed_Particles(sphere,[=](const TV& X){return angular_velocity.Cross(X-sphere.center);},
                [=](const TV&){return Cross_Product_Matrix(angular_velocity);}
                ,density(rid),particles_per_cell);
            T total_mass=particles.mass.Sum();
            TV total_momentum=particles.V.Weighted_Sum(particles.mass);
            TV dV=total_momentum/total_mass;
            particles.V-=dV;
        } break;
        case 2:{ // Oscillating sphere
            Set_Grid(RANGE<TV>::Unit_Box()*m);
            SPHERE<TV> sphere(TV(.5,.5,.5)*m,.3*m);
            density(rid)=2*unit_rho*scale_mass;
            Seed_Particles(sphere,0,[=](const TV&){return MATRIX<T,3>();},
                density(rid),particles_per_cell);
        } break;
        case 3:{ // Freefall sphere
            Set_Grid(RANGE<TV>::Unit_Box()*m);
            SPHERE<TV> sphere(TV(.5,.5,.5)*m,.3*m);
            density(rid)=2*unit_rho*scale_mass;
            Seed_Particles(sphere,0,[=](const TV&){return MATRIX<T,3>();},
                density(rid),particles_per_cell);
        } break;
        case 4:{ // Dam break
            Set_Grid(RANGE<TV>::Unit_Box()*m);
            Seed_Particles(RANGE<TV>(TV(.0,.0,.0),TV(m/3,2*m/3,m)),0,0,density(rid),particles_per_cell);
            gravity=TV(0,-1,0)*m/sqr(s);
        } break;
        case 28:{ // inlet-cube
            T velocity=extra_T.m>=1?extra_T(0):0.25,stop_time=extra_T.m>=2?extra_T(1):200*frame_dt;
            Set_Grid(RANGE<TV>::Unit_Box());
            Seed_Particles(grid.domain,0,0,density(rid),particles_per_cell);
            T inlet_center=0.625,inlet_half_width=0.125;
            side_bc_type.Fill(MPM::BC_SLIP);
            side_bc_type(2)=MPM::BC_MIXED;
            this->bc_type(2)=[=,this](const TV& X,T time)
            {
                if(time>stop_time) return MPM::BC_SLIP;
                if(abs(X.x-0.2)<=0.125/2 && abs(X.z-0.2)<=0.125/2) return MPM::BC_FREE;
                else if(abs(X.x-0.9)<=0.125/2 && abs(X.z-0.9)<=0.125/2) return MPM::BC_FREE;
                else return MPM::BC_SLIP;
            };
            bc_velocity=[=,this](const TV& X,T time)
            {
                return TV(0,(time<=stop_time &&
                    abs(X.x-inlet_center)<=inlet_half_width &&
                    abs(X.z-inlet_center)<=inlet_half_width && X.y<0.001)?velocity:0,0);
            };
            RANGE<TV> source_range=grid.domain;
            source_range.min_corner.y-=.5*grid.dX.y;
            source_range.min_corner.x=inlet_center-inlet_half_width;
            source_range.max_corner.x=inlet_center+inlet_half_width;
            source_range.min_corner.z=inlet_center-inlet_half_width;
            source_range.max_corner.z=inlet_center+inlet_half_width;
            source_range.max_corner.y=0;
            TV X0=source_range.min_corner;
            X0.x+=2*inlet_half_width;
            X0.z+=2*inlet_half_width;
            Add_Source(X0,TV(0,1,0),TV(0,velocity,0),TV(),
                Make_IO(source_range),density(rid),particles_per_cell,true,[=,this](){return time<=stop_time;});
        } break;
        
        case 46:{ // Generalization of TG to 3D
            T mu=viscosity(rid);
            T A=1,B=1,C=1,k=1; // Arbitrary
            side_bc_type.Fill(MPM::BC_PERIODIC);
            Set_Grid(RANGE<TV>::Centered_Box()*pi*m/k);
            use_analytic_field=true;
            Add_Velocity(
                [=,this](auto X,auto t)
                {
                    auto x=X(0), y=X(1), z=X(2);
                    return Make_Vector<T>((A*sin(k*z)+C*cos(k*y))*exp(-mu*k*k*t), (B*sin(k*x)+A*cos(k*z))*exp(-mu*k*k*t), (C*sin(k*y)+B*cos(k*x))*exp(-mu*k*k*t));
                });
            Add_Pressure(
                [=,this](auto X,auto t)
                {
                    auto x=X(0), y=X(1), z=X(2);
                    return -exp(-2*mu*k*k*t)*(A*B*sin(k*x)*cos(k*z)+A*C*sin(k*z)*cos(k*y)+B*C*sin(k*y)*cos(k*x));
                });
            Setup_Analytic_Boundary_Conditions();
            Seed_Particles_Analytic(grid.domain,density(rid),particles_per_cell);
            end_frame.Append([=,this](int frame){Check_Analytic_Velocity();});
        } break;
        
        case 47:{ // Weird analytic fields as a general test case; Periodic BC; Divergence Free
            side_bc_type.Fill(MPM::BC_PERIODIC);
            Set_Grid(RANGE<TV>::Centered_Box()*pi*m);
            use_analytic_field=true;
            Add_Velocity(
                [=,this](auto X,auto t)
                {
                    auto x=X(0), y=X(1), z=X(2);
                    return Make_Vector<T>(2*cos(t+pi/6)*cos(2*x)*sin(3*y)*sin(z) + exp(t)*cos(y)/5,
                                           -cos(t+pi/6)*sin(2*x)*cos(3*y)*sin(z) + (1-t+t*t*5)*sin(z)/5,
                                           -cos(t+pi/6)*sin(2*x)*sin(3*y)*cos(z) + (1-t+t*t*5)*sin(x)/5);
                });
            Add_Pressure(
                [=,this](auto X,auto t)
                {
                    auto x=X(0), y=X(1), z=X(2);
                    return sin(t-pi/5)*exp(cos(2*x)*cos(y)*sin(3*z)-t);
                });
            Setup_Analytic_Boundary_Conditions();
            Seed_Particles_Analytic(grid.domain,density(rid),particles_per_cell);
            end_frame.Append([=,this](int frame){Check_Analytic_Velocity();});
        } break;
        
        case 100: this->Commandline_Analytic_test(); break;
        default: PHYSBAM_FATAL_ERROR("test number not implemented");
    }
    if(forced_collision_type!=-1)
        for(int i=0;i<collision_objects.m;i++)
            collision_objects(i)->type=(MPM::BC_TYPE)forced_collision_type;
}
template class STANDARD_TESTS<VECTOR<float,3> >;
template class STANDARD_TESTS<VECTOR<double,3> >;
}
