//#####################################################################
// Copyright 2015, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Analytic_Tests/ANALYTIC_SCALAR_IMPLICIT_OBJECT.h>
#include <Core/Fourier_Transforms/FFT.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/NODE_ITERATOR.h>
#include <Core/Images/PNG_FILE.h>
#include <Core/Implicit_Objects/IMPLICIT_OBJECT.h>
#include <Core/Interpolation/INTERPOLATED_COLOR_MAP.h>
#include <Core/Level_Sets/LEVELSET.h>
#include <Core/Log/DEBUG_SUBSTEPS.h>
#include <Core/Math_Tools/pow.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Meshes/TETRAHEDRALIZED_VOLUME.h>
#include <Core/Meshes/TRIANGULATED_AREA.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Seeding/POISSON_DISK.h>
#include <Deformables/Forces/DEFORMABLE_GRAVITY.h>
#include <Deformables/Forces/FINITE_VOLUME.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_PARTICLES.h>
#include <Hybrid_Methods/Iterators/GATHER_SCATTER.h>
#include <Hybrid_Methods/Iterators/PARTICLE_GRID_WEIGHTS_SPLINE.h>
#include <Hybrid_Methods/Projection/MPM_PROJECTION_SYSTEM.h>
#include <Hybrid_Methods/Seeding/MPM_PARTICLE_SOURCE.h>
#include "STANDARD_TESTS_BASE.h"
#include <fstream>
#ifdef USE_OPENMP
#include <omp.h>
#endif
namespace PhysBAM{
//#####################################################################
// Constructor
//#####################################################################
template<class TV> STANDARD_TESTS_BASE<TV>::
STANDARD_TESTS_BASE(const STREAM_TYPE stream_type_input,PARSE_ARGS& parse_args)
    :MPM_MAC_EXAMPLE<TV>(stream_type_input),test_number(0),resolution(32),user_resolution(false),stored_last_frame(0),
    user_last_frame(false),order(2),seed(1234),particles_per_cell(1<<TV::m),regular_seeding(false),
    no_regular_seeding(false),scale_mass(1),override_output_directory(false),
    m(1),s(1),kg(1),forced_collision_type(-1),dump_collision_objects(false),
    test_diff(false),bc_periodic(false),use_analytic_field(false),poisson_disk(*new POISSON_DISK<TV>(1))

{
    T framerate=0;
    bool use_quasi_exp_F_update=false;
    bool use_separate=false,use_slip=false,use_stick=false;
    bool print_stats=false;
    parse_args.Extra(&test_number,"example number","example number to run");
    parse_args.Add("-restart",&restart,"frame","restart frame");
    parse_args.Add("-resolution",&resolution,&user_resolution,"resolution","grid resolution");
    parse_args.Add("-substeps",&write_substeps_level,"level","output-substep level");
    parse_args.Add("-substeps_delay",&substeps_delay_frame,"frame","delay substeps until after this frame");
    parse_args.Add("-last_frame",&last_frame,&user_last_frame,"frame","number of frames to simulate");
    parse_args.Add("-threads",&threads,"threads","Number of threads");
    parse_args.Add("-o",&viewer_dir.output_directory,&override_output_directory,"dir","Output directory");
    parse_args.Add("-framerate",&framerate,"rate","Number of frames per second");
    parse_args.Add("-frame_dt",&frame_dt,"rate","Number of frames per second");
    parse_args.Add("-min_dt",&min_dt,"dt","Minimum time step size");
    parse_args.Add("-max_dt",&max_dt,"dt","Maximum time step size");
    parse_args.Add("-order",&order,"order","Interpolation basis order");
    parse_args.Add_Not("-no_affine",&opt.use_affine,"Use affine PIC");
    parse_args.Add("-affine",&opt.use_affine,"Use affine PIC");
    parse_args.Add("-print_stats",&print_stats,"Print momentum/energy stats");
    parse_args.Add("-only_write_particles",&only_write_particles,"Only write particle data (ignore grid data, restart data etc)");
    parse_args.Add("-only_log",&only_log,"Output log file only");
    parse_args.Add("-cfl",&cfl,"cfl","CFL number");
    parse_args.Add("-solver_tolerance",&solver_tolerance,"tol","Solver tolerance");
    parse_args.Add("-solver_iterations",&solver_iterations,"iter","Solver iterations");
    parse_args.Add("-seed",&seed,"seed","Random number seed");
    parse_args.Add("-particles_per_cell",&particles_per_cell,"num","Number of particles per cell");
    parse_args.Add("-scale_mass",&scale_mass,"scale","Scale mass of particles");
    parse_args.Add("-regular_seeding",&regular_seeding,"use regular particle seeding");
    parse_args.Add_Not("-no_regular_seeding",&no_regular_seeding,"use regular particle seeding");
    parse_args.Add("-use_exp_F",&use_quasi_exp_F_update,"Use an approximation of the F update that prevents inversion");
    parse_args.Add("-m",&m,"scale","meter scale");
    parse_args.Add("-s",&s,"scale","second scale");
    parse_args.Add("-kg",&kg,"scale","kilogram scale");
    parse_args.Add("-slip",&use_slip,"force slip collisions");
    parse_args.Add("-stick",&use_stick,"force stick collisions");
    parse_args.Add("-separate",&use_separate,"force separating collisions");
    parse_args.Add("-T",&extra_T,"float","extra float argument");
    parse_args.Add("-I",&extra_int,"int","extra int argument");
    parse_args.Add("-dump_collisions",&dump_collision_objects,"dump out collision objects");
    parse_args.Add("-test_system",&test_system,"run sanity checks on Krylov systems");
    parse_args.Add("-print_matrix",&print_matrix,"print Krylov matrix");
    parse_args.Add("-test_output_prefix",&test_output_prefix,&use_test_output,"","prefix to use for test output");
    parse_args.Add_Not("-no_preconditioner",&this->use_preconditioner,"disable preconditioner");
    parse_args.Add("-flip",&opt.flip,"frac","Use this fraction of flip in transfers");
    parse_args.Add("-div_free",&div_free_advection,"Use div-free interpolated velocity to move particles");
    parse_args.Add("-cubic",&cubic_advection,"Use cubic interpolated velocity to move particles");
    parse_args.Add("-linear",&linear_advection,"Use linear interpolated velocity to move particles");
    parse_args.Add("-second",&second_order,"Use second order interpolated advection");
    parse_args.Add("-test_diff",&test_diff,"test analytic derivatives");
    parse_args.Add("-bc_periodic",&bc_periodic,"set boundary condition periodic");
    parse_args.Add("-test_periodic",&use_periodic_test_shift,"test periodic bc");
    parse_args.Add("-mu",&arg_mu,"mu","viscosity");
    parse_args.Add("-analyze_u_modes",&analyze_u_modes,"Perform FFT analysis on velocity");
    parse_args.Add("-analyze_energy_vort",&analyze_energy_vort,"Analyze energy and vorticity");
    parse_args.Add("-dump_modes_freq",&dump_modes_freq,"num","Dump FFT modes every num time steps");
    parse_args.Add("-max_ke",&max_ke,"value","Normalization for FFT images");
    parse_args.Add_Not("-no_clamp",&clamp_particles,"clamp particles on wall");
    parse_args.Add("-particle_vort",&particle_vort,"output particle vorticity");
    parse_args.Add("-xpic",&xpic,"int","XPIC order");
    parse_args.Add("-pos_update",&position_update,"char","position update");
    parse_args.Add("-obj_extrap",&use_object_extrap,"extrapolate velocities inside objects");
    parse_args.Add("-rk",&rk_particle_order,"order","rk order");
    parse_args.Add("-mls",&this->use_mls_xfers,"Use moving least squares transfers");
    parse_args.Add("-lsproj",&this->use_level_set_projection,"Use level sets for projection");
    parse_args.Add("-no_surface",&this->disable_free_surface,"Do not construct free surface level set");
    parse_args.Add("-reseed",&this->use_reseeding,"Reseed particles");
    parse_args.Add("-analytic_u",&analytic_u_expr,"expr","Analytic velocity expression");
    parse_args.Add("-analytic_p",&analytic_p_expr,"expr","Analytic pressure expression");
    parse_args.Add("-analytic_phi",&analytic_phi_expr,"expr","Analytic level set expression");
    parse_args.Add("-grid_min",&analytic_grid_range.min_corner,"vec","Min corner of grid domain");
    parse_args.Add("-grid_max",&analytic_grid_range.max_corner,"vec","Max corner of grid domain");
    parse_args.Add("-analytic_bc",&analytic_bc_types,"expr","BC types [fsnp], 2*d entries");
    parse_args.Add("-warm_start",&this->use_warm_start,"use warm starts for pressure projection");
    parse_args.Add("-mls_zero_invalid",&this->zero_invalid,"Use zero velocity for invalid face");
    parse_args.Add("-dump_bc",&this->dump_bc,"Use zero velocity for invalid face");
    parse_args.Add_Not("-no_push",&this->push_particles_from_objects,"Do not push particles out from objects");
    parse_args.Add("-face_fractions",&this->output_face_fractions,"Dump face fractions for last time step of frame");
    parse_args.Add_Not("-no_move",&this->enable_advection,"Disable advection");
    parse_args.Add_Not("-no_f",&this->enable_force,"Disable external and analytic force");
    parse_args.Add_Not("-no_p",&this->enable_pressure_projection,"Disable pressure projection");
    parse_args.Add_Not("-no_mu",&this->enable_viscosity,"Disable viscosity");
    parse_args.Add("-C_order",&this->C_order,"int","C matrix update order. Default is 1; 2 for 2nd order; other numbers for the inverse version");
    parse_args.Add("-refine",&refine,"int","refine grid resolution and frame_dt by this number. Default is 0 which means this isn't turned on. If turned on, max_dt and min_dt will be ignored.");
    parse_args.Add("-fix_dt",&fix_dt,"fix time step size");
    parse_args.Add("-use_bdf2",&this->use_bdf2,"Use BDF-2 for time integration");
    parse_args.Add("-pred_p",&this->predictive_pressure,"Use predictive pressure");
    parse_args.Add("-ab_p",&this->ab_pressure,"Use Adam-Bashforth time integration on pressure");
    parse_args.Add("-f_timepoint",&this->f_timepoint,"timepoint","evaluate analytic force at t_n + dt * f_timepoint");
    parse_args.Add("-adv_v_timepoint",&this->adv_v_timepoint,"timepoint","evaluate velocity for move at t_{n+k} where k is this number. Compute v_{n+k} as linear combination of v_n and v_{n-1}");
    parse_args.Add("-C_INT",&this->C_INT,"Perform time integration on C matrix");
    parse_args.Add("-C_JH",&this->C_JH,"Turn on Han's C update formula");
    parse_args.Add("-H_INT",&this->H_INT,"Perform time integration on H tensor");
    parse_args.Add("-polypic6",&opt.use_polypic,"Use PolyPIC-6 for locally quadratic v field");
    parse_args.Add("-init_prev",&this->init_prev,"Initialze v, C and p at timestep -1, using analytic expressions. Only useful when advection is turned off");
    parse_args.Add("-one_substep",&this->one_substep,"1 substep per frame");
    parse_args.Add_Not("-no_check",&this->write_output,"Disable writing output file");
    parse_args.Add_Not("-no_particle_err",&this->check_particle_error,"Disable checking particle velocity error");
    parse_args.Add_Not("-no_grid_err",&this->check_grid_error,"Disable checking grid velocity error");
    parse_args.Add_Not("-no_pressure_err",&this->check_pressure_error,"Disable checking grid velocity error");
    parse_args.Add("-second_order_pic",&second_order_pic,"Turn on everything required for second order PIC");
    parse_args.Add("-seed_threaded",&seed_threaded,"Turn on everything required for second order PIC");
    
    parse_args.Parse(true);
    PHYSBAM_ASSERT((int)use_slip+(int)use_stick+(int)use_separate<=1);
    if(use_slip) forced_collision_type=MPM::BC_SLIP;
    if(use_stick) forced_collision_type=MPM::BC_NOSLIP;
    if(use_separate) forced_collision_type=MPM::BC_SEP;

    if(this->use_level_set_projection && !this->disable_free_surface)
        this->use_phi=true;

    if(second_order_pic)
    {
        this->use_bdf2=true;
        this->f_timepoint=1;
        this->adv_v_timepoint=0.5;
        this->C_INT=true;
        this->H_INT=true;
        opt.use_polypic=true;
    }

    unit_p=kg*pow<2-TV::m>(m)/(s*s);
    unit_rho=kg*pow<-TV::m>(m);
    unit_mu=kg*pow<2-TV::m>(m)/s;
    min_dt*=s;
    max_dt*=s;
    if(refine)
    {
        max_dt=frame_dt/refine;
        min_dt=max_dt;
        resolution*=refine;
        user_resolution=true;
    }
    viscosity*=unit_mu;

    this->interpolated_advection = div_free_advection | cubic_advection | linear_advection;
    
    if(fix_dt) min_dt=max_dt;

    if(framerate) frame_dt=1/framerate;
    frame_dt*=s;

#ifdef USE_OPENMP
    omp_set_num_threads(threads);
#pragma omp parallel
#pragma omp single
    {
        if(omp_get_num_threads()!=threads) PHYSBAM_FATAL_ERROR();
        LOG::cout<<"Running on "<<threads<<" threads"<<std::endl;
    }
#else
    PHYSBAM_ASSERT(threads==1);
#endif

    stored_last_frame=last_frame;
    random.Set_Seed(seed);

    this->reseed_volume=grid.dX.Product()/particles_per_cell;
    
    particles.Store_C(opt.use_affine);
    particles.Store_H(opt.use_polypic);

    if(analyze_u_modes){
        Add_Callbacks(false,"p2g",[this](){
                static bool done=false;
                if(!done) Velocity_Fourier_Analysis();
                done=true;});
        Add_Callbacks(false,"time-step",[this](){
                Velocity_Fourier_Analysis();});}

    if(analyze_energy_vort)
        end_frame.Append([=,this](int frame){Energy_Vorticity_Analysis();});

    if(use_periodic_test_shift){
        auto shift_func=[this](int sign){
                TV shift=TV(sign*periodic_test_shift)*grid.dX;
                for(int i=0;i<particles.X.m;i++)
                    particles.X(i)=wrap(particles.X(i)+shift,grid.domain.min_corner,grid.domain.max_corner);};
        Add_Callbacks(true,"time-step",[=,this](){shift_func(1);});
        Add_Callbacks(false,"time-step",[=,this](){shift_func(-1);});
        Add_Callbacks(false,"initialize",[=,this](){
                RANDOM_NUMBERS<T> local_random;
                for(int i=0;i<TV::m;i++)
                    periodic_test_shift(i)=local_random.Get_Uniform_Integer(0,grid.numbers_of_cells(i)-1);
                LOG::printf("shift %P\n",periodic_test_shift);});}

    if(print_stats){
        auto stats_p=[=,this](const char* name){Add_Callbacks(false,name,[=,this](){Print_Particle_Stats(name);});};
        auto stats_g=[=,this](const char* name){Add_Callbacks(false,name,[=,this](){Print_Grid_Stats(name);});};
        stats_p("update-weights");
        stats_g("p2g");
        stats_g("forces");
        stats_g("projection");
        stats_g("viscosity");
        stats_p("g2p");}
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> STANDARD_TESTS_BASE<TV>::
~STANDARD_TESTS_BASE()
{
    for(int i=0;i<destroy.m;i++) destroy(i)();
    delete &poisson_disk;
    delete analytic_velocity;
    delete analytic_pressure;
    delete analytic_level_set;
}
//#####################################################################
// Function Seed_Particles
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Seed_Particles_Poisson(IMPLICIT_OBJECT<TV>& object,std::function<TV(const TV&)> V,
    std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,T particles_per_cell,
    std::function<VECTOR<T,3>(const TV&)>* color)
{
    ARRAY<TV> X;
    for(int i=0;i<TV::m;i++) poisson_disk.is_periodic(i)=(side_bc_type(i)==MPM::BC_PERIODIC);
    poisson_disk.periodic_boundary=grid.domain;
    poisson_disk.Set_Distance_By_Volume(grid.dX.Product()/particles_per_cell);
    poisson_disk.Sample(random,object,X,0,seed_threaded);
    T volume=grid.dX.Product()/particles_per_cell;
    T mass=density*volume;
    for(int i=0;i<X.m;i++)
        if(color) Add_Particle(X(i),V,dV,mass,volume,(*color)(X(i)));
        else Add_Particle(X(i),V,dV,mass,volume);
}
//#####################################################################
// Function Seed_Particles
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Seed_Particles_Uniform(IMPLICIT_OBJECT<TV>& object,std::function<TV(const TV&)> V,
    std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,const GRID<TV>& seed_grid,
    std::function<VECTOR<T,3>(const TV&)>* color)
{
    T volume=seed_grid.dX.Product();
    T mass=density*volume;
    for(CELL_ITERATOR<TV> it(seed_grid);it.Valid();it.Next())
    {
        TV X=it.Location();
        if(object.Lazy_Inside(X))
        {
            if(color) Add_Particle(X,V,dV,mass,volume,(*color)(X));
            else Add_Particle(X,V,dV,mass,volume);
        }
    }
}
//#####################################################################
// Function Seed_Particles
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Seed_Particles(IMPLICIT_OBJECT<TV>& object,std::function<TV(const TV&)> V,
    std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,T particles_per_cell,
    std::function<VECTOR<T,3>(const TV&)>* color)
{
    if(test_diff) Test_dV(V,dV);

    if(!regular_seeding) return Seed_Particles_Poisson(object,V,dV,density,particles_per_cell,color);

    object.Update_Box();
    RANGE<IV> range=grid.Clamp_To_Cell(object.Box(),3).Intersect(grid.Cell_Indices());
    TV LB=grid.Node_X(range.min_corner);
    TV UB=grid.Node_X(range.max_corner);
    T scale=pow<1,TV::m>((T)particles_per_cell);
    GRID<TV> seed_grid(IV(TV(range.Edge_Lengths())*scale),RANGE<TV>(LB,UB));
    Seed_Particles_Uniform(object,V,dV,density,seed_grid,color);
}
//#####################################################################
// Function Add_Particle
//#####################################################################
template<class TV> int STANDARD_TESTS_BASE<TV>::
Add_Particle(const TV& X,std::function<TV(const TV&)> V,std::function<MATRIX<T,TV::m>(const TV&)> dV,
    const T mass,const T volume,const VECTOR<T,3>& color)
{
    return Add_Particle(X,V?V(X):TV(),dV?dV(X):MATRIX<T,TV::m>(),mass,volume,color);
}
//#####################################################################
// Function Add_Particle
//#####################################################################
template<class TV> int STANDARD_TESTS_BASE<TV>::
Add_Particle(const TV& X,const TV& V,const MATRIX<T,TV::m>& dV,const T mass,const T volume,const VECTOR<T,3>& color)
{
    int p=particles.Add_Element_From_Deletion_List();
    particles.valid(p)=true;
    particles.X(p)=X;
    particles.V(p)=V;
    particles.F(p)=MATRIX<T,TV::m>()+1;
    if(particles.store_Fp) particles.Fp(p).Set_Identity_Matrix();
    if(particles.store_C)
        for(int a=0;a<TV::m;a++)
            particles.C(p).Set_Row(a,dV.Row(a));
    if(particles.store_S) particles.S(p)=SYMMETRIC_MATRIX<T,TV::m>()+1;
    particles.mass(p)=mass;
    particles.volume(p)=volume;
    ARRAY_VIEW<VECTOR<T,3> >* color_attribute=particles.template Get_Array<VECTOR<T,3> >("color");
    (*color_attribute)(p)=color;
    return p;
}
//#####################################################################
// Function Set_Grid
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Set_Grid(const RANGE<TV>& domain,IV resolution_scale,int default_resolution)
{
    if(!user_resolution) resolution=default_resolution;
    grid.Initialize(resolution_scale*resolution,domain);
    Set_Weights(order);
}
//#####################################################################
// Function Set_Grid
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Set_Grid(const RANGE<TV>& domain,IV resolution_scale,IV resolution_padding,
    int resolution_multiple,int default_resolution)
{
    if(!user_resolution) resolution=default_resolution;
    int scaled_resolution=(resolution+resolution_multiple-1)/resolution_multiple;
    grid.Initialize(resolution_scale*scaled_resolution+resolution_padding,domain);
    Set_Weights(order);
}
//#####################################################################
// Function Test_dV
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Test_dV(std::function<TV(const TV&)> V,std::function<MATRIX<T,TV::m>(const TV&)> dV) const
{
    PHYSBAM_ASSERT(V || !dV);
    if(!dV) return;
    RANDOM_NUMBERS<T> rand;
    T eps=(T)1e-6;
    TV x0,dx;
    rand.Fill_Uniform(x0,-(T)1,(T)1);
    rand.Fill_Uniform(dx,-eps,eps);
    TV v0=V(x0),v1=V(x0+dx);
    MATRIX<T,TV::m> dv0=dV(x0),dv1=dV(x0+dx);
    TV a=(v1-v0)/eps,b=(T).5/eps*(dv0+dv1)*dx,c=a-b;
    T ma=a.Magnitude(),mb=b.Magnitude(),mc=c.Magnitude();
    LOG::printf("dV %g %g %g  rel %g\n",ma,mb,mc,mc/max(ma,mb,(T)1e-30));
}
//#####################################################################
// Function Check_Analytic_Velocity
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Check_Analytic_Velocity(std::function<bool(const FACE_INDEX<TV::m>&)> valid_face) const
{
    if(!use_analytic_field) return;
    REGION_ID rid(0);
    T max_error=0,l2_error=0;
    int num_l2_samples=0;
    
    if(this->check_grid_error)
    {
        for(REGION_ID r(0);r<mass.u.m;r++)
        {
            for(FACE_RANGE_ITERATOR<TV::m> it(mass.u(r).domain);it.Valid();it.Next())
            {
                REGION_FACE_INDEX<TV::m> ri={r,it.face};
                if(!mass(ri)) continue;
                if(!valid_face(it.face)) continue;
                if(this->use_mls_xfers && !this->valid_xfer_data(ri)) continue;
                T u=velocity(ri);
                TV X=grid.Face_X(it.face);
                TV v=analytic_velocity->f(X,time);
                T e=abs(u-v(it.face.axis));
                max_error=max(max_error,e);
                l2_error+=sqr(e);
                num_l2_samples++;
                Add_Ball(X,e,VECTOR<T,3>(1,1,0));
            }
        }
        PHYSBAM_DEBUG_WRITE_SUBSTEP("grid error",1);
        if(num_l2_samples) l2_error/=num_l2_samples;
        l2_error=sqrt(l2_error);
        LOG::printf("grid velocity error: L-inf: %g L-2: %g N: %i\n",max_error,l2_error,num_l2_samples);
    }
    
    if(this->check_particle_error)
    {
        max_error=0;
        l2_error=0;
        num_l2_samples=0;
    
        for(int p=0;p<particles.X.m;p++){
            if(!particles.valid(p)) continue;
            TV e=particles.V(p)-analytic_velocity->f(particles.X(p),time);
            max_error=max(max_error,e.Max_Abs());
            l2_error+=e.Magnitude_Squared();
            num_l2_samples+=TV::m;
            PHYSBAM_ASSERT(l2_error<=num_l2_samples*max_error*max_error+1e-10);
            Add_Arrow(particles.X(p),e,VECTOR<T,3>(0,1,0));
        }
        PHYSBAM_DEBUG_WRITE_SUBSTEP("particle error",1);
        if(num_l2_samples) l2_error/=num_l2_samples;
        l2_error=sqrt(l2_error);
        LOG::printf("particle velocity error: L-inf: %g L-2: %g N: %i\n",max_error,l2_error,num_l2_samples);
    }
    
    if(this->check_pressure_error)
    {
        max_error=0;
        l2_error=0;
        num_l2_samples=0;
        for(REGION_ID r(0);r<mass.u.m;r++)
        {
            for(CELL_ITERATOR<TV> it(this->grid);it.Valid();it.Next()){
                T e=this->pressure(this->cell_index(it.index))-this->Compute_Analytic_Pressure(it.Location(),time);
                max_error=max(max_error,abs(e));
                l2_error+=e*e;
                num_l2_samples++;
                Add_Ball(it.Location(),e,VECTOR<T,3>(0,1,1));
            }
            PHYSBAM_DEBUG_WRITE_SUBSTEP("pressure error",1);
            if(num_l2_samples) l2_error/=num_l2_samples;
            l2_error=sqrt(l2_error);
            LOG::printf("pressure error: L-inf: %g L-2: %g N: %i\n",max_error,l2_error,num_l2_samples);
        }
    }
}
//#####################################################################
// Function Compute_Analytic_Force
//#####################################################################
template<class TV> TV STANDARD_TESTS_BASE<TV>::
Compute_Analytic_Force(const TV& X,T time) const
{
    if(!use_analytic_field) return TV();
    const ANALYTIC_VECTOR<TV>* av=analytic_velocity;
    const ANALYTIC_SCALAR<TV>* ap=analytic_pressure;
    REGION_ID rid(0);
    TV F=av->dt(X,time);
    if(this->enable_advection) F+=av->dX(X,time)*av->f(X,time);
    if(this->enable_pressure_projection) F+=1/density(rid)*ap->dX(X,time);
    if(this->enable_viscosity) F-=viscosity(rid)/density(rid)*av->L(X,time);
    return F;
}
//#####################################################################
// Function Compute_Analytic_Pressure
//#####################################################################
template<class TV> TV::SCALAR STANDARD_TESTS_BASE<TV>::
Compute_Analytic_Pressure(const TV& X,T time) const
{
    if(!use_analytic_field) return T();
    const ANALYTIC_SCALAR<TV>* ap=analytic_pressure;
    return ap->f(X,time);
}
//#####################################################################
// Function Initialize_Velocity_Hessians
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Initialize_Velocity_Hessians() const
{
    if(!use_analytic_field) return;
    const ANALYTIC_VECTOR<TV>* av=analytic_velocity;
#pragma omp parallel for
    for(int p=0;p<this->particles.X.m;p++){
        this->particles.H(p)=av->ddX(particles.X(p),0);
    }
}
//#####################################################################
// Function Initialize_Prev_Velocity_Fields
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Initialize_Prev_Velocity_Fields() const
{
    if(!use_analytic_field) return;
    const ANALYTIC_VECTOR<TV>* av=analytic_velocity;
    auto& v_prev=*this->particles.template Get_Array<TV>("V_prev");
    auto& C_prev=*this->particles.template Get_Array<MATRIX<T,TV::m>>("C_prev");
#pragma omp parallel for
    for(int p=0;p<this->particles.X.m;p++){
        v_prev(p)=av->f(particles.X(p),-max_dt);
        C_prev(p)=av->dX(particles.X(p),-max_dt);
    }
    if(opt.use_polypic){
        auto& H_prev=*this->particles.template Get_Array<SYMMETRIC_TENSOR<T,0,TV::m,TV::m>>("H_prev");
#pragma omp parallel for
        for(int p=0;p<this->particles.X.m;p++){
            H_prev(p)=av->ddX(particles.X(p),-max_dt);
        }
    }
}
//#####################################################################
// Function Dump_Image
//#####################################################################
template<class T> static void
Dump_Image(const std::string& file,const ARRAY<T,VECTOR<int,2> >& ke)
{
    INTERPOLATED_COLOR_MAP<T> icm;
    icm.Initialize_Colors((T)1e-8,1,true,true,false);
    
    ARRAY<VECTOR<T,3>,VECTOR<int,2> > image(ke.domain);
    for(RANGE_ITERATOR<2> it(image.domain);it.Valid();it.Next())
        image(it.index)=icm(ke(it.index));

    ARRAY<VECTOR<T,3>,VECTOR<int,2> > bar(VECTOR<int,2>(1000,1));
    for(int i=0;i<1000;i++)
        bar(VECTOR<int,2>(i,0))=icm((i/(T)999)*(icm.mx-icm.mn)+icm.mn);
    PNG_FILE<T>::Write("bar.png",bar);

    PNG_FILE<T>::Write(file,image);
}
//#####################################################################
// Function Dump_Image
//#####################################################################
template<class T,int d> static void
Dump_Image(const std::string& file,const ARRAY<T,VECTOR<int,d> >& ke)
{
}
//#####################################################################
// Function Energy_Vorticity_Analysis
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Energy_Vorticity_Analysis() const
{
    static int id=-1;
    id++;
    T pe_particle=0;
    for(int p=0;p<particles.X.m;p++){
        if(!particles.valid(p)) continue;
        pe_particle-=particles.mass(p)*this->gravity.Dot(particles.X(p));}

    auto* vort=particles.template Get_Array<typename TV::SPIN>("vort");
    if(vort){
        std::ofstream fout(LOG::sprintf("%s/pvort-%i.txt",viewer_dir.output_directory,id).c_str());
        fout<<"vort x y\n";
        for(int p=0;p<particles.X.m;p++){
            if(!particles.valid(p)) continue;
            fout<<(*vort)(p)<<" "<<particles.X(p)[0]<<" "<<particles.X(p)[1]<<"\n";}}
    LOG::printf("ke grid %P  ke particle %P\n",
        this->Total_Grid_Kinetic_Energy(),
        this->Total_Particle_Kinetic_Energy());
    LOG::printf("te grid %P  te particle %P\n",
        this->Total_Grid_Kinetic_Energy()+this->Potential_Energy(time),
        this->Total_Particle_Kinetic_Energy()+pe_particle);
    LOG::printf("vort particle %P\n",this->Total_Particle_Vorticity());
}
//#####################################################################
// Function Velocity_Fourier_Analysis
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Velocity_Fourier_Analysis() const
{
    static int id=-1;
    id++;
    REGION_ID rid(0);
    LOG::printf("analysis id %i\n",id);
    FFT<TV> fft;
    ARRAY<T,IV> ua(grid.Cell_Indices()),ke(ua.domain);
    ARRAY<std::complex<T>,IV> out(ua.domain);
    
    TV coefficients=(T)(2*pi)/grid.domain.Edge_Lengths();
    IV counts=grid.numbers_of_cells,hi=counts/2,lo=hi-counts;
    ARRAY<T> bins((int)rint((coefficients*TV((hi-1).Componentwise_Max(-lo))).Magnitude())+1);
    int taylor_modes=extra_int.m>=1?extra_int(0):1;
    T total_taylor=0;
    for(int a=0;a<TV::m;a++){
        ua.Put(velocity.u(rid).Component(a),ua);
        fft.Transform(ua,out);
        out.array/=sqrt((T)ke.domain.Size());
        for(RANGE_ITERATOR<TV::m> it(ke.domain);it.Valid();it.Next()){
            TV k=coefficients*TV(wrap(it.index,lo,hi));
            ke(wrap(it.index+counts/2,IV(),counts))=sqr(abs(out(it.index))/max_ke);
            bins((int)rint(k.Magnitude()))+=ke(it.index);}
        if(id%dump_modes_freq==0)
            Dump_Image(LOG::sprintf("%s/modes-%i-%c.png",viewer_dir.output_directory,id,"xyz"[a]),ke);
        if(taylor_modes>0)
            for(RANGE_ITERATOR<TV::m> it(RANGE<IV>::Unit_Box()*2);it.Valid();it.Next())
                total_taylor+=sqr(abs(out(wrap((it.index*2-1)*taylor_modes,IV(),grid.numbers_of_cells))));
        else total_taylor+=sqr(abs(out(IV())));}
    std::ofstream fout(LOG::sprintf("%s/bins-%i.txt",viewer_dir.output_directory,id).c_str());
    for(int i=0;i<bins.m;i++)
        fout<<i<<" "<<bins(i)<<std::endl;

    T l2_u=0;
    int num_l2_u=0;
    auto valid=[&](FACE_INDEX<TV::m> face){return mass.u(rid)(face) && (this->psi_N.domain.Empty() || !this->psi_N(face));};
    for(REGION_ID r(0);r<mass.u.m;r++)
    {
        for(FACE_RANGE_ITERATOR<TV::m> it(mass.u(r).domain);it.Valid();it.Next())
        {
            REGION_FACE_INDEX<TV::m> ri={r,it.face};
            if(side_bc_type(2*it.face.axis)==MPM::BC_PERIODIC)
                if(it.face.index(it.face.axis)==grid.numbers_of_cells(it.face.axis))
                    continue;
            if(!valid(it.face)) continue;
            l2_u+=sqr(velocity(ri));
            num_l2_u++;
        }
    }
    if(num_l2_u) l2_u/=num_l2_u;
    if(num_l2_u) total_taylor/=num_l2_u;

    T l2_omega=0;
    int num_l2_omega=0;
    for(REGION_ID r(0);r<mass.u.m;r++)
    {
        auto& m=mass.u(r);
        auto& v=velocity.u(r);
        for(RANGE_ITERATOR<TV::m> it(m.domain.Thickened(-1));it.Valid();it.Next())
        {
            bool ok=true;
            MATRIX<T,TV::m> dV;
            for(int i=0;i<TV::m;i++)
                for(int j=0;j<TV::m;j++)
                {
                    if(i==j) continue;
                    FACE_INDEX<TV::m> A(i,it.index.Add_Axis(j,1));
                    FACE_INDEX<TV::m> B(i,it.index.Add_Axis(j,-1));
                    FACE_INDEX<TV::m> C(i,A.index.Add_Axis(i,1));
                    FACE_INDEX<TV::m> D(i,B.index.Add_Axis(i,1));
                    if(m(A) && m(B) && m(C) && m(D))
                        dV(i,j)=(T).25*grid.one_over_dX(j)*(v(A)-v(B)+v(C)-v(D));
                    else ok=false;
                }
            if(!ok) continue;
            l2_omega+=Contract_Permutation_Tensor(dV).Magnitude_Squared();
            num_l2_omega++;
        }
    }
    if(num_l2_omega) l2_omega/=num_l2_omega;

    LOG::printf("l2 velocity %P  l2 vorticity %P\n",l2_u,l2_omega);
    LOG::printf("taylor total %P\n",total_taylor);
}
//#####################################################################
// Function Add_Source
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Add_Source(const TV& X0,const TV& n,const TV& V0,const TV& g,
    IMPLICIT_OBJECT<TV>* io,T density,T particles_per_cell,bool owns_io,std::function<bool()> cond)
{
    MPM_PARTICLE_SOURCE<TV>* source=new MPM_PARTICLE_SOURCE<TV>(
        poisson_disk,random,X0,n,V0,g,io);
    T volume=grid.dX.Product()/particles_per_cell;
    T mass=density*volume;
    // NOTE: assumes initial particles are already added.
    source->Seed_Points(particles.X);
    Add_Callbacks(false,"move-particles",[=,this](){
            if(!cond()) return;
            ARRAY<TV> X,V;
            ARRAY<MATRIX<T,TV::m> > dV;
            source->Seed(dt,X,V,&dV);
            for(int i=0;i<X.m;i++)
                Add_Particle(X(i),V(i),dV(i),mass,volume);});
    if(owns_io) destroy.Append([io,source](){delete io;delete source;});

    static int source_id=0;
    this->viewer_output.Add([=,this](){Write_To_File(stream_type,viewer_dir.Frame_File(LOG::sprintf("source-%d",source_id)),*source);},false);
    this->read_output_files.Append([=,this](){
            Read_From_File(viewer_dir.Frame_File(LOG::sprintf("source-%d",source_id)),*source);});
    source_id++;
}
//#####################################################################
// Function Setup_Analytic_Boundary_Conditions
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Setup_Analytic_Boundary_Conditions()
{
    bc_velocity=[this](const TV& X,T t){return analytic_velocity->f(X,t);};
    bc_pressure=[this](const TV& X,T t){return analytic_pressure->f(X,t);};
}
//#####################################################################
// Function Commandline_Analytic_test
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Commandline_Analytic_test()
{
    REGION_ID rid(0);
    Set_Grid(analytic_grid_range*m);
    LOG::printf("domain %P\n",grid.domain);
    density.Append(unit_rho*scale_mass);
    use_analytic_field=true;
    analytic_velocity=new ANALYTIC_VECTOR_PROGRAM<TV>(analytic_u_expr,"U");
    analytic_pressure=new ANALYTIC_SCALAR_PROGRAM<TV>(analytic_p_expr,"p");
    auto V=[=,this](const TV& X){return analytic_velocity->f(X,0);};
    auto dV=[=,this](const TV& X){return analytic_velocity->dX(X,0);};
    if(analytic_phi_expr.size()){
        analytic_level_set=new ANALYTIC_SCALAR_PROGRAM<TV>(analytic_phi_expr,"p");
        auto io=Make_IO(analytic_level_set);
        io->box=grid.domain;
        Seed_Particles(*io,V,dV,density(rid),particles_per_cell);
        delete io;}
    else Seed_Particles(grid.domain,V,dV,density(rid),particles_per_cell);
    Setup_Analytic_Boundary_Conditions();
    if(analytic_bc_types.size()){
        PHYSBAM_ASSERT(analytic_bc_types.size()==2*TV::m);
        for(int i=0;i<2*TV::m;i++){
            switch(analytic_bc_types[i]){
                case 'p':this->side_bc_type(i)=MPM::BC_PERIODIC;break;
                case 'f':this->side_bc_type(i)=MPM::BC_FREE;break;
                case 'n':this->side_bc_type(i)=MPM::BC_NOSLIP;break;
                case 's':this->side_bc_type(i)=MPM::BC_SLIP;break;
                default: PHYSBAM_FATAL_ERROR();break;}}}
    end_frame.Append([=,this](int frame){Check_Analytic_Velocity();});
}
template class STANDARD_TESTS_BASE<VECTOR<float,2> >;
template class STANDARD_TESTS_BASE<VECTOR<float,3> >;
template class STANDARD_TESTS_BASE<VECTOR<double,2> >;
template class STANDARD_TESTS_BASE<VECTOR<double,3> >;
}
