//#####################################################################
// Copyright 2012.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################

#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_INDEX.h>
#include <Core/Grids/GRID.h>
#include <Core/Images/PNG_FILE.h>
#include <Core/Interpolation/FACE_LOOKUP_UNIFORM.h>
#include <Core/Interpolation/INTERPOLATED_COLOR_MAP.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Vectors/VECTOR.h>
#include <limits>
#include <string>
using namespace PhysBAM;

typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,TV::m> IV;

struct FRAME_DATA
{
    GRID<TV> grid;
    ARRAY<T,FACE_INDEX<TV::m> > velocity;
    ARRAY<T,IV> vorticity;
};

void Read_Output_Files(FRAME_DATA& fd,const VIEWER_DIR& viewer_dir)
{
    Read_From_File(viewer_dir.Common_File("grid"),fd.grid);
    Read_From_File(viewer_dir.Frame_File("mac_velocities"),fd.velocity);
}

template<class LOOKUP>
T Cell_Centered_Face_Velocity_Difference(const GRID<TV>& grid,const int axis,const int difference_axis,const LOOKUP& lookup,const IV& cell_index)
{T one_over_four_dh=(T).25*grid.one_over_dX[difference_axis];IV difference_offset=IV::Axis_Vector(difference_axis);
    IV face_1=grid.First_Face_Index_In_Cell(axis,cell_index),face_2=grid.Second_Face_Index_In_Cell(axis,cell_index);
    return one_over_four_dh*(lookup(axis,face_1+difference_offset)+lookup(axis,face_2+difference_offset)-lookup(axis,face_1-difference_offset)-lookup(axis,face_2-difference_offset));}


VECTOR<T,2> Compute_Vorticity(FRAME_DATA& fd)
{
    VECTOR<T,2> r(std::numeric_limits<T>::max(),std::numeric_limits<T>::min());
    RANGE<IV> domain_indices(fd.grid.Cell_Indices());
    domain_indices.Change_Size(-IV::All_Ones_Vector());
    ARRAY<typename TV::SCALAR,FACE_INDEX<TV::m> > lookup(fd.velocity);
    fd.vorticity.Resize(fd.grid.Cell_Indices());
    for(CELL_ITERATOR<TV> it(fd.grid,domain_indices);it.Valid();it.Next()){
        VECTOR<int,2> index=it.Cell_Index();
        T vort_mag=abs(Cell_Centered_Face_Velocity_Difference(fd.grid,1,0,lookup,index)-Cell_Centered_Face_Velocity_Difference(fd.grid,0,1,lookup,index));


        fd.vorticity(index)=vort_mag;
        r(0)=min(r(0),vort_mag);
        r(1)=max(r(1),vort_mag);}
    return r;
}

void Dump(const FRAME_DATA& fd,const std::string& output,T vmin,T vmax)
{
    INTERPOLATED_COLOR_MAP<T> color_map;
    color_map.Initialize_Matlab_Jet(vmin,vmax,false,false);
    ARRAY<VECTOR<T,3>,VECTOR<int,2> > image(fd.vorticity.domain);
    for(RANGE_ITERATOR<2> it(image.domain);it.Valid();it.Next())
        image(it.index)=color_map(fd.vorticity(it.index));
    PNG_FILE<T>::Write(output+".png",image);
}

int main(int argc, char* argv[])
{
    PARSE_ARGS parse_args(argc,argv);
    int frame=0;
    T vmin=std::numeric_limits<T>::min(),vmax=std::numeric_limits<T>::max();
    std::string input="input",output="vort";
    parse_args.Extra(&input,"input","Input directory");
    parse_args.Add("-o",&output,"output","Output image");
    parse_args.Add("-frame",&frame,"frame","Frame");
    parse_args.Add("-vmin",&vmin,"vorticity","Minimum vorticity");
    parse_args.Add("-vmax",&vmax,"vorticity","Maximum vorticity");
    parse_args.Parse();

    FRAME_DATA fd;
    VIEWER_DIR viewer_dir(input);
    viewer_dir.Set(frame);
    Read_Output_Files(fd,viewer_dir);
    auto bound=Compute_Vorticity(fd);
    LOG::printf("Min vort: %P Max vort: %P\n",bound(0),bound(1));
    Dump(fd,output,vmin,vmax);
    return 0;
}

