//#####################################################################
// Copyright 2012.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################

#include <Core/Boundaries/BOUNDARY_PERIODIC.h>
#include <Core/Fourier_Transforms/FFT.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Images/EPS_FILE.h>
#include <Core/Images/PNG_FILE.h>
#include <Core/Interpolation/INTERPOLATED_COLOR_MAP.h>
#include <Core/Interpolation/INTERPOLATION_CURVE.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Vectors/VECTOR.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_PARTICLES.h>
#include <Hybrid_Methods/Iterators/GATHER_SCATTER.h>
#include "../MPM_MAC_DRIVER.h"
#include "../MPM_MAC_EXAMPLE.h"
#include <fstream>
#ifdef USE_OPENMP
#include <omp.h>
#endif
#include <fftw3.h>

using namespace PhysBAM;

typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,TV::m> IV;

template<class TV>
struct FOURIER_EXAMPLE: public MPM_MAC_EXAMPLE<TV>
{
    FOURIER_EXAMPLE()
        :MPM_MAC_EXAMPLE<TV>(STREAM_TYPE::type_float)
    {
    }

    void Initialize() override {}
};

// X is particle seeding in unit box; replicate it in every grid cell
void Replicate_Particles(MPM_PARTICLES<TV>& particles,const GRID<TV>& grid,const ARRAY<TV>& X)
{
    for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next()){
        TV offset=grid.Node_X(it.index);
        for(int i=0;i<X.m;i++){
            int p=particles.Add_Element();
            particles.X(p)=offset+grid.dX*X(i);
            particles.mass(p)=1;
            particles.valid(p)=true;}}
}

void Sample_Box_Regularly(ARRAY<TV>& X,int particles_per_dim)
{
    GRID<TV> init_grid(IV()+particles_per_dim,RANGE<TV>::Unit_Box());
    for(CELL_ITERATOR<TV> it(init_grid);it.Valid();it.Next())
        X.Append(it.Location()+(T).0001);
}

void Sample_Box_Random(RANDOM_NUMBERS<T>& rand,ARRAY<TV>& X,int number_of_particles)
{
    for(int i=0;i<number_of_particles;i++)
    {
        TV Y;
        rand.Fill_Uniform(Y,0,1);
        X.Append(Y);
    }
}

static const std::complex<T>& centered_fft(const ARRAY<std::complex<T>,IV>& f,const IV& index)
{
    IV counts=f.domain.Edge_Lengths();
    return f(wrap(index-counts/2,IV(),counts));
}

int main(int argc, char* argv[])
{
#ifdef USE_OPENMP
    omp_set_num_threads(1);
#endif

    bool use_affine=false;
    bool use_polypic=false;
    T flip=0;
    int order=3;
    int resolution=16;
    int size=256;
    int particles_per_dim=2;
    int irregular_seeding=0;
    bool use_px=false,use_py=false;
    T px=0,py=0;
    int seed=-1;
    std::string output_filename="eigen.png";
    std::string viewer_directory="output";
    bool dump_particles=false;
    bool dump_eigenvalues=false;
    PARSE_ARGS parse_args(argc,argv);
    parse_args.Add("-resolution",&resolution,"num","transfer resolution");
    parse_args.Add("-size",&size,"num","analyze transfer as though this resolution");
    parse_args.Add("-affine",&use_affine,"use affine transfers");
    parse_args.Add("-polypic",&use_polypic,"use PolyPIC transfers");
    parse_args.Add("-flip",&flip,"num","flip ratio");
    parse_args.Add("-order",&order,"order","interpolation order");
    parse_args.Add("-ppd",&particles_per_dim,"num","particles per cell per dimension");
    parse_args.Add("-o",&output_filename,"file.png","filename for output image");
    parse_args.Add("-v",&viewer_directory,"dir","viewer directory");
    parse_args.Add("-irreg",&irregular_seeding,"num","each cell is seeded identicially with num particles");
    parse_args.Add("-px",&px,&use_px,"px","particle x position");
    parse_args.Add("-py",&py,&use_py,"py","particle y position");
    parse_args.Add("-seed",&seed,"seed","random number generator seed (-1 = timer)");
    parse_args.Add("-dump_particles",&dump_particles,"Output particle distribution");
    parse_args.Add("-dump_eigenvalues",&dump_eigenvalues,"Output eigenvalues");
    parse_args.Parse();
    fs::create_directories(viewer_directory);

    PHYSBAM_ASSERT(resolution<=size);
    
    FOURIER_EXAMPLE<TV> example;
    MPM_MAC_DRIVER<TV> driver(example);

    IV center=IV()+resolution/2;
    RANDOM_NUMBERS<T> rand;
    if(seed!=-1) rand.Set_Seed(seed);
    
    example.grid.Initialize(IV()+resolution,RANGE<TV>::Unit_Box());
    ARRAY<TV> unit_X;
    if(use_px&&use_py) unit_X.Append(TV(px,py));
    else if(irregular_seeding) Sample_Box_Random(rand,unit_X,irregular_seeding);
    else Sample_Box_Regularly(unit_X,particles_per_dim);
    Replicate_Particles(example.particles,example.grid,unit_X);

    REGION_ID rid(0);
    example.dt=0;
    example.opt.flip=flip;
    example.opt.use_affine=use_affine;
    example.opt.use_polypic=use_polypic;
    example.Set_Weights(order);
    example.density.Append(1);
    example.viscosity.Append(0);
    ARRAY<RANGE<IV>,REGION_ID> region_ranges(rid+1);
    region_ranges(rid)=example.grid.Cell_Indices(3);
    example.velocity.Resize(region_ranges);
    example.mass.Resize(region_ranges);
    example.gather_scatter=new GATHER_SCATTER<TV>(example.grid,example.simulated_particles);
    example.gather_scatter->threads=example.threads;
    example.gather_scatter->face_weights=example.weights;
    example.gather_scatter->weights=example.weights(0);
    example.gather_scatter->Prepare_Scatter(example.particles);
    example.periodic_boundary.is_periodic.Fill(true);
    example.side_bc_type.Fill(MPM::BC_PERIODIC);
    example.particles.Store_C(example.opt.use_affine || example.opt.use_polypic);
    example.particles.Store_H(example.opt.use_polypic);
    example.velocity.u(rid)(FACE_INDEX<TV::m>(0,center))=1;
    driver.Update_Simulated_Particles();
    driver.Update_Particle_Weights();
    if(use_polypic) driver.Grid_To_Particle_Polypic();
    else driver.Grid_To_Particle();
    driver.Particle_To_Grid();
    
    
    ARRAY<T,IV> row(IV()+size);
    for(RANGE_ITERATOR<TV::m> it(RANGE<IV>::Unit_Box()*resolution);it.Valid();it.Next()){
        IV index=it.index-center;
        for(int i=0;i<TV::m;i++) if(index(i)<0) index(i)+=size;
        row(index)=example.velocity.u(rid)(FACE_INDEX<TV::m>(0,it.index));
        PHYSBAM_ASSERT(!example.velocity.u(rid)(FACE_INDEX<TV::m>(1,it.index)));}
    T max_diff=0;
    for(RANGE_ITERATOR<TV::m> it(RANGE<IV>::Centered_Box()*4);it.Valid();it.Next())
        max_diff=max(max_diff,abs(example.velocity.u(rid)(FACE_INDEX<TV::m>(0,center+it.index))-example.velocity.u(rid)(FACE_INDEX<TV::m>(0,center-it.index))));
    LOG::printf("symmetric error: %g\n",max_diff);
    ARRAY<std::complex<T>,IV> out(row.domain);

    GRID<TV> fft_grid(out.domain.Edge_Lengths(),RANGE<TV>::Unit_Box());
    FFT<TV> fft;
    fft.Transform(row,out);

    INTERPOLATED_COLOR_MAP<T> icm;
    icm.colors.Add_Control_Point(1.00001,VECTOR<T,3>(1,1,1));
    icm.colors.Add_Control_Point(1,VECTOR<T,3>(.5,0,0));
    icm.colors.Add_Control_Point(1-.01,VECTOR<T,3>(1,0,0));
    icm.colors.Add_Control_Point(1-.02,VECTOR<T,3>(1,.5,0));
    icm.colors.Add_Control_Point(1-.04,VECTOR<T,3>(1,1,0));
    icm.colors.Add_Control_Point(1-.08,VECTOR<T,3>(0,1,0));
    icm.colors.Add_Control_Point(1-.16,VECTOR<T,3>(0,1,1));
    icm.colors.Add_Control_Point(1-.32,VECTOR<T,3>(0,0,1));
    icm.colors.Add_Control_Point(1-.64,VECTOR<T,3>(.5,0,1));
    icm.colors.Add_Control_Point(0,VECTOR<T,3>(0,0,0));

    ARRAY<VECTOR<T,3>,VECTOR<int,2> > bar(VECTOR<int,2>(1000,1));
    for(int i=0;i<1000;i++)
        bar(VECTOR<int,2>(i,0))=icm.colors.Value((i/(T)999)*(icm.mx-icm.mn)+icm.mn);
    PNG_FILE<T>::Write("bar.png",bar);

    ARRAY<VECTOR<T,3>,IV> image(out.domain);
    for(RANGE_ITERATOR<TV::m> it(image.domain);it.Valid();it.Next())
        image(it.index)=icm(abs(centered_fft(out,it.index)));

    PNG_FILE<T>::Write(output_filename,image);

    if(dump_particles){
        VIEWER_DIR viewer_dir(viewer_directory);
        VIEWER_OUTPUT viewer_output(STREAM_TYPE::type_float,viewer_dir);
        viewer_output.Add("debug_particles",DEBUG_PARTICLES<TV>::Get(),false,{});
        for(int i=0;i<example.particles.X.m;i++)
            Add_Point(example.particles.X(i),VECTOR<T,3>(1,0,0));
        Flush_Frame("particles");
        Flush_Frame("end");
    }

    if(dump_eigenvalues){
        std::ofstream eig_x((viewer_directory+"/eig-x.txt").c_str());
        std::ofstream eig_y((viewer_directory+"/eig-y.txt").c_str());
        std::ofstream eig_xy((viewer_directory+"/eig-xy.txt").c_str());
        eig_x<<"waven x eig\n";
        eig_y<<"waven x eig\n";
        eig_xy<<"waven x eig\n";
        int shift=size/2;
        for(int i=out.domain.min_corner(0);i<out.domain.max_corner(0);++i){
            eig_x<<i-shift<<" "<<(T)(i-shift)/size<<" "<<centered_fft(out,IV(i,shift)).real()<<"\n";
            eig_y<<i-shift<<" "<<(T)(i-shift)/size<<" "<<centered_fft(out,IV(shift,i)).real()<<"\n";
            eig_xy<<i-shift<<" "<<(T)(i-shift)/size<<" "<<centered_fft(out,IV(i,i)).real()<<"\n";}}

    return 0;
}

