//#####################################################################
// Copyright 2012.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Analytic_Tests/ANALYTIC_SCALAR.h>
#include <Core/Analytic_Tests/ANALYTIC_VECTOR.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/NODE_ITERATOR.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Projection/CUT_CELL_PROJECTION.h>
#include <Core/Vectors/VECTOR.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>

using namespace PhysBAM;

typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,TV::m> IV;

int main(int argc, char* argv[])
{
    CUT_CELL_PROJECTION<TV> proj;
    // Negative means inside fluid for both surface and object
    std::string surface_str="p=-1;",object_str="p=-1",u_str="u=0;v=0;",p_str="p=0;",bc_str="pppp";
    int ghost=3;
    RANGE<TV> domain=RANGE<TV>::Centered_Box();
    IV size;
    int resolution=32;
    T density=100;
    T dt=.01;

    PARSE_ARGS parse_args(argc,argv);
    parse_args.Add("-sf",&surface_str,"prog","free surface level set");
    parse_args.Add("-ob",&object_str,"prog","object level set");
    parse_args.Add("-u",&u_str,"prog","solution velocity");
    parse_args.Add("-p",&p_str,"prog","solution pressure");
    parse_args.Add("-bc",&bc_str,"prog","boundary conditions [fsp]");
    parse_args.Add("-print_matrix",&proj.print_matrix,"print system matrix and vectors");
    parse_args.Add("-print_residual",&proj.print_residual,"print residual after linear solve");
    parse_args.Add_Not("-no_precon",&proj.use_preconditioner,"disable preconditioner");
    parse_args.Add("-test_system",&proj.test_system,"print system matrix and vectors");
    parse_args.Add("-solver_tolerance",&proj.solver_tolerance,"tol","Krylov tolerance");
    parse_args.Add("-solver_iterations",&proj.solver_iterations,"num","Krylov iterations");
    parse_args.Add("-resolution",&resolution,"num","grid resolution");
    parse_args.Add("-lo",&domain.min_corner,"vec","grid domain lower corner");
    parse_args.Add("-hi",&domain.max_corner,"vec","grid domain upper corner");
    parse_args.Add("-rho",&density,"value","fluid density");
    parse_args.Add("-dt",&dt,"value","time step size");
    parse_args.Parse();

    GRID<TV> grid(IV()+resolution,domain);
    GRID<TV> phi_grid=grid.Get_MAC_Grid_At_Regular_Positions();
    VIEWER_DIR viewer_dir("proj");
    VIEWER_OUTPUT viewer_output(STREAM_TYPE::type_float,viewer_dir);
    if(TV::m==3) viewer_output.template Add_Axes<TV::m>("axes",{});
    viewer_output.Add("debug_particles",DEBUG_PARTICLES<TV>::Get(),false,{});
    viewer_output.Add("grid",grid,true,{.draw=(TV::m==2)});

    ARRAY<T,FACE_INDEX<TV::m> > u(grid.Cell_Indices(ghost));
    viewer_output.Add("mac_velocities","grid",u,false,{});
    ARRAY<bool,FACE_INDEX<TV::m> > valid_u(grid.Cell_Indices(ghost));
    ARRAY<T,IV> object_phi(grid.Node_Indices(ghost));
    ARRAY<T,IV> surface_phi(grid.Node_Indices(ghost));
    ARRAY<T,IV> pressure(grid.Cell_Indices(ghost));
    ARRAY<bool,IV> valid_p(grid.Cell_Indices(ghost));
    proj.object_phi=&object_phi;
    proj.surface_phi=&surface_phi;
    proj.valid_u=&valid_u;
    proj.valid_p=&valid_p;
    proj.pressure=&pressure;

    ANALYTIC_VECTOR_PROGRAM<TV> av(u_str,"U");
    ANALYTIC_SCALAR_PROGRAM<TV> ap(p_str,"p");
    ANALYTIC_SCALAR_PROGRAM<TV> as(surface_str,"p");
    ANALYTIC_SCALAR_PROGRAM<TV> ao(object_str,"p");
    proj.bc_v=[&av](const TV& X){return av.f(X,0);};
    proj.bc_p=[&ap](const TV& X){return ap.f(X,0);};

    for(NODE_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next()){
        object_phi(it.index)=ao.f(it.Location(),0);
        surface_phi(it.index)=as.f(it.Location(),0);}

    for(FACE_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next()){
        TV X=it.Location();
        TV V=av.f(X,0),dp=ap.dX(X,0);
        u(it.face)=(V+dt*dp/density)(it.face.axis);}

    for(int i=0;i<2*TV::m;i++){
        if(bc_str[i]=='p') proj.bc_type(i)=0;
        else if(bc_str[i]=='f') proj.bc_type(i)=1;
        else if(bc_str[i]=='s') proj.bc_type(i)=2;
        else PHYSBAM_FATAL_ERROR();}

    Dump_Levelset(phi_grid,at_cells,object_phi,VECTOR<T,3>(1,1,0));
    Dump_Levelset(phi_grid,at_cells,surface_phi,VECTOR<T,3>(1,0,1));
    Flush_Frame("before projection");

    proj.Cut_Cell_Projection(grid,u,density,dt);

    Dump_Levelset(phi_grid,at_cells,object_phi,VECTOR<T,3>(1,1,0));
    Dump_Levelset(phi_grid,at_cells,surface_phi,VECTOR<T,3>(1,0,1));
    Flush_Frame("after projection");
    
    T u_l2=0,u_inf=0,u_max=0;
    int u_cnt=0;
    for(FACE_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next()){
        if(!valid_u(it.face)) continue;
        TV X=it.Location();
        T v=av.f(X,0)(it.face.axis),w=u(it.face),e=abs(v-w);
        Add_Ball(X,e,VECTOR<T,3>(1,0,0));
        u_inf=max(u_inf,e);
        u_max=max(u_max,abs(v));
        u_l2+=e*e;
        u_cnt++;}
    u_l2=sqrt(u_l2/u_cnt);
    LOG::printf("u: %.3f %.3f   %.3f\n",u_inf,u_l2,u_max);
    Dump_Levelset(phi_grid,at_cells,object_phi,VECTOR<T,3>(1,1,0));
    Dump_Levelset(phi_grid,at_cells,surface_phi,VECTOR<T,3>(1,0,1));
    Flush_Frame("velocity error");

    T p_l2=0,p_inf=0,p_max=0;
    int p_cnt=0;
    for(CELL_ITERATOR<TV> it(grid,ghost);it.Valid();it.Next()){
        if(!valid_p(it.index)) continue;
        TV X=it.Location();
        T p=ap.f(X,0),q=pressure(it.index),e=abs(p-q);
        Add_Ball(X,e,VECTOR<T,3>(0,1,0));
        p_inf=max(p_inf,e);
        p_max=max(p_max,abs(p));
        p_l2+=e*e;
        p_cnt++;}
    p_l2=sqrt(p_l2/p_cnt);
    LOG::printf("p: %.3f %.3f   %.3f\n",p_inf,p_l2,p_max);
    Dump_Levelset(phi_grid,at_cells,object_phi,VECTOR<T,3>(1,1,0));
    Dump_Levelset(phi_grid,at_cells,surface_phi,VECTOR<T,3>(1,0,1));
    Flush_Frame("pressure error");

    return 0;
}

