import argparse
import subprocess as sp
import numpy as np
import os
import sys

def parse_args(): # Default to BDF2 on
    parser = argparse.ArgumentParser(description="Options to run MPM_MAC tests")
    parser.add_argument("-no", "--test_case_number", type=int, default=41)
    parser.add_argument("-threads", "--threads", type=int, default=1)
    parser.add_argument("-d", "--dimension", type=int, default=2)
    parser.add_argument("-n", "--last_frame", type=int, default=24)
    parser.add_argument("-C", "--C_update_ver", type=int, default=1)
    parser.add_argument("-reseed", "--reseed", type=int, default=0)
    parser.add_argument("-adv", "--advection", type=int, default=1)
    parser.add_argument("-f", "--force", type=int, default=1)
    parser.add_argument("-p", "--pressure", type=int, default=1)
    parser.add_argument("-mu", "--viscosity", type=float, default=0.2)
    parser.add_argument("-ft", "--force_timepoint", type=float, default=1)
    parser.add_argument("-k", "--adv_v_timepoint", type=float, default=0.5)
    parser.add_argument("-dt0", "--frame_dt", type=float, default=-1)
    parser.add_argument("-dx0", "--base_dx", type=int, default=-1)
    parser.add_argument("-rf", "--refine_scales", type=int, default=4)
    parser.add_argument("-scheme", "--integration_scheme", default="BDF2")
    parser.add_argument("-cjh", "--cjh", type=int, default=0)
    parser.add_argument("-cint", "--cint", type=int, default=1)
    parser.add_argument("-hint", "--hint", type=int, default=1)
    parser.add_argument("-pred_p", "--pred_p", type=int, default=0)
    parser.add_argument("-ab_p", "--ab_p", type=int, default=0)
    parser.add_argument("-init_prev", "--init_prev", type=int, default=0)
    parser.add_argument("-qpic", "--qpic", type=int, default=1)
    parser.add_argument("-spline", "--spline", type=int, default=3)
    parser.add_argument("-flip", "--flip", type=float, default=0)
    parser.add_argument("-o", "--output_dir", default="")
    parser.add_argument("-c", "--comments", default="")
    parser.add_argument("-fn", "--filename", default="")
    parser.add_argument("-fm", "--format", default="")
    parser.add_argument("-check", "--check", type=int, default=0)
    parser.add_argument("-err", "--err", default="uvp2inf")
    return parser.parse_args()
    
def extract_err(output): # input: string; output: dict of float / double
    oList = output.split()
    oList.reverse()
    errs={}
    if "u" in args.err:
        u_idx = oList.index("grid")
        if "inf" in args.err:
    	    inf_idx = u_idx - 4
    	    errs["uinf"]=float(oList[inf_idx])
        if "2" in args.err:
    	    l2_idx = u_idx - 6
    	    errs["u2"]=float(oList[l2_idx])
    if "v" in args.err:
        v_idx = oList.index("particle")
        if "inf" in args.err:
    	    inf_idx = v_idx - 4
    	    errs["vinf"]=float(oList[inf_idx])
        if "2" in args.err:
    	    l2_idx = v_idx - 6
    	    errs["v2"]=float(oList[l2_idx])
    if "p" in args.err:
        p_idx = oList.index("pressure")
        if "inf" in args.err:
    	    inf_idx = p_idx - 3
    	    errs["pinf"]=float(oList[inf_idx])
        if "2" in args.err:
    	    l2_idx = p_idx - 5
    	    errs["p2"]=float(oList[l2_idx])
    return errs
    
if __name__ == '__main__':
    args = parse_args()
    Cname = {1:"C1", 2:"C2", 3:"Cinv"}
    Switch_name = {0: "OFF", 1: " ON"}
    viscosity_on = 0
    if args.viscosity != 0:
    	viscosity_on = 1
    errmsg = f"Test {args.test_case_number}; Advection: {Switch_name[args.advection]}; Force: {Switch_name[args.force]}; Pressure: {Switch_name[args.pressure]}; Viscosity: {Switch_name[viscosity_on]}; "
    errmsg += args.comments
    errmsg += "\n"
    err_types=[]
    if 'u' in args.err and '2' in args.err:
        err_types.append('u2')
    if 'u' in args.err and 'inf' in args.err:
        err_types.append('uinf')
    if 'v' in args.err and '2' in args.err:
        err_types.append('v2')
    if 'v' in args.err and 'inf' in args.err:
        err_types.append('vinf')
    if 'p' in args.err and '2' in args.err:
        err_types.append('p2')
    if 'p' in args.err and 'inf' in args.err:
        err_types.append('pinf')
    if err_types==[]:
        raise NotImplementedError("Please set -err correctly (u,v,p followed by 2,inf)")
    err_fullnames={'u2':'    grid L-2  ', 'uinf':'    grid L-inf', 'v2': 'particle L-2  ', 'vinf': 'particle L-inf', 'p2': 'pressure L-2  ', 'pinf': 'pressure L-inf'}
    err_latex={'u2':'$\\uu_{ia}$-$L_2$', 'uinf':'$\\uu_{ia}$-$L_\\infty$', 'v2': '$\\vv_p$-$L_2$', 'vinf': '$\\vv_p$-$L_\\infty$', 'p2': '$p$-$L_2$', 'pinf': '$p$-$L_\\infty$'}
    print(f"qpic {args.qpic}; sp {args.spline}; chint {args.cint}; no {args.test_case_number};")
    for C_order in [1,2,3]:
        if args.C_update_ver and C_order != args.C_update_ver:
            continue
        errs = {'u2':[], 'uinf':[], 'v2':[], 'vinf':[], 'p2':[], 'pinf':[]}
        for refine in 2**np.arange(args.refine_scales):
            command = ["./mpm_mac",
                       str(args.test_case_number),
                       "-C_order", str(C_order),
                       "-refine", str(int(refine))]
            command.append("-last_frame")
            if args.last_frame == 0:
                command.append("1")
                command.append("-one_substep")
            else:
                command.append(str(args.last_frame))
            if "u" not in args.err:
                command.append("-no_grid_err")
            if "v" not in args.err:
                command.append("-no_particle_err")
            if "p" not in args.err:
                command.append("-no_pressure_err")
            command.append("-threads")
            if args.threads==0:
                command.append(str(int(refine)))
            else:
                command.append(str(args.threads))
            if args.output_dir:
                command.append("-o")
                command.append(args.output_dir)
            if args.integration_scheme == "BDF2":
                command.append("-use_bdf2")
            if args.reseed != 0:
                command.append("-reseed")
            if args.advection == 0:
                command.append("-no_move")
            if args.force == 0:
                command.append("-no_f")
            if args.pressure == 0:
                command.append("-no_p")
            if args.viscosity == 0:
                command.append("-no_mu")
            else:
                command.append("-mu")
                command.append(str(args.viscosity))
            if args.force_timepoint != 0:
                command.append("-f_timepoint")
                command.append(str(args.force_timepoint))
            if args.adv_v_timepoint != 0:
                command.append("-adv_v_timepoint")
                command.append(str(args.adv_v_timepoint))
            if args.frame_dt > 0:
                command.append("-frame_dt")
                command.append(str(args.frame_dt))
            if args.base_dx > 0:
                command.append("-resolution")
                command.append(str(args.base_dx))
            if args.flip > 0:
                command.append("-flip")
                command.append(str(args.flip))
            if args.check == 0:
                command.append("-no_check")
            if args.cjh != 0:
                command.append("-C_JH")
            if args.cint != 0:
                command.append("-C_INT")
            if args.hint != 0:
                command.append("-H_INT")
            if args.pred_p != 0:
                command.append("-pred_p")
            if args.ab_p != 0:
                command.append("-ab_p")
            if args.qpic != 0:
                command.append("-polypic6")
            if args.init_prev != 0:
                command.append("-init_prev")
            if args.spline==3:
                command.append("-order")
                command.append("3")
            elif args.spline!=2:
                raise NotImplementedError
            if args.dimension==3:
                command.append("-3d")
            elif args.dimension!=2:
                raise NotImplementedError
            p = sp.Popen(command, stdout=sp.PIPE)
            output, err = p.communicate()
            if err:
                print(err.decode('ascii'))
            output = output.decode('ascii')
            if args.check:
                print(output)
            print(f"{Cname[C_order]};\tRefine: {refine};", end=' ')
            err_dict = extract_err(output)
            for err_type in err_types:
                print(f"{err_fullnames[err_type]} error: {err_dict[err_type]:.4e}", end=' ')
                if errs[err_type] != []:
                    order = np.log2(errs[err_type][-1]/err_dict[err_type])
                    print(f"Order: {order:.2f};", end=' ')
                else:
                    print("            ", end=' ')
                errs[err_type].append(err_dict[err_type])
            print()
        for err_type in err_types:
            errs[err_type] = np.array(errs[err_type])
        orders={}
        for err_type in err_types:
            orders[err_type] = np.log2(errs[err_type][:-1]/errs[err_type][1:])
        
        single_errmsg = f"{Cname[C_order]}"
        for err_type in err_types:
            single_errmsg += f"\t{err_fullnames[err_type]} error: "
            for err in errs[err_type]:
                single_errmsg += f"{err:.4e} "
            single_errmsg += f"\t\t{err_fullnames[err_type]} Order: "
            for order in orders[err_type]:
                single_errmsg += f"{order:.2f} "
            single_errmsg += '\n'
        errmsg += single_errmsg
        
        if "t" in args.format:
            single_errmsg = "\n% command: python3 "
            for arg in sys.argv:
                single_errmsg += f"{arg} "
            single_errmsg += "\n"
            single_errmsg += "\\begin{tabular}{|c|"
            for err_type in err_types:
                single_errmsg += "l c|"
            single_errmsg += "}\n\\hline\n"
            for err_type in err_types:
                single_errmsg += "&\multicolumn{2}{c|}{" + f"{err_latex[err_type]}" + "} "
            single_errmsg += "\\\\\nresolution "
            for err_type in err_types:
                single_errmsg += "& error & order "
            single_errmsg += "\\\\\n\\hline\n"
            for i in range(args.refine_scales):
                single_errmsg += f"{(args.base_dx if args.base_dx != -1 else 32)*2**i} "
                for err_type in err_types:
                    estr = f"{errs[err_type][i]:.2e}"
                    basestr = estr[0:4]
                    powerstr = -int(estr[-2:])
                    if i:
                        single_errmsg += f"& ${basestr}\\times 10^{ {powerstr} }$ & {orders[err_type][i-1]:.2f} "
                    else:
                        single_errmsg += f"& ${basestr}\\times 10^{ {powerstr} }$ & "
                single_errmsg += f"\\\\\n\\hline\n"
            single_errmsg += "\\end{tabular}\n"
            errmsg += single_errmsg
        
        if args.advection == 0:
            break
    errmsg += "\n"
    
    if args.filename and os.path.exists(os.path.join('2023-test-logs',args.filename)):
        with open(os.path.join('2023-test-logs',args.filename), "r") as f:
            errmsg = f.read() + errmsg
    if args.filename:
        with open(os.path.join('2023-test-logs',args.filename), "w") as f:
            f.write(errmsg)
                  
